import os
import sys
import pickle
import getpass

from sqlalchemy import *
from subprocess import call
from sqlalchemy import event
from subprocess import Popen, PIPE
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base

if sys.version_info.major > 2:
    from core_framework.db_engine import DbEngine

schema = 'eoglasna'

framework_folder = r'C:\Users\{}\Documents\crawler_framework'.format(getpass.getuser())
python_path = r'{}\py_config.pkl'.format(framework_folder)

BasePstg = declarative_base()

# creating schema if it doesn't exist
event.listen(BasePstg.metadata, 'before_create', DDL("CREATE SCHEMA IF NOT EXISTS {}".format(schema)))


# ======================================================================
#                               TABLES
# ======================================================================
class CourtList(BasePstg):
    """All gathered court links"""
    __tablename__ = 'court_list'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('court_list_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    court_group = Column(String(500))
    court_url = Column(String(1024))
    court_name = Column(String(1024))
    court_id = Column(Integer())

    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class DetailsRbr(BasePstg):
    """Assigns unique id for every details_link"""
    __tablename__ = 'details_rbr'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('det_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    rbr_seq = Sequence('rbr_id_seq',minvalue=2421855, metadata=BasePstg.metadata, schema=schema)
    rbr = Column(
        BigInteger, rbr_seq,
        server_default=rbr_seq.next_value(), index=True)

    sha = Column(String(64))  # hash value of the row
    details_link = Column(String(1024), index=True)
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class SearchResults(BasePstg):
    """History of all gathered search results"""
    __tablename__ = 'search_results'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('search_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    rid = Column(Integer())
    decision = Column(String(500), index=True)  # BROJ_RJESENJA
    all_decisions = Column(String(4000))  # link on all decisions with same decision number
    badge = Column(String(500))  # VRSTA_RJESENJA badge is flag on web that says decision_type
    decision_type = Column(String(500))  # is evaluated decision type using decision
    court_name = Column(String(1024))  # NAZIV_SUDA
    court_id = Column(Integer(), index=True)  # SUD_ID
    court_group = Column(String(1024))
    announce_date = Column(DateTime)  # DATUM_OBJAVE
    erase_date = Column(DateTime)  # DATUM_BRISANJA

    participant = Column(String(4000))  # SUDIONIK
    vat_number = Column(String(4000))  # OIB
    participant_type = Column(String(1024))  # VRSTA_SUDIONIKA

    details_link = Column(String(1024), index=True)  # DETALJI_OBJAVE_LINK
    decision_title = Column(String(4000))  # DETALJI_OBJAVE_TEXT
    status = Column(Integer())  # DETALJI --CRAWLED STATUS
    status_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class CourtRss(BasePstg):
    """History of all gathered search results"""
    __tablename__ = 'court_rss'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('rss_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    court_name = Column(String(1024))
    court_id = Column(Integer(), index=True)
    court_group = Column(String(1024))
    title = Column(String(4000))
    details_link = Column(String(1024), index=True)
    description = Column(String(4000))
    description_type = Column(String(500))
    status = Column(Integer())
    status_date = Column(DateTime)
    announce_date = Column(DateTime)
    erase_date = Column(DateTime)
    author = Column(String(1024))

    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class Details(BasePstg):
    """History of all gathered details"""
    __tablename__ = 'details'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('details_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    rbr = Column(BigInteger())
    rid = Column(Integer())
    decision = Column(String(500), index=True)
    decision_title = Column(String(4000))  # DETALJI_OBJAVE_TEXT
    all_decisions = Column(String(4000))
    badge = Column(String(500))  # VRSTA_RJESENJA badge is flag on web that says decision_type
    decision_type = Column(String(500))  # is evaluated decision type using decision

    court_name = Column(String(1024))
    court_id = Column(Integer(), index=True)
    court_group = Column(String(1024))
    details_link = Column(String(1024), index=True)
    announce_date = Column(DateTime)
    erase_date = Column(DateTime)
    ecase = Column(String(1024))
    insolvent = Column(String(1024))

    doc_link = Column(String(1240), index=True)

    public_notary = Column(String(3000))
    public_notary_address = Column(String(3000))
    public_notary_case = Column(String(1024))

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class Participants(BasePstg):
    """History of all gathered participants"""
    __tablename__ = 'participants'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('part_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    rbr = Column(BigInteger())
    vat_number = Column(String(1240))
    participant = Column(String(3000))
    participant_role = Column(String(1240))
    address = Column(String(3000))
    details_link = Column(String(1240), index=True)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class Documents(BasePstg):
    """History of all downloaded documents"""
    __tablename__ = 'documents'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('doc_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    rbr = Column(BigInteger())
    decision = Column(String(500), index=True)
    court_name = Column(String(1024))
    court_id = Column(Integer(), index=True)
    court_group = Column(String(1024))

    details_link = Column(String(1024), index=True)
    doc_link = Column(String(1240))
    dl_status = Column(Integer(), index=True)
    dl_retries = Column(Integer())
    dl_date = Column(DateTime)

    doc_size = Column(BigInteger())
    doc_sha = Column(String(64))
    doc_name = Column(String(1240))
    doc_name_raw = Column(String(1240))
    doc_path = Column(String(2480))

    up_status = Column(Integer(), index=True)
    up_path = Column(String(2480))
    up_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class ArchiveTables(BasePstg):
    """History of all xlsx archive tables thata re published"""
    __tablename__ = 'archive_tables'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('arch_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    table_name = Column(String(1024), index=True)
    badge = Column(String(500))  # Vrsta objave
    court_name = Column(String(1024))  # Sud
    decision = Column(String(500), index=True)  # Oznaka spisa
    decision_title = Column(String(4000))  # Naslov

    debtor = Column(String(4000))  # Stečajni dužnik
    debtor_address = Column(String(4000))  # Adresa stečajnog dužnika
    vat_number = Column(String(128))  # OIB stečajnog dužnika
    participants = Column(VARCHAR(1000000))  # Sudionici
    announce_date = Column(DateTime)  # Početni dan objave
    erase_date = Column(DateTime)  # Posljednji dan objave
    details_link = Column(String(1024), index=True)  # Oglas (link)

    doc_link = Column(String(1240))  # Dokumenti (link)
    doc_name = Column(String(1240))  # Dokumenti (datoteke)
    ecase = Column(String(1024))  # e-Predmet (link)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created

# ======================================================================
#                               LOGGERS
# ======================================================================
class CrawlerSemaphore(BasePstg):
    __tablename__ = 'l_crawl'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('crawl_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    crawl_date = Column(Date, index=True)
    crawl_status = Column(Integer())  # NULL = nothing passed, 1 = all passed
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class CrawlerInfoLog(BasePstg):
    __tablename__ = 'l_info'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('loginfo_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    runtime_id = Column(String(64), index=True)  # hash value of single run cycle of program
    sha = Column(String(64), index=True)  # hash value of the row
    start_time = Column(DateTime)
    end_time = Column(DateTime)
    step_name = Column(String(256))
    status = Column(Integer())  # 1 = success  2=errors
    errors = Column(Integer())  # >0 there was and error
    note = Column(String(1024))
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class CrawlerErrorLog(BasePstg):
    __tablename__ = 'l_errors'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('logerr_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    runtime_id = Column(String(64), index=True)  # hash value of single run cycle of program
    sha = Column(String(64), index=True)  # hash value of the row
    start_time = Column(DateTime)
    end_time = Column(DateTime)
    step_name = Column(String(256))
    err_type = Column(String(3072))
    err_line = Column(Integer())
    err_desc = Column(String(3072))
    err_cnt = Column(Integer())
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


def deploy(conn_id):
    db = DbEngine()
    engine = db.connect(conn_id, echo=False)
    # BasePstg.metadata.drop_all(engine)
    BasePstg.metadata.create_all(engine)


if __name__ == '__main__':
    deploy(1)