import re
import sys
import hashlib
import urllib.parse
import pandas as pd

base_url = 'https://e-oglasna.pravosudje.hr'


def court_list_parse(html_bs):
    court_list = []

    def parse(html):
        court_group = html.find('h5')
        if court_group is not None:
            court_group = court_group.text
            links = html.find_all('a', {'href': True})
            for link in links:
                court_url = link['href']
                court_name = link.text.strip()
                court_list.append({'court_group': court_group, 'court_url': court_url, 'court_name': court_name})

    section = html_bs.find('section', {'class': 'court-list'})
    if section is None:
        error = {'err_desc': 'court_list_parser: cannot find section element during gathering courts'}
        return error

    sections = section.find_all('div', {'class': 'section'})
    for section in sections:
        court_group_multi = section.find_all('div', {'class': 'print-break-none'})
        if court_group_multi:
            for court_group in court_group_multi:
                parse(court_group)
        else:
            parse(section)

    if not court_list:
        error = {'err_desc': 'court_list_parser: court list is empty'}
        return error

    return court_list


def sudionici_fix(sudionik_data):
    """sudionici fix različiti formati gore parser je rješavao oibove koji se nalaze u sivim kvadratćima
    ovaj fix će izvući oib iz naziva sudionika i maknuti ga od tamo radi se zasebno kako se
    hash vrijednost nebi promjenila za već postojeće zapise u bazi"""

    sudionik = sudionik_data.get('sudionik')
    oib = sudionik_data.get('oib')

    regex_patters = ['OIB:\s+(\d+)', '\|\s+(\d+)', 'OIB\s+(\d+)']
    sub_paterns = ['OIB:\s+\d+', '\|\s+\d+', 'OIB\s+\d+']

    for regex_pattern in regex_patters:
        if sudionik is not None:
            oib_search = re.search(regex_pattern, sudionik, re.DOTALL | re.IGNORECASE)
            if oib_search is not None:
                oib = oib_search.group(1)

            for sub_patern in sub_paterns:
                sudionik = re.sub(sub_patern, '', sudionik)

            for i in range(6):
                sudionik = sudionik.strip(',').strip().strip('|').strip()

    sudionik_data.update({'sudionik': sudionik, 'oib': oib})
    return sudionik_data


def parse(html, courts, rid, err_dir=None):
    all_data, old_data, sha_list = [], [], []
    try:
        result_selection = html.find('section', {'class': 'rezultati'})
        ul = result_selection.find('ul')

        lis = ul.find_all('li')
        for li in lis:

            # broj rješenja
            broj_rjesenja = None
            search_link_broj = li.find('a', {'title': 'Sve objave vezane za ovaj premet'})
            if search_link_broj is not None:
                broj_rjesenja = search_link_broj.text.strip()
                if broj_rjesenja == '':
                    broj_rjesenja = None

            # detalji objave
            detalji_objave_link = None
            detalji_objave_text = None
            search_detalji = li.find('a', {'title': 'Detalji objave'})
            if search_detalji is not None:
                detalji_objave_link = search_detalji['href']
                detalji_objave_text = search_detalji.text.strip()

            # naziv suda
            naziv_suda, sud_id = None, None
            search_sud = li.find('a', {'title': 'Sve objave ovog suda'})
            if search_sud is not None:
                naziv_suda = search_sud.text.strip()
                if courts is not None:
                    get_sud_data = courts.get(naziv_suda)
                    if get_sud_data is not None:
                        sud_id = get_sud_data.get('sud_id')

            # vrsta rješenja
            vrsta_rjesenja = None
            search_badge = li.find('span', {'class': 'badge'})
            if search_badge is not None:
                vrsta_rjesenja = search_badge.text.strip()

            # Sudionici
            sudionici, sudionici_sha = list(), list()
            divs = li.find_all('div')
            divs = [div for div in divs if re.search('Sudionici', div.text, re.DOTALL | re.MULTILINE | re.IGNORECASE)]
            for div in divs:
                dblock_search = div.find_all('div', {'class': 'd-block'})
                for dblock in dblock_search:
                    sha_dblock = hashlib.sha3_256(dblock.encode()).hexdigest()
                    if sha_dblock not in sudionici_sha:
                        sudionik, oib, vrsta_sudionika = None, None, None
                        sudionik_search = dblock.find('span', {'data-content': 'Sudionici predmeta'})
                        if sudionik_search is not None:
                            sudionik = sudionik_search.text
                        oib_search = dblock.find('small', {'data-content': 'OIB sudionika'})

                        if oib_search is not None:
                            oib_numbers = re.search("(\d+)", oib_search.text)
                            if oib_numbers is not None:
                                oib = oib_numbers.group(1)

                        duznik_search = dblock.find('span', {'data-content': 'Sudionik je dužnik'})
                        if duznik_search is not None:
                            vrsta_sudionika = duznik_search.text

                        sudionici.append({'sudionik': sudionik, 'oib': oib, 'vrsta_sudionika': vrsta_sudionika})
                        sudionici_sha.append(sha_dblock)

            # Datum objave
            datum_objave = None
            datum_objave_search = re.search('<small class="text-muted d-block">Datum i vrijeme objave</small>.*?<div class="d-block"><span class="m-date">(.*?)</span></div>', li.decode(), re.MULTILINE | re.IGNORECASE | re.DOTALL)
            if datum_objave_search is not None:
                datum_objave = datum_objave_search.group(1)
                datum_objave = pd.to_datetime(datum_objave, dayfirst=True, errors='coerce')
                if pd.isna(datum_objave) is True:
                    datum_objave = None

            # Posljedni datum objave
            datum_brisanja = None
            datum_brisanja_search = re.search('<small class="text-muted d-block">Posljednji datum objave</small>.*?<div class="d-block"><span class="m-date">(.*?)</span></div>', li.decode(), re.MULTILINE | re.IGNORECASE | re.DOTALL)
            if datum_brisanja_search is not None:
                datum_brisanja = datum_brisanja_search.group(1)
                datum_brisanja = pd.to_datetime(datum_brisanja, dayfirst=True, errors='coerce')
                if pd.isna(datum_brisanja) is True:
                    datum_brisanja = None

            # Dokument link
            doc_link = None
            doc_link_search = li.find('a', {'data-content': 'Preuzimanje dokumenta objave', 'href': True})
            if doc_link_search is not None:
                doc_link = doc_link_search['href']

            if doc_link_search is None:
                doc_link_search = li.find('a', {'data-content': 'Preuzimanje objavljenih dokumenata', 'href': True})
            if doc_link_search is not None:
                doc_link = doc_link_search['href']

            base_data = {'broj_rjesenja': broj_rjesenja, 'detalji_objave_link': detalji_objave_link,
                         'detalji_objave_text': detalji_objave_text, 'vrsta_rjesenja': vrsta_rjesenja,
                         'naziv_suda': naziv_suda, 'datum_objave': datum_objave,
                         'datum_brisanja': datum_brisanja, 'sud_id': sud_id, 'doc_link': doc_link, 'rid': rid}

            # doc_fix dodavanje base web url kako bi klijenti mogli kliknuti na puni link
            doc_link_new = None
            if doc_link is not None:
                doc_link_new = ''.join(['https://e-oglasna.pravosudje.hr', doc_link])

            if sudionici:
                for sudionik in sudionici:

                    row = base_data.copy()
                    row.update(sudionik)
                    sha = hashlib.sha3_256(str(row).encode()).hexdigest()

                    sudionik_fix = sudionici_fix(sudionik)
                    row.update(sudionik_fix)

                    row.update({'sha': sha, 'doc_link': doc_link_new})

                    if sha not in sha_list:
                        all_data.append(row)
                        sha_list.append(sha)

            else:
                solo_data = base_data.copy()
                sha = hashlib.sha3_256(str(solo_data).encode()).hexdigest()

                solo_data.update({'sha': sha, 'doc_link': doc_link_new})
                if sha not in sha_list:
                    all_data.append(solo_data)
                    sha_list.append(sha)

        return all_data

    except Exception as e:
        doc_loc = None
        if err_dir is not None:
            sha = hashlib.sha3_256(str(html).encode()).hexdigest()
            doc_loc = '/'.join([err_dir, f'{sha}.html'])
            with open(doc_loc, 'w') as f:
                f.write(str(html))

        if doc_loc is not None:
            e = f"doc: {doc_loc} err: {str(e)}"
        else:
            e = str(e)

        exc_type, exc_obj, exc_tb = sys.exc_info()
        error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': e}
        # print("-- 8", error)
        return error


def detalji(html, rbr, detalj_url, rid):

    eoglasna_detalji = {'rbr': rbr, 'detalj_url': detalj_url, 'datum_objave': None, 'datum_brisanja': None, 'epredmet': None, 'nesolventnost': None,
                        'zip_link': None, 'javni_biljeznik_naziv': None, 'javni_biljeznik_adresa': None, 'jb_predmet_broj': None, 'rid': rid}

    zip_link = html.find('a', {'href': True, 'data-content': 'ZIP arhiva svih dokumenata'})
    if zip_link is not None:
        zip_link = ''.join([base_url, zip_link['href']])

    eoglasna_detalji.update({'zip_link': zip_link})

    # print(html)
    javni_bilj_data = html.find('dl', {'class':'invert-dl row'})

    javni_biljeznik = re.search('<dt class=".*?">Objavio javnobilježnički ured:</dt>\s+<dd class=".*?"><span>(.*?)</span><span>(.*?)</span></dd>', javni_bilj_data.decode(), re.MULTILINE | re.DOTALL | re.IGNORECASE)
    if javni_biljeznik is not None:
        javni_biljeznik_naziv, javni_biljeznik_adresa = javni_biljeznik.groups()
        javni_biljeznik_adresa = javni_biljeznik_adresa.strip().strip(';').strip()
        javni_biljeznik_naziv = javni_biljeznik_naziv.strip().strip(';').strip()
        eoglasna_detalji.update({'javni_biljeznik_naziv': javni_biljeznik_naziv, 'javni_biljeznik_adresa': javni_biljeznik_adresa})

    jb_predmet_broj = re.search('<dt class=".*?">Broj javnobilježničkog predmeta:</dt>\s+<dd class=".*?"><span>(.*?)</span></dd>', javni_bilj_data.decode(), re.MULTILINE | re.DOTALL | re.IGNORECASE)
    if jb_predmet_broj is not None:
        jb_predmet_broj = jb_predmet_broj.group(1)
        jb_predmet_broj = jb_predmet_broj.strip()
        eoglasna_detalji.update({'jb_predmet_broj': jb_predmet_broj})

    datum_detalji = html.find('section', {'id': 'detalji'})
    if datum_detalji is not None:
        txt = datum_detalji.text
        datum_objave = re.search('Datum i vrijeme objave:\s+(\d+.\d+.\d{4}.?\s+\d+:\d+)', txt, re.DOTALL | re.MULTILINE)
        if datum_objave is not None:
            datum_objave = datum_objave.group(1).strip()
            datum_objave = pd.to_datetime(datum_objave, dayfirst=True)
            eoglasna_detalji.update({'datum_objave': datum_objave})

        datum_skidanja = re.search('Posljednji datum objave:\s+(\d+.\d+.\d{4})', txt, re.DOTALL | re.MULTILINE)
        if datum_skidanja is not None:
            datum_skidanja = datum_skidanja.group(1).strip()
            datum_skidanja = pd.to_datetime(datum_skidanja, dayfirst=True)
            eoglasna_detalji.update({'datum_brisanja': datum_skidanja})

    epredmet = html.find('a', {'href': re.compile('e-predmet'), 'title': False})
    if epredmet is not None:
        epredmet = epredmet['href']
        eoglasna_detalji.update({'epredmet': epredmet})

    nesolventnost = html.find('a', {'href': re.compile('nesolventnost'), 'title': False})
    if nesolventnost is not None:
        nesolventnost = nesolventnost['href']
        eoglasna_detalji.update({'nesolventnost': nesolventnost})

    return eoglasna_detalji


def detalji_sudionik(html, rbr, detalj_url):
    section = html.find('section', {'id': 'subjekti'})
    uls = []
    if section is not None:
        uls = section.find_all('ul')

    sudionici = {}

    for ul in uls:
        lis = ul.find_all('li')
        for li in lis:
            sudionik = {'oib': None, 'adresa': None, 'sudionik': None, 'sudionik_uloga': None, 'rbr': rbr, 'detalj_url': detalj_url}
            divs = li.find_all('div', {'class': 'col-12'})
            for div in divs:
                # print(div)
                # print("=====*20")
                for k in sudionik.keys():
                    search_result = re.search(f'<small class="text-muted d-block">{k}</small>.*?<span class="d-block">(.*?)</span>.*?</div>', div.decode(), re.MULTILINE | re.IGNORECASE | re.DOTALL)
                    if search_result is not None:
                        sudionik.update({k: search_result.group(1)})

                # get sudionik name and roll in process
                dblock = div.find('div', {'class': 'd-block'})
                if dblock is not None:
                    sudionik_name = dblock.find('strong')
                    if sudionik_name is not None:
                        sudionik.update({'sudionik': sudionik_name.text.strip()})

                    roll_type = dblock.find('span', {'title': True})
                    if roll_type is not None:
                        sudionik_uloga = roll_type['title'].strip()
                        if sudionik_uloga != '':
                            sudionik.update({'sudionik_uloga': sudionik_uloga})

                    roll_type = dblock.find('span', {'title': False, 'class': 'badge-light'})
                    if roll_type is not None:
                        sudionik_uloga = roll_type.text.strip()
                        if sudionik_uloga != '':
                            sudionik.update({'sudionik_uloga': sudionik_uloga})

                    roll_type = dblock.find('small', {'title': True})
                    if roll_type is not None:
                        sudionik_uloga = roll_type['title'].strip()
                        if sudionik_uloga != '':
                            sudionik.update({'sudionik_uloga': sudionik_uloga})

            sha = hashlib.sha3_256(str(sudionik).encode()).hexdigest()
            sudionik.update({'sha': sha})
            sudionici.update({len(sudionici): sudionik})

    return sudionici


def doc_filenames(response_headers):
    filename, filename2 = None, None
    if response_headers is not None:
        content_disposition = response_headers.get('Content-Disposition')
        if content_disposition is not None:

            filename = re.search('name="(.*?)"', content_disposition)
            if filename is not None:
                filename = filename.group(1)

            content_disposition2 = content_disposition.replace("\\", '')
            filename2 = re.search("filename.?=.*?''(.*)", content_disposition2)
            if filename2 is not None:
                filename2 = filename2.group(1)
                filename2 = urllib.parse.unquote(filename2)
    return filename, filename2


def generate_sha(data):
    sha = hashlib.sha3_256(str(data).encode()).hexdigest()
    data.update({'sha': sha})
    return data


def parse2(html, courts, court_groups, rid=1, err_dir=None):
    result_selection = html.find('section', {'class': 'rezultati'})
    if result_selection is None:
        return {'err_desc': 'parse2 cant find section with class rezultati'}

    ul = result_selection.find('ul')
    if ul is None:
        return {'err_desc': 'parse2 cant find ul within class rezultati'}

    lis = ul.find_all('li')
    if not lis:
        return {'err_desc': 'parse2 cant find li within ul element'}

    result_data = dict()
    for li in lis:
        row_data = dict()
        divs = li.find_all('div')
        divs = [div for div in divs if div.parent.name == 'li']
        for div in divs:

            link_col = {'Sve objave vezane za ovaj premet': 'all_decisions', 'Detalji objave': 'details_link'}
            text_col = {'Sve objave vezane za ovaj premet': 'decision', 'Detalji objave': 'decision_title', 'Sve objave ovog suda': 'court_name',}

            # parsing row link data
            links = div.find_all('a', {'href': True, 'title': True})
            for link in links:
                link_text = link.text.strip()
                link_title = link['title']
                link_path = link['href']

                link_col_name = link_col.get(link_title)
                if link_col_name is not None:
                    row_data.update({link_col_name: link_path})
                text_col_name = text_col.get(link_title)
                if text_col_name is not None:
                    if text_col_name == 'decision':
                        decision_type = 'ostalo'
                        decision_types = {'^st': 'stečaj', '^sp': 'stečaj potrošača'}
                        for decision_type_rex, decision_type_group in decision_types.items():
                            if re.search(decision_type_rex, link_text, re.DOTALL |re.IGNORECASE) is not None:
                                decision_type = decision_type_group
                        row_data.update({'decision_type': decision_type})
                    if len(link_text) > 0:
                        row_data.update({text_col_name: link_text})

            # parsing badge description
            badge = div.find('span', {'class': 'badge'})
            if badge is not None:
                if badge['class'].__contains__('text-uppercase'):
                    badge = badge.text.strip()
                    if len(badge) > 0:
                        row_data.update({'badge': badge})

            # parsing small element data
            small_elements = {'Datum i vrijeme objave': 'announce_date', 'Posljednji datum objave': 'erase_date'}
            for small_element_title, small_element_colname in small_elements.items():
                small_element = div.find('small', text=re.compile(small_element_title))
                if small_element is not None:
                    span = div.find('span', {'class': 'm-date'})
                    if span is not None:
                        span_text = span.text.strip()
                        if len(span_text) > 0:
                            span_text = pd.to_datetime(span_text, dayfirst=True)
                            row_data.update({small_element_colname: span_text})

        court_name = row_data.get('court_name')
        court_id = courts.get(court_name)
        court_group = court_groups.get(court_name)
        if court_id is not None:
            row_data.update({'court_id': court_id, 'court_group': court_group, 'rid': rid})

        row_data_c = row_data.copy()
        link_details = row_data.get('details_link')
        link_data = result_data.get(link_details)
        if link_data is None:
            result_data.update({link_details: {}})

        sudionici_check = li.find('small', text=re.compile('Sudionici'))
        if sudionici_check is None:
            link_data = result_data.get(link_details)
            link_data.update({len(link_data): generate_sha(row_data_c)})

        for div in divs:
            link_data = result_data.get(link_details)
            # parsing participants
            check_small_element = div.find('small', text=re.compile('Sudionici'))
            if check_small_element is not None:
                participant_block_datas = div.find_all('div', {'class': 'd-block'})
                if not participant_block_datas:
                    link_data.update({len(link_data): generate_sha(row_data_c)})
                for participant_block_data in participant_block_datas:
                    row_data_c = row_data.copy()
                    if participant_block_data:
                        participant = {}
                        oib = participant_block_data.find('small', {'title': 'OIB'})
                        if not oib:
                            oib = participant_block_data.find('small', {'data-original-title': 'OIB'})

                        if oib is not None:
                            oib = re.search(r'(\d+)', oib.text.strip(), re.MULTILINE| re.DOTALL)
                            if oib is not None:
                                participant.update({'vat_number': oib.group(1).strip().zfill(11)})

                        participant_data = re.search('<span.*?>(?P<participant>.*?)</span>.*?<span.*?>(?P<participant_type>.*?)</span>', participant_block_data.decode(), re.MULTILINE| re.DOTALL)
                        if not participant_data:
                            participant_data = re.search('<span>(?P<participant>.*?)</span>', participant_block_data.decode(), re.MULTILINE | re.DOTALL)

                        if participant_data:
                            participant_data = {k: v.strip() for k, v in participant_data.groupdict().items() if len(v.strip()) > 0}
                            participant.update(participant_data)

                        if participant:
                            row_data_c.update(participant)
                            link_data.update({len(link_data): generate_sha(row_data_c)})

    return [result_data]


def details(html, detalj_url, rss_data, rbr, rid=1):
    court_name = rss_data.get('court_name')
    court_id = rss_data.get('court_id')
    court_group = rss_data.get('court_group')

    fina_court = html.find('span', {'class': 'court'}, text=re.compile('Financijska agencija'))
    if fina_court is not None:
        court_id = 99999
        court_name = 'Financijska agencija'
        court_group = 'Fina'

    eoglasna_detalji = {'details_link': detalj_url, 'announce_date': None, 'erase_date': None, 'ecase': None, 'insolvent': None,
                        'doc_link': None, 'public_notary': None, 'public_notary_address': None, 'public_notary_case': None, 'rid': rid,
                        'court_name': court_name, 'court_id': court_id, 'court_group': court_group, 'decision': None, 'rbr': rbr}

    doc_links = {'rbr': rbr, 'details_link': detalj_url, 'court_name': court_name, 'court_id': court_id,
                 'court_group': court_group, 'decision': None, 'doc_link': None}

    zip_link = html.find('a', {'href': True, 'data-content': 'ZIP arhiva svih dokumenata'})
    if zip_link is not None:
        zip_link = ''.join([base_url, zip_link['href']])

    eoglasna_detalji.update({'doc_link': zip_link})

    # broj rješenja
    broj_rjesenja = None
    search_link_broj = html.find('a', {'data-content': 'Sve objave vezane za ovaj predmet'})
    if search_link_broj is not None:
        broj_rjesenja = search_link_broj.text.strip()
        if broj_rjesenja == '':
            broj_rjesenja = None
        eoglasna_detalji.update({'decision': broj_rjesenja})

    search_link_broj = html.find('a', {'data-content': 'Sve objave vezane za ovaj predmet', 'href': True})
    if search_link_broj is not None:
        sve_objave_rjesenja = search_link_broj['href']
        eoglasna_detalji.update({'all_decisions': sve_objave_rjesenja})

    decision_title = html.find('strong', {'itemprop': 'name'})
    if decision_title is not None:
        decision_title = re.sub('\n', ' ', decision_title.text.strip())
        if len(decision_title) > 0:
            eoglasna_detalji.update({'decision_title': decision_title})

    # parsing badge description
    badge = html.find('span', {'class': 'badge'})
    if badge is not None:
        if badge['class'].__contains__('text-uppercase'):
            badge = badge.text.strip()
            if len(badge) > 0:
                eoglasna_detalji.update({'badge': badge})

    javni_bilj_data = html.find('dl', {'class':'invert-dl row'})

    javni_biljeznik = re.search('<dt class=".*?">Objavio javnobilježnički ured:</dt>\s+<dd class=".*?"><span>(.*?)</span><span>(.*?)</span></dd>', javni_bilj_data.decode(), re.MULTILINE | re.DOTALL | re.IGNORECASE)
    if javni_biljeznik is not None:
        javni_biljeznik_naziv, javni_biljeznik_adresa = javni_biljeznik.groups()
        javni_biljeznik_adresa = javni_biljeznik_adresa.strip().strip(';').strip()
        javni_biljeznik_naziv = javni_biljeznik_naziv.strip().strip(';').strip()
        eoglasna_detalji.update({'public_notary': javni_biljeznik_naziv, 'public_notary_address': javni_biljeznik_adresa, 'rid': 2})

    jb_predmet_broj = re.search('<dt class=".*?">Broj javnobilježničkog predmeta:</dt>\s+<dd class=".*?"><span>(.*?)</span></dd>', javni_bilj_data.decode(), re.MULTILINE | re.DOTALL | re.IGNORECASE)
    if jb_predmet_broj is not None:
        jb_predmet_broj = jb_predmet_broj.group(1)
        jb_predmet_broj = jb_predmet_broj.strip()
        eoglasna_detalji.update({'public_notary_case': jb_predmet_broj})
        if broj_rjesenja is None and len(jb_predmet_broj) >0:
            broj_rjesenja = jb_predmet_broj
            eoglasna_detalji.update({'decision': jb_predmet_broj})

    if broj_rjesenja is not None:
        decision_type = 'ostalo'
        decision_types = {'^st': 'stečaj', '^sp': 'stečaj potrošača'}
        for decision_type_rex, decision_type_group in decision_types.items():
            if re.search(decision_type_rex, broj_rjesenja, re.DOTALL | re.IGNORECASE) is not None:
                decision_type = decision_type_group
        eoglasna_detalji.update({'decision_type': decision_type})

    datum_detalji = html.find('section', {'id': 'detalji'})
    if datum_detalji is not None:
        txt = datum_detalji.text
        datum_objave = re.search('Datum i vrijeme objave:\s+(\d+.\d+.\d{4}.?\s+\d+:\d+)', txt, re.DOTALL | re.MULTILINE)
        if datum_objave is not None:
            datum_objave = datum_objave.group(1).strip()
            datum_objave = pd.to_datetime(datum_objave, dayfirst=True)
            eoglasna_detalji.update({'announce_date': datum_objave})

        datum_skidanja = re.search('Posljednji datum objave:\s+(\d+.\d+.\d{4})', txt, re.DOTALL | re.MULTILINE)
        if datum_skidanja is not None:
            datum_skidanja = datum_skidanja.group(1).strip()
            datum_skidanja = pd.to_datetime(datum_skidanja, dayfirst=True)
            eoglasna_detalji.update({'erase_date': datum_skidanja})

    epredmet = html.find('a', {'href': re.compile('e-predmet'), 'title': False})
    if epredmet is not None:
        epredmet = epredmet['href']
        eoglasna_detalji.update({'ecase': epredmet})

    nesolventnost = html.find('a', {'href': re.compile('nesolventnost'), 'title': False})
    if nesolventnost is not None:
        nesolventnost = nesolventnost['href']
        eoglasna_detalji.update({'insolvent': nesolventnost})

    for key in list(doc_links.keys()):
        key_data = eoglasna_detalji.get(key)
        doc_links.update({key: key_data})

    return eoglasna_detalji, doc_links


def participants_details(html, detalj_url, rbr):

    section = html.find('section', {'id': 'subjekti'})
    uls = []
    if section is not None:
        uls = section.find_all('ul')

    sudionici = {}

    for ul in uls:
        lis = ul.find_all('li')
        for li in lis:
            sudionik = {'rbr': rbr, 'vat_number': None, 'address': None, 'participant': None, 'participant_role': None, 'details_link': detalj_url}
            divs = li.find_all('div', {'class': 'col-12'})
            for div in divs:
                # print(div)
                # print("=====*20")

                search_result = re.search(f'<small class="text-muted d-block">OIB</small>.*?<span class="d-block">(\d+)</span>.*?</div>', div.decode(), re.MULTILINE | re.IGNORECASE | re.DOTALL)
                if search_result is not None:
                    sudionik.update({'vat_number': search_result.group(1).strip().zfill(11)})
                search_result = re.search(f'<small class="text-muted d-block">Adresa</small>.*?<span class="d-block">(.*?)</span>.*?</div>', div.decode(), re.MULTILINE | re.IGNORECASE | re.DOTALL)
                if search_result is not None:
                    sudionik.update({'address': search_result.group(1).strip()})

                # get sudionik name and roll in process
                dblock = div.find('div', {'class': 'd-block'})
                if dblock is not None:
                    sudionik_name = dblock.find('strong')
                    if sudionik_name is not None:
                        sudionik.update({'participant': sudionik_name.text.strip()})

                    roll_type = dblock.find('span', {'title': True})
                    if roll_type is not None:
                        sudionik_uloga = roll_type['title'].strip()
                        if sudionik_uloga != '':
                            sudionik.update({'participant_role': sudionik_uloga})

                    roll_type = dblock.find('span', {'title': False, 'class': 'badge-light'})
                    if roll_type is not None:
                        sudionik_uloga = roll_type.text.strip()
                        if sudionik_uloga != '':
                            sudionik.update({'participant_role': sudionik_uloga})

                    roll_type = dblock.find('small', {'title': True})
                    if roll_type is not None:
                        sudionik_uloga = roll_type['title'].strip()
                        if sudionik_uloga != '':
                            sudionik.update({'participant_role': sudionik_uloga})

                    roll_type = dblock.find('span', {'class': 'badge'})
                    if roll_type is not None:
                        sudionik_uloga = roll_type.text.strip().capitalize()
                        if sudionik_uloga != '':
                            sudionik.update({'participant_role': sudionik_uloga})

            sha = hashlib.sha3_256(str(sudionik).encode()).hexdigest()
            sudionik.update({'sha': sha})
            sudionici.update({len(sudionici): sudionik})

    return sudionici


def doc_filename(response_headers):
    filename = None
    extension = 'unknown'
    if response_headers is not None:
        content_disposition = response_headers.get('Content-Disposition')
        if content_disposition is not None:

            filename = re.search('name="(.*?)"', content_disposition)
            if filename is not None:
                filename = filename.group(1)
                ext_dot_position = filename.rfind(".")
                if ext_dot_position != -1:
                    extension = filename[ext_dot_position + 1:]
    return filename, extension

def check_login(soup):
    prijava_link = soup.find('a', {'href': True}, text=re.compile('Пријави се'))


# import os
# from bs4 import BeautifulSoup
#
# from core_framework.db_engine import DbEngine
#
# class Eoglasna(DbEngine):
#     def __init__(self, ):
#         DbEngine.__init__(self)
#         self.connect(1, connect_args={"application_name": "eoglasna/crawler.py"})
#
#
# if __name__ == '__main__':
#     api = Eoglasna()
#     courts = {row.get('court_name'): row.get('court_id') for row in api.select('court_list', filters={'court_id': True}, schema='eoglasna')}
#
#     # directories = os.listdir(r'E:\TuzbeDocs\eoglasna_htmls')
#     # for directory in directories:
#     with open(r'E:\Crawleri\crawlers_sm\crawlers_sm\croatia\eoglasna\example.html', 'rb') as f:
#         html = f.read()
#         html = BeautifulSoup(html, 'lxml')
#         parse2(html,courts)

# import re
# import hashlib
# from bs4 import BeautifulSoup
#
# with open(r'E:\Crawleri\crawlers_sm\crawlers_sm\croatia\eoglasna\htmls\detalji3.html', 'r', encoding='utf-8') as f:
#     html = f.read()
#     html = BeautifulSoup(html, 'lxml')
#     data = detalji_sudionik(html)
#     print(len(data))
#     for d in data:
#         print(d)

