import re
import hashlib
import pandas as pd
from bs4 import BeautifulSoup

repl_cols = {'OIB': 'vat_number', 'Klasa': 'doc_class',
             'Naziv/Ime i prezime': 'subject', 'Datum odluke': 'decision_date',
             'Naziv odluke': 'decision', 'Vijeće': 'council', '': 'details_url',
             'Postupak': 'procedure', 'Uvjeti stečeni': 'conditions', 'Iznos obveza': 'amount',
             'Vrijeme otvaranja': 'open_date', 'RC': 'rc', 'Naziv': 'cname', 'Ime i prezime': 'person_name',
             'Sjedište (ulica i mjesto)': 'hq', 'Poslovna adresa (ulica i mjesto)': 'business',
             'Prebivalište (ulica i mjesto)': 'residence', 'Djelatnost': 'activity',
             'Broj zaposlenih': 'workers', 'Nadležna PU': 'jurisdiction', 'Telefon': 'phone',
             'E-mail': 'email', 'Ur. broj': 'doc_id', 'Vrsta': 'doc_type', 'Datum objave': 'doc_date'}


class ParseTableBucket:
    def __init__(self):
        self.search_data = {}
        self.details_url = None


def row_sha(data, sha_cols):
    sha_data = [v for k,v in data.items() if k in sha_cols]
    return hashlib.sha3_256(str(sha_data).encode()).hexdigest()


def parse_table(html):
    table = html.find('table')
    if table is None:
        return {'err_desc': 'cannot find table in search results'}

    column_names = [repl_cols.get(th.text.strip()) for th in table.find_all('th')]
    data = []
    for row in table.find_all('tr'):
        row_data = []
        for td in row.find_all('td'):
            td_check = td.find('a')
            if td_check is not None:
                link = td.a['href']
                row_data.append(link)
            else:
                not_link = ''.join(td.stripped_strings)
                if not_link == '':
                    not_link = None
                row_data.append(not_link)
        data.append(row_data)
    df = pd.DataFrame(data[1:], columns=column_names)
    table_data = df.to_dict('records')
    docs_class = {}

    for record in table_data:
        sha_rows = ['doc_class', 'vat_number', 'debtor', 'decision_date', 'decision']
        decision_date = pd.to_datetime(record.get('decision_date'), errors='coerce', dayfirst=True)
        record.update({'sha': row_sha(record, sha_rows), 'decision_date': decision_date})
        doc_class = record.get('doc_class')
        if docs_class.get(doc_class) is None:
            data_bucket = ParseTableBucket()
            record_copy = record.copy()
            details_url = record_copy.get('details_url')
            if details_url is not None:
                record_copy.pop('details_url')
            data_bucket.search_data.update(record_copy)
            data_bucket.details_url = details_url
            docs_class.update({doc_class: data_bucket})
    return [docs_class]


def parse_details(html):
    rows = html.find_all('div', {'class': ['span5', 'span10']})
    details_data = {}
    for row in rows:
        label = row.find('label')
        span = row.find('span')
        if label is not None and span is not None:
            label_text = re.sub(':', '', label.text.strip())
            span_text = span.text.strip()
            if len(span_text) == 0:
                span_text = None
            label_text = repl_cols.get(label_text)
            if label_text == 'open_date':
                span_text = pd.to_datetime(span_text, errors='coerce', dayfirst=True)
            if label_text == 'amount':
                span_text = span_text.replace('.', '').replace(',', '.')
                span_text = pd.to_numeric(span_text)
            if label_text is not None:
                details_data.update({label_text: span_text})

    sha = row_sha(details_data, list(repl_cols.values()))
    details_data.update({'sha': sha})

    # documents parsing
    table = html.find('table')
    if table is None:
        return {'err_desc': 'cannot find table in details results'}

    column_names = [repl_cols.get(th.text.strip()) for th in table.find_all('th')]

    data = []
    for row in table.find_all('tr'):
        row_data = []
        for td in row.find_all('td'):
            td_check = td.find('input')
            if td_check is not None:
                link = td.input['onclick']
                link_data = re.search("(/pn-public-web.*?)'", link)
                if link_data is not None:
                    link = link_data.group(1)
                row_data.append(link)
            else:
                not_link = ''.join(td.stripped_strings)
                if not_link == '':
                    not_link = None
                row_data.append(not_link)
        data.append(row_data)
    df = pd.DataFrame(data[1:], columns=column_names)
    documents_data = df.to_dict('records')
    vat_number = details_data.get('vat_number')
    for document_data in documents_data:
        sha_cols = ['doc_class', 'doc_id', 'doc_type', 'doc_date']
        doc_date = document_data.get('doc_date')
        doc_date = pd.to_datetime(doc_date, errors='coerce', dayfirst=True)
        document_data.update({'doc_date': doc_date, 'vat_number': vat_number})
        sha = row_sha(document_data, sha_cols)
        document_data.update({'sha': sha})
    return details_data, documents_data
