"""UNOPTIMIZED VERSION OF CRAWLER"""
import sys
import asyncio
import logging
import hashlib
import cyrtranslit
from time import sleep
from bs4 import BeautifulSoup
from datetime import datetime, timedelta
from core_framework.crawlers import CrawlerBase
from crawlers_sm.serbia.alsu.parser import parse

_print = False
schema = 'alsu'
db_conn_id = 4
base_url = "http://alsu.gov.rs"
stec_url = "/".join([base_url, 'la', 'stecajevi'])
next_page = base_url +"/la/stecajevi/page/{}/"


def pprint(text):
    if _print is True:
        print(text)


class Alsu(CrawlerBase):
    def __init__(self, webpage, n_pages):
        CrawlerBase.__init__(self, webpage)
        self.pages = []
        self.all_pages = []
        self.n_pages = n_pages + 1

    def convert_none(self, text):
        if text is None:
            return ''
        return text

    def gather_links(self):
        step_name = 'gathering links'
        start_date = datetime.now()

        # load list of pages that need to be checked or rechecked
        last_checked = datetime.now() - timedelta(hours=24)
        self.connect(db_conn_id)

        pages = self.select('pages', filters={'or': [{'status_date': f"<={last_checked}"}, {'status': '1'}]}, schema=schema)
        self.pages = [row.get('page') for row in pages]

        all_pages = self.select('pages', columns=['page'], schema=schema)
        self.all_pages = [row.get('page') for row in all_pages]

        # gather links
        future = asyncio.Task(self.get_pages())
        loop = asyncio.get_event_loop()
        loop.run_until_complete(future)
        link_list = future.result()

        # create connection on database
        self.connect(db_conn_id)

        # iterate trough results and insert data or errors in tables
        all_errors = list()
        errors_history = dict()
        error_sha = hashlib.sha3_256(str(datetime.now()).encode()).hexdigest()
        if link_list:
            for links, errors in link_list:
                if errors:
                    pprint(f"errors: {errors}")
                    all_errors.append(errors)
                    for page, error in errors.items():
                        error_desc = error.get('RequestError')
                        if error_desc is not None:
                            desc_sha = hashlib.sha3_256(str(error_desc).encode()).hexdigest()

                            # this block of code controls that we insert errors that are unique
                            step_errors = errors_history.get(error_sha)
                            if step_errors is None:
                                errors_history.update({error_sha: []})

                            step_errors = errors_history.get(error_sha)
                            if desc_sha not in step_errors:
                                step_errors.append(desc_sha)
                                insert_data = {'sha': error_sha, 'step_name': step_name, 'error_desc': error_desc, 'desc_sha': desc_sha}
                                self.insert('errors', {0: insert_data}, schema=schema)

                        merge_data = {'page': page, 'status': 1, 'error_desc': error_desc}
                        self.merge('pages', {0: merge_data}, filters={'page': page}, on=['page'], update=False, schema=schema)

                        update_data = {'status': 1, 'status_date': datetime.now(), 'error_desc': error_desc}
                        self.update('pages', filters={'page': page}, values=update_data, schema=schema)
                if links:
                    pprint(f"link: {links}")
                    page = None
                    for rowid, data in links.items():
                        page = data.get('page')
                        break
                    self.merge('links', links, filters={}, update=False, on=['link_id'], schema=schema)

                    merge_data = {'page': page, 'status': 0}
                    self.merge('pages', {0: merge_data}, filters={'page': page}, on=['page'], update=False, schema=schema)

                    update_data = {'status': 0, 'status_date': datetime.now(), 'error_desc': None}
                    self.update('pages', filters={'page': page}, values=update_data, schema=schema)

                pprint("=====" * 20)

        error_count = len(all_errors)
        log_data = {'step_name': step_name, 'start_time': start_date, 'end_time': datetime.now()}
        if error_count > 0:
            log_data.update({'error_count': error_count, 'error_sha': error_sha})
        self.insert('log', {0: log_data}, schema=schema)

    async def get_pages(self):
        n_pages = range(1, self.n_pages)
        futures, results = list(), list()
        for page in n_pages:
            if page in self.pages or page not in self.all_pages:
                futures.append(self.get_page(page))

        for future in asyncio.as_completed(futures):
            try:
                results.append((await future))
            except Exception as e:
                logging.warning('Exception in Alsu.get_pages: {}'.format(e))
        return results

    async def get_page(self, page):
        # proc_id is important when we have multithread or multiprocessing in play
        proc_id = 1

        # prepare new request for this thread/process and return proxy sha
        sha = self.async_new_request(proc_id=proc_id, protocols=['http'], new=True)
        links, errors = {}, {}
        try:
            # tic_time is used for long loops where releasing same proxy will not happen longer than 30 min
            self.tic_time(sha, proc_id)

            # acquire request and set request type on 2 (number 2 is GET request with proxy)
            r = self.requests.get(proc_id)
            r.request_type = 2

            url = next_page.format(page)
            await r.go(url)
            html = r.response.content
            html_raw = r.response.content_raw

            if not type(html) is dict:
                html = BeautifulSoup(html_raw, 'lxml')
                articles = html.find_all('article')
                for article in articles:
                    link = article.find('a', {'href': True})
                    if link is not None:
                        link = link['href']
                        link_reverse = link[::-1]
                        link_id = link_reverse[0:link_reverse.find("/")][::-1].strip()
                        links.update({len(links): {"page_link": link, 'link_id': int(link_id), 'page': page}})
            else:
                err_desc = html.get('RequestError')
                if err_desc != '' and err_desc != 'Page server is down':
                    self.bad_proxy(sha, r.proxy_data)
                html.update({'error_date': datetime.now()})
                errors.update({page: html})
        finally:
            # release proxy back in use if it is not released it will be not be available for that web_base next 30min
            self.release_proxy(sha)
        return links, errors

    # ===================================================================================================
    # ===================================================================================================
    # ===================================================================================================

    def gather_data(self):
        step_name = 'gathering main data'
        start_date = datetime.now()

        # load list of pages that need to be checked or rechecked
        last_checked = datetime.now() - timedelta(hours=24)
        self.connect(db_conn_id)

        self.links = self.select('links', filters={'or': [{'processed': None}, {'status': '1'}, {'bankruptcy_status': 'U toku', 'status_date': f'<={last_checked}'}]}, schema=schema)

        # gather link data
        future = asyncio.Task(self.get_link_data())
        loop = asyncio.get_event_loop()
        loop.run_until_complete(future)
        gather_results = future.result()

        # create connection on database
        self.connect(db_conn_id)

        # iterate trough results and insert data or errors in tables
        all_errors = list()
        errors_history = dict()
        error_sha = hashlib.sha3_256(str(datetime.now()).encode()).hexdigest()

        if gather_results:
            for link_data, errors in gather_results:
                if errors:
                    all_errors.append(errors)
                    for link_id, error in errors.items():
                        pprint(f"link data error: {error}")
                        error_desc = error.get('RequestError')
                        if error_desc is not None:
                            desc_sha = hashlib.sha3_256(str(error_desc).encode()).hexdigest()
                            # this block of code controls that we insert errors that are unique
                            step_errors = errors_history.get(error_sha)
                            if step_errors is None:
                                errors_history.update({error_sha: []})
                            step_errors = errors_history.get(error_sha)
                            if desc_sha not in step_errors:
                                step_errors.append(desc_sha)
                                insert_data = {'sha': error_sha, 'step_name': step_name, 'error_desc': error_desc, 'desc_sha': desc_sha}
                                self.insert('errors', {0: insert_data}, schema=schema)

                        update_data = {'status': 1, 'status_date': datetime.now(), 'error_desc': error_desc}
                        self.update('links', filters={'link_id': link_id}, values=update_data, schema=schema)

                if link_data:
                    pprint(f"link data: {link_data}")
                    for rowid, data in link_data.items():
                        link_id = data.get('link_id')
                        bankruptcy_status = data.get('bankruptcy_status')

                        self.merge('data', {0: data}, filters={'link_id': link_id}, schema=schema)
                        values = {'processed': datetime.now(), 'bankruptcy_status': bankruptcy_status, 'status': 0, 'status_date': datetime.now(), 'error_desc': None}

                        self.update('links', {'link_id': link_id, 'processed': None}, values, schema=schema)
                        values.pop('processed')
                        self.update('links', {'link_id': link_id, 'processed': True}, values, schema=schema)
                pprint("=====" * 20)

        error_count = len(all_errors)
        log_data = {'step_name': step_name, 'start_time': start_date, 'end_time': datetime.now()}
        if error_count > 0:
            log_data.update({'error_count': error_count, 'error_sha': error_sha})
        self.insert('log', {0: log_data}, schema=schema)

    async def get_link_data(self):
        semaphore = asyncio.Semaphore(20)
        futures, results = list(), list()
        for link_data in self.links:
            link = link_data.get('page_link')
            link_id = link_data.get('link_id')
            futures.append(self.get_data(link, link_id, semaphore))

        for future in asyncio.as_completed(futures):
            try:
                results.append((await future))
            except Exception as e:
                logging.warning('Exception in Alsu.get_link_data: {}'.format(e))
        return results

    async def get_data(self, link, link_id, semaphore):
        # proc_id is important when we have multithread or multiprocessing in play
        proc_id = 1
        async with semaphore:
            # prepare new request for this thread/process and return proxy sha
            sha = self.async_new_request(proc_id=proc_id, protocols=['http'], new=True)
            parsed_data, errors = {}, {}
            try:
                # tic_time is used for long loops where releasing same proxy will not happen longer than 30 min
                self.tic_time(sha, proc_id)

                # acquire request and set request type on 2 (number 2 is GET request with proxy)
                r = self.requests.get(proc_id)
                r.request_type = 2

                await r.go(link)
                html = r.response.content
                html_raw = r.response.content_raw

                if not type(html) is dict:
                    parsed = parse(html_raw, link_id)
                    if parsed == 400:
                        error = {'error_date': datetime.now(), 'RequestError': 'No Vat number'}
                        errors.update({link_id: error})
                    else:
                        parsed_data.update({len(parsed_data): parsed})
                else:
                    err_desc = html.get('RequestError')
                    if err_desc != '' and err_desc != 'Page server is down':
                        self.bad_proxy(sha, r.proxy_data)
                    html.update({'error_date': datetime.now()})
                    errors.update({link_id: html})
            finally:
                # release proxy back in use if it is not released it will be not be available for that web_base next 30min
                self.release_proxy(sha)
            return parsed_data, errors

    def master(self):
        print("+ gathering links")
        self.gather_links()
        print("+ gathering data from links")
        self.gather_data()


if __name__ == '__main__':
    api = Alsu('alsu.gov.rs', 50)
    api.master()

