"""OPTIMIZED VERSION OF CRAWLER"""
import asyncio
import logging
import hashlib
from abc import ABC
from datetime import datetime, timedelta
from core_framework.crawlers import CrawlerBase
from crawlers_sm.serbia.alsu.parser import parse, extract_links

_print = False
schema = 'alsu'
db_conn_id = 1
base = 'alsu.gov.rs'
page_depth = 80
base_url = "http://alsu.gov.rs"
stec_url = "/".join([base_url, 'la', 'stecajevi'])
next_page = base_url + "/la/stecajevi/page/{}/"


def pprint(text):
    if _print is True:
        print(text)


class Alsu(CrawlerBase, ABC):
    def __init__(self, webpage, n_pages):
        CrawlerBase.__init__(self, webpage)
        self.n_pages = n_pages + 1
        self.pages, self.all_pages, self.links = [], [], []
        self.last_checked = datetime.now() - timedelta(hours=24)
        self.steps = {0: 'gathering links', 1: 'gathering main data'}

    @staticmethod
    def convert_none(text):
        if text is None:
            return ''
        return text

    def prepare_list(self, step_id):
        if step_id == 0:
            pages = self.select("pages", filters={'or': [{'status_date': f"<={self.last_checked}"}, {'status': '1'}]}, schema=schema)
            self.pages = [row.get('page') for row in pages]
            all_pages = self.select("pages", columns=['page'], schema=schema)
            self.all_pages = [row.get('page') for row in all_pages]

        if step_id == 1:
            or_ = [{'processed': None}, {'status': '1'}, {'bankruptcy_status': 'U toku', 'status_date': f'<={self.last_checked}'}]
            self.links = self.select('links', filters={'or': or_}, schema=schema)

    def gather(self):

        for step_id, step_data in self.steps.items():
            step_name = self.steps.get(step_id)
            print("+", step_name)

            # start time of step
            start_date = datetime.now()

            # create connection/session
            self.connect(db_conn_id, connect_args={"application_name": "serbia/alsu/crawler_opt/gather"})

            # load list of pages that need to be checked or rechecked
            self.prepare_list(step_id)

            # gather links/linkdata
            future = asyncio.Task(self.get_all_data(step_id))
            loop = asyncio.get_event_loop()
            loop.run_until_complete(future)
            gathered_data = future.result()

            # process gathered data
            # iterate trough results and insert data or errors in tables
            all_errors = list()
            errors_history = dict()
            error_sha = hashlib.sha3_256(str(datetime.now()).encode()).hexdigest()

            if gathered_data:
                for link_data, errors in gathered_data:
                    if errors:
                        pprint(f"errors: {errors}")
                        all_errors.append(errors)
                        for page, error in errors.items():
                            error_desc = error.get('RequestError')
                            if error_desc is not None:
                                desc_sha = hashlib.sha3_256(str(error_desc).encode()).hexdigest()

                                # this block of code controls that we insert errors that are unique
                                step_errors = errors_history.get(error_sha)
                                if step_errors is None:
                                    errors_history.update({error_sha: []})

                                step_errors = errors_history.get(error_sha)
                                if desc_sha not in step_errors:
                                    step_errors.append(desc_sha)
                                    insert_data = {'sha': error_sha, 'step_name': step_name, 'error_desc': error_desc, 'desc_sha': desc_sha}
                                    self.insert('errors', {0: insert_data}, schema=schema)

                            if step_id == 0:
                                merge_data = {'page': page, 'status': 1, 'error_desc': error_desc}
                                self.merge('pages', {0: merge_data}, filters={'page': page}, on=['page'], update=False, schema=schema)

                            update_data = {'status': 1, 'status_date': datetime.now(), 'error_desc': error_desc}

                            if step_id == 0:
                                self.update('pages', filters={'page': page}, values=update_data, schema=schema)
                            if step_id == 1:
                                self.update('links', filters={'link_id': page}, values=update_data, schema=schema)
                    if link_data:
                        pprint(f"link_data: {link_data}")
                        page = None

                        for rowid, data in link_data.items():
                            if step_id == 0:
                                page = data.get('page')
                                break
                            if step_id == 1:
                                link_id = data.get('link_id')
                                bankruptcy_status = data.get('bankruptcy_status')
                                self.merge('data', {0: data}, filters={'link_id': link_id}, schema=schema)
                                values = {'processed': datetime.now(), 'bankruptcy_status': bankruptcy_status, 'status': 0, 'status_date': datetime.now(), 'error_desc': None}
                                self.update('links', {'link_id': link_id, 'processed': None}, values, schema=schema)
                                values.pop('processed')
                                self.update('links', {'link_id': link_id, 'processed': True}, values, schema=schema)
                        if step_id == 0:
                            self.merge('links', link_data, filters={}, update=False, on=['link_id'], schema=schema)
                            merge_data = {'page': page, 'status': 0}
                            self.merge('pages', {0: merge_data}, filters={'page': page}, on=['page'], update=False, schema=schema)
                            update_data = {'status': 0, 'status_date': datetime.now(), 'error_desc': None}
                            self.update('pages', filters={'page': page}, values=update_data, schema=schema)
                    pprint("=====" * 20)

            error_count = len(all_errors)
            log_data = {'step_name': step_name, 'start_time': start_date, 'end_time': datetime.now()}
            if error_count > 0:
                log_data.update({'error_count': error_count, 'error_sha': error_sha})
            self.insert('log', {0: log_data}, schema=schema)

            # close connection/session
            self.disconnect()

    async def get_all_data(self, step_id):
        n_pages = range(1, self.n_pages)
        semaphore = asyncio.Semaphore(20)
        futures, results = list(), list()

        if step_id == 0:
            for page in n_pages:
                if page in self.pages or page not in self.all_pages:
                    futures.append(self.get_data(page, semaphore))

        if step_id == 1:
            for link_data in self.links:
                link = link_data.get('page_link')
                link_id = link_data.get('link_id')
                futures.append(self.get_data(link, semaphore, link_id, step_id=1))

        for future in asyncio.as_completed(futures):
            try:
                results.append((await future))
            except Exception as e:
                step_name = self.steps.get(step_id)
                logging.warning('Exception in step Alsu.{}: {}'.format(step_name, e))
        return results

    async def get_data(self, link, semaphore, link_id=1, step_id=0):
        # proc_id is important when we have multi thread or multiprocessing in play
        proc_id = 1

        async with semaphore:
            # prepare new request for this thread/process and return proxy sha
            sha = self.async_new_request(proc_id=proc_id, protocols=['http'], new=True)
            parsed_data, errors = {}, {}
            try:
                # acquire request and set request type on 2 (number 2 is GET request with proxy)
                r = self.requests.get(proc_id)
                r.request_type = 2

                url = None

                if step_id == 0:
                    url = next_page.format(link)
                if step_id == 1:
                    url = link
                if url is not None:
                    await r.go(url)
                    html = r.response.content
                    html_raw = r.response.content_raw
                    if not type(html) is dict:
                        if step_id == 0:
                            parsed_data = extract_links(html_raw, link)
                        if step_id == 1:
                            parsed = parse(html_raw, link_id)
                            if parsed == 400:
                                error = {'error_date': datetime.now(), 'RequestError': 'No Vat number'}
                                errors.update({link_id: error})
                            else:
                                parsed_data.update({len(parsed_data): parsed})
                    else:
                        err_desc = html.get('RequestError')
                        if err_desc != '' and err_desc != 'Page server is down':
                            self.bad_proxy(sha, r.proxy_data)
                        html.update({'error_date': datetime.now()})
                        if step_id == 0:
                            errors.update({link: html})
                        if step_id == 1:
                            errors.update({link_id: html})
            finally:
                await r.async_ses.close()
                r.clean()
                # release proxy back in use if it is not released it will be not be available for that web_base next 30min
                self.release_proxy(sha)
            return parsed_data, errors

    def master(self):
        print("+ gathering starts")
        self.gather()


if __name__ == '__main__':
    api = Alsu(base, page_depth)
    api.master()
