import os
import sys
import pickle
import getpass

from sqlalchemy import *
from subprocess import call
from sqlalchemy import event
from subprocess import Popen, PIPE
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base


if sys.version_info.major > 2:
    from core_framework.db_engine import DbEngine

schema = 'alsu'
framework_folder= r'C:\Users\{}\Documents\crawler_framework'.format(getpass.getuser())
python_path = r'{}\py_config.pkl'.format(framework_folder)

BasePstg = declarative_base()

# creating schema if it doesn't exist
event.listen(BasePstg.metadata, 'before_create', DDL("CREATE SCHEMA IF NOT EXISTS alsu"))


class AlsuLinks(BasePstg):
    """Contains all gathered links from alsu pages"""
    __tablename__ = 'links'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('links_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        Integer, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    page_link = Column(String(500))  # full link to the page we are parsing
    link_id = Column(Integer(), index=True)  # link id of the page we are parsing
    processed = Column(DateTime)  # datetime of first parsing. If it is a blank then it is not yet parsed
    bankruptcy_status = Column(String(3000))  # if bankruptcy_status is not yet finished ('U toku') we will recheck periodically this case
    status = Column(Integer(), index=True)  # last status of crawling that page, 0 = 0k, 1 = Done with errors
    status_date = Column(DateTime)  # datetime of status when it was updated
    error_desc = Column(String(3000))  # if there was an error for that page description will be here
    date_created = Column(DateTime, server_default=func.now())  # datetime when record has been added


class AlsuStatus(BasePstg):
    """Contains last status of crawled page. Servers to crawl again pages that resulted in error/exception"""
    __tablename__ = 'status'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('status_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        Integer, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    status_id = Column(Integer(), index=True)  # status code how did crawling finished
    status_desc = Column(String(3000))  # description of status code
    sha = Column(String(64), index=True)  # hash value of the row
    date_created = Column(DateTime, server_default=func.now())  # datetime when record has been added


class AlsuPages(BasePstg):
    """Contains last status of crawled page. Servers to crawl again pages that resulted in error/exception"""
    __tablename__ = 'pages'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('pages_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        Integer, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    page = Column(Integer(), index=True)  # number of page we are crawling
    status = Column(Integer(),  index=True)  # status of crawling that page
    status_date = Column(DateTime)  # datetime of status when it was updated
    error_desc = Column(String(3000))  # if there was an error for that page description will be here
    date_created = Column(DateTime, server_default=func.now())  # datetime when record has been added


class AlsuLog(BasePstg):
    """Contains log how did crawling job finished."""
    __tablename__ = 'log'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('log_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        Integer, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    step_name = Column(String(500))  # name of step like gathering links, or parsing details
    start_time = Column(DateTime)  # datetime when step started
    end_time = Column(DateTime)  # datetime when step ended
    error_count = Column(Integer())  # number of errors that occurred during step
    error_sha = Column(String(64), index=True)  # hash value of datetime when errors occurred and this is also connection on sha column in errors table


class AlsuErrors(BasePstg):
    """Contains errors in one or more steps that happened during crawling."""
    __tablename__ = 'errors'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('errors_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        Integer, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of datetime when errors occurred
    step_name = Column(String(500))  # name of step where errors occurred like gathering links, or parsing details
    error_desc = Column(String(4000))  # description of error that occurred
    error_date = Column(DateTime, server_default=func.now())  # date and time when error occurred
    desc_sha = Column(String(64), index=True)  # hash value of description for easier filtering and grouping
    date_created = Column(DateTime, server_default=func.now())  # date and time when this record is inserted in database


class AlsuData(BasePstg):
    """Contains parsed data from every individual link"""
    __tablename__ = 'data'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('data_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        Integer, id_seq,
        server_default=id_seq.next_value(), primary_key=True)
    sha = Column(String(64))  # hash value of row
    link_id = Column(Integer(), index=True)
    company_id = Column(Integer(), index=True)
    registry_number = Column(String(20), index=True)
    vat_number = Column(String(3000))
    company_name = Column(String(3000))
    court = Column(String(3000))
    decision = Column(String(3000))
    decision_date = Column(DateTime)
    activity = Column(String(3000))
    bankruptcy_status = Column(String(3000))
    municipality = Column(String(3000))
    address = Column(String(3000))
    city = Column(String(3000))
    legal_form = Column(String(3000))
    judge = Column(String(3000))
    manager = Column(String(3000))
    archive = Column(DateTime)
    date_created = Column(DateTime, server_default=func.now())


class AlsuTables(BasePstg):
    """Contains description of all tables in alsu schema"""
    __tablename__ = 'tables'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('tables_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        Integer, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    table_name = Column(String(3000))
    table_description = Column(String(3000))
    date_created = Column(DateTime, server_default=func.now())


def deploy(conn_id):
    db = DbEngine()
    engine = db.connect(conn_id, echo=False)
    BasePstg.metadata.drop_all(engine)
    BasePstg.metadata.create_all(engine)

    # add or remove data from tables
    session = sessionmaker(bind=engine)()

    table_desc = [{'table_name': AlsuData.__tablename__, 'table_description': AlsuData.__doc__},
                  {'table_name': AlsuErrors.__tablename__, 'table_description': AlsuErrors.__doc__},
                  {'table_name': AlsuLog.__tablename__, 'table_description': AlsuLog.__doc__},
                  {'table_name': AlsuPages.__tablename__, 'table_description': AlsuPages.__doc__},
                  {'table_name': AlsuStatus.__tablename__, 'table_description': AlsuStatus.__doc__},
                  {'table_name': AlsuLinks.__tablename__, 'table_description': AlsuLinks.__doc__},
                  {'table_name': AlsuTables.__tablename__, 'table_description': AlsuTables.__doc__},
                  ]

    status_desc = [{'status_id': 0, 'status_desc': 'OK'},
                   {'status_id': 1, 'status_desc': 'Done with errors'}]

    tables = {AlsuStatus.__tablename__: status_desc,AlsuTables.__tablename__: table_desc}
    for table, data in tables.items():
        rows = dict()
        [rows.update({len(rows): k}) for k in data]
        db.merge(table, rows, {}, delete=True, schema='alsu')
    session.commit()


def input_handler():
    while True:
        if sys.version_info.major < 3:
            option = raw_input()
        else:
            option = input()
        if option.isdigit() is False:
            print("\nOption must be numeric value. Please try again...\n")
            continue
        return int(option)


def version_call(argv):
    if not argv:
        sys.stdout.write("Provide connection id where alsu schema and tables will be deployed:")
        conn_id = input_handler()

        with open(python_path, 'rb') as fr:
            data = pickle.load(fr)
            version = data.get('version')
            script_path = os.path.realpath(__file__)
            if version != 'venv':
                cmd = 'py -{} {} {}'.format(version, script_path, conn_id)
            else:
                p = Popen('py -c "import sys; import os; sys.stdout.write(sys.executable)" ', stdout=PIPE)
                lines = p.stdout.readlines()
                real_executable = lines[0]
                if type(real_executable) is bytes:
                    real_executable = real_executable.decode()
                cmd = '{} {} {}'.format(real_executable, script_path, conn_id)
            call(cmd)
    else:
        deploy(int(argv[0]))
        os.system('pause')


if __name__ == '__main__':
    version_call(sys.argv[1:])



