import os
import sys
import pickle
import getpass
import hashlib
from sqlalchemy import *
from subprocess import call
from sqlalchemy import event
from subprocess import Popen, PIPE
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base

if sys.version_info.major > 2:
    from core_framework.db_engine import DbEngine

from crawlers_sm.serbia.apr.companies.codebook import *

schema = 'apr'

framework_folder = r'C:\Users\{}\Documents\crawler_framework'.format(getpass.getuser())
python_path = r'{}\py_config.pkl'.format(framework_folder)

BasePstg = declarative_base()

# creating schema if it doesn't exist
event.listen(BasePstg.metadata, 'before_create', DDL("CREATE SCHEMA IF NOT EXISTS {}".format(schema)))


# =====================================================================
#                               DATA TABLES
# =====================================================================
# class AprChanges(BasePstg):
#     """Contains all gathered unique company changes that came over APR webservice"""
#     __tablename__ = 'company_changes'
#     __table_args__ = {'schema': schema}
#
#     id_seq = Sequence('comp_change_id_seq', metadata=BasePstg.metadata, schema=schema)
#     id = Column(
#         BigInteger, id_seq,
#         server_default=id_seq.next_value(), primary_key=True)
#
#     sha = Column(String(64), index=True)  # hash value of the row
#     registry_number = Column(String(128), index=True)  # original mbr
#     company_type = Column(Integer())  # id for type of company. original= TipPrivrednogSubjekta
#     change_id = Column(Integer(), index=True)  # id of what kind of change occurred
#     change_date = Column(DateTime, index=True)  # datetime when change happened. original = DatumPromene
#     change_mode_id = Column(Integer())  # id on what way did change occurred. original = Identifikatornacinapromene
#     regproc_type = Column(Integer())  # id for type of registration procedure. original = Identifikatortiparegistracionogpostupka
#     processed = Column(Integer(), index=True)  # flag that tell us did crawler already processed this data if iz is 1 then it is
#     proc_date = Column(DateTime)  # date and time when did crawler mark this record as processed
#     transfer = Column(Integer(), index=True)
#     transfer_date = Column(DateTime)
#     date_created = Column(DateTime, server_default=func.now())  # date and time when record was created

# class HtmlDownload(BasePstg):
#     """Contains all companies that are not under jurisdiction of apr, so link """
#     __tablename__ = 'html_download'
#     __table_args__ = {'schema': schema}
#
#     id_seq = Sequence('html_id_seq', metadata=BasePstg.metadata, schema=schema)
#     id = Column(
#         BigInteger, id_seq,
#         server_default=id_seq.next_value(), primary_key=True)
#     sha = Column(String(64), index=True)  # hash value of the row
#
#     registry_number = Column(String(128), index=True)  # RegistryNumber
#     tab_name = Column(String(1280))  # name from tab link
#     registry = Column(String(1280))  # registry group
#     proc_date = Column(Date)  # date when it was processed
#
#
#     dl_stat = Column(Integer())  # html download status 1 it is done
#     dl_date = Column(DateTime)  # date and time when this html was last downloaded
#     dl_path = Column(String(1280))  # location where html is
#     html_size = Column(Integer()) # size of html
#     current_link = Column(String(1280))  # current dynamic link for reusage
#
#     parse_stat = Column(Integer())  # was downloaded html parsed without errors
#     parse_date = Column(DateTime)  # date and time when this html was parsed
#
#     date_created = Column(DateTime, server_default=func.now())  # DateCreated


class Decisions(BasePstg):
    """Contains all gathered info about company public decisions"""
    __tablename__ = 'decisions'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('decis_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row (case_number, act_type, publication_date)
    registry_number = Column(String(128), index=True)  # RegistryNumber

    case_number = Column(String(300))  # CaseNumber
    act_type = Column(String(4000))  # ActType
    publication_date = Column(DateTime())  # PublicationDate

    sha_doc = Column(String(64), index=True)  # sha_doc
    doc_dl = Column(Integer())  # DocDl  - download status
    doc_ext = Column(String(10))  # ext
    doc_loc = Column(String(4000))  # local path to doc

    doc_size = Column(BigInteger())
    doc_upl = Column(Integer())  # upload - upload status
    upl_loc = Column(String(4000))  # server path to doc
    doc_link = Column(String(4000))  # temp_link

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class BlackoutCompanies(BasePstg):
    """Contains all companies that are not under jurisdiction of apr, so link """
    __tablename__ = 'blackout'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('blackout_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    registry_number = Column(String(128), index=True)  # RegistryNumber

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated
    date_checked = Column(DateTime)  # DateChecked


class ParsingResults(BasePstg):
    """Contains result of parsing crawled data"""
    __tablename__ = 'parsing_results'
    __table_args__ = {'schema': schema}
    id_seq = Sequence('parsing_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber
    tab_name = Column(String(1024), index=True, nullable=False)
    parsing_stats = Column(ARRAY(Integer))

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class SearchInfo(BasePstg):
    """Contains all gathered company info from search results"""
    __tablename__ = 'search_info'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('search_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber
    ctype = Column(String(500), index=True)  # Type
    cname = Column(String(4000))  # Name
    cstatus = Column(String(500))  # Status
    status_group = Column(String(500))
    registry = Column(String(500))  # Registry
    web_link = Column(String(4000))  # WebLink
    link_date = Column(DateTime)  # time when link was parsed
    link_status = Column(Integer, index=True)  # status of link is it still usable or not

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated
    date_checked = Column(DateTime)  # DateChecked


class BasicInfo(BasePstg):
    """Contains all gathered company info from basic results in details"""
    __tablename__ = 'basic_info'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('basic_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber
    basic_name = Column(String(4000))  # BasicName
    cname = Column(String(4000))  # CompanyName
    cstatus = Column(String(500))  # Status
    status_group = Column(String(500))
    legal_form = Column(String(500))  # LegalForm
    foun_date = Column(Date)  # FounDate
    erase_date = Column(Date)  # EraseDate
    activity_start = Column(Date)   # ActivityStartDate - Datum početka obavljanja delatnosti
    activity_end = Column(Date)  # ActivityEndDate - Datum prestanka obavljanja delatnosti

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class ActivityPause(BasePstg):
    """Contains all gathered company info about economic activity pause periods in details"""
    __tablename__ = 'activity_pause'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('activityp_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber
    pause_start = Column(Date)  # PauseStartDate
    pause_end = Column(Date)  # PauseEndDate

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class AddressData(BasePstg):
    """Contains all gathered company info about company addresses in details"""
    __tablename__ = 'address_data'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('address_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber
    municipality = Column(String(4000))  # Municipality
    place = Column(String(4000))  # Place
    street = Column(String(4000))  # Street
    address_type = Column(String(4000))  # AddressType
    address_extra = Column(String(4000))  # AddressExtra
    zip_code = Column(String(150))  # ZipCode

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class Annotations(BasePstg):
    """Contains all gathered company info about annotation in details"""
    __tablename__ = 'annotations'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('anno_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber

    post_date = Column(DateTime)  # PostDate
    text_lines = Column(VARCHAR(1000000))  # TextLines
    submitter = Column(VARCHAR(1000000))  # Submitter

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class AssociationExtra(BasePstg):
    """Contains all gathered company info about association additional data in details"""
    __tablename__ = 'association_extra'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('assex_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber
    statute_date = Column(Date)  # StatuteDate
    statute_change = Column(Date)  # StatuteDateChange
    activity_code2 = Column(String(50))  # ActivityCode2
    activity_name2 = Column(VARCHAR(1000000))  # ActivityName2

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class BankAccounts(BasePstg):
    """Contains all gathered company info about company bank accounts in details"""
    __tablename__ = 'bank_accounts'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('bacc_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber
    bank_number = Column(String(500), index=True)  # BankNumber
    # active = Column(Integer(), index=True)  # Active

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class Bankruptcy(BasePstg):  # stecajne mase
    """Contains all gathered company info about company bankruptcy data in details"""
    __tablename__ = 'bankruptcy'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('bankrupcy_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber
    vat_number = Column(String(50))  # VatNumber
    bankruptcy_status = Column(String(50))  # BankruptcyStatus
    cregistry_number = Column(String(50))  # CompanyRegNum
    cname = Column(String(3000))  # CompanyName
    pname = Column(String(3000))  # Name - aka Person name
    psurname = Column(String(3000))  # Surname - aka person surname
    pin = Column(String(50))  # PIN - akak personal identification number
    debtor_reg_num = Column(String(50))  # DebtorRegNum
    debtor = Column(String(3000))  # Debtor
    creation_type = Column(String(3000))  # CreationType
    court_name = Column(String(3000))  # CourtName
    case_number = Column(String(500))  # CaseNumber
    case_date = Column(DateTime)  # CaseDate
    bankruptcy_value = Column(Numeric(18,2))  # BankruptcyValue
    # active = Column(Integer(), index=True)  # Active

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class BranchOffices(BasePstg):
    """Contains all gathered company info about branch offices in details"""
    __tablename__ = 'branch_offices'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('branch_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber

    branch = Column(String(500))  # Branch
    bname = Column(VARCHAR(1000000))  # Name - aka branch name
    activity_code = Column(String(20))  # ActivityCode
    activity_name = Column(VARCHAR(1000000))  # ActivityName
    municipality = Column(VARCHAR(1000000))  # Municipality
    place = Column(VARCHAR(1000000))  # Place
    street = Column(VARCHAR(1000000))  # Street
    address_extra = Column(VARCHAR(1000000))  # AddressExtra
    description = Column(VARCHAR(1000000))  # Description
    # active = Column(Integer(), index=True)  # Active

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class BusinessData(BasePstg):
    """Contains all gathered company info about additional business data ids in details"""
    __tablename__ = 'business_data'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('bussdata_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber
    vat_number = Column(String(128), index=True)  # VatNumber
    rzzo = Column(String(128), index=True)  # RZZO
    pio = Column(String(128), index=True)  # PIO
    activity_code = Column(String(20))  # ActivityCode
    activity_name = Column(VARCHAR(1000000))  # ActivityName

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class BusinessName(BasePstg):
    """Contains all gathered company info about company name variations"""
    __tablename__ = 'business_name'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('bussname_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber

    cname = Column(String(4000))  # CompanyName
    short_name = Column(String(4000))  # ShortName
    other_name = Column(String(4000))  # CompanyNameOther
    short_other_name = Column(String(4000))  # ShortNameOther
    language = Column(String(350))  # language

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class Capital(BasePstg):
    """Contains all gathered company info about company capital variations"""
    __tablename__ = 'capital'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('capital_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber

    share_group = Column(String(256))  # ShareGroup
    capital_type = Column(String(256))  # CapitalType
    amount = Column(Numeric(18, 2))  # Amount
    currency = Column(String(256))  # Currency
    event_date = Column(Date)  # EventDate
    share_in = Column(String(4000))  # ShareIn
    # active = Column(Integer(), index=True)  # Active

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class CapitalAmounts(BasePstg):
    """Contains all gathered info about company split of capital amounts"""
    __tablename__ = 'capital_amounts'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('capitalam_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber

    group_id = Column(Integer())  # GroupId
    person_group_id = Column(Integer())  # PersonGroupId
    owner_type = Column(String(256))  # OwnerType
    share_group = Column(String(256))  # ShareGroup
    capital_type = Column(String(256))  # CapitalType
    amount = Column(Numeric(18, 2))  # Amount
    currency = Column(String(256))  # Currency
    event_date = Column(Date())  # EventDate
    share_in = Column(String(4000))  # ShareIn
    # active = Column(Integer(), index=True)  # Active

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class CapitalGroups(BasePstg):
    """Contains all gathered info about company split of capital groups"""
    __tablename__ = 'capital_groups'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('capitalgr_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber

    pin = Column(String(50))  # PIN
    passport_number = Column(String(256))  # PassportNum
    creg_num = Column(String(50))  # CompanyNumber
    group_id = Column(Integer())  # GroupId
    person_group_id = Column(Integer())  # PersonGroupId
    owner_type = Column(String(256))  # OwnerType
    # active = Column(Integer(), index=True)  # Active

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class CapitalShares(BasePstg):
    """Contains all gathered info about company split of capital shares"""
    __tablename__ = 'capital_shares'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('capitalsh_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber

    group_id = Column(Integer())  # GroupId
    person_group_id = Column(Integer())  # PersonGroupId
    capital_share = Column(Numeric(18, 4))  # CapitalShare
    in_group_share = Column(Numeric(18, 4))  # InGroupShare
    owner_type = Column(String(256))  # OwnerType
    # active = Column(Integer(), index=True)  # Active

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class CompanyPeople(BasePstg):
    """Contains all gathered info about company peoples"""
    __tablename__ = 'company_people'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('people_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber

    pin = Column(String(50))  # PIN
    passport_number = Column(String(256))  # PassportNum
    passport_state = Column(String(4000))  # PassportState
    pname = Column(String(4000))  # Name -- aka person name
    pfunction = Column(String(4000))  # PersonFunction -- aka person function
    ptype = Column(String(256))  # PersonType
    pgroup = Column(String(500))  # PersonGroup
    represent = Column(String(256))  # Represent
    limitations = Column(String(4000))  # Limitations
    creg_num = Column(String(50))  # CompanyNumber -- aka company registry number
    cname = Column(String(4000))  # CompanyName
    branch = Column(String(4000))  # Branch
    # active = Column(Integer(), index=True)  # Active

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class Email(BasePstg):
    """Contains all gathered info about company public email"""
    __tablename__ = 'email'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('email_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber
    email = Column(String(4000))  # Email

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class EraseDetails(BasePstg):
    """Contains all gathered info about company details about why it is closed/erased"""
    __tablename__ = 'erase_details'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('eraser_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber

    erase_date = Column(Date())  # EraseDate
    activity_end_date = Column(Date())  # ActivityEndDate
    activity_end_reason = Column(String(4000))  # ActivityEndReason

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class Fax(BasePstg):
    """Contains all gathered info about company public fax numbers"""
    __tablename__ = 'fax'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('fax_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber
    fax_number = Column(String(4000))  # FaxNumber
    # active = Column(Integer(), index=True)  # Active

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class LimitedDuration(BasePstg):
    """Contains all gathered info about company limited duration of doing business"""
    __tablename__ = 'limited_duration'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('limited_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber
    end_date = Column(Date())  # EndDate

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class Phone(BasePstg):
    """Contains all gathered info about company webpage"""
    __tablename__ = 'phone'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('phone_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber
    phone_number = Column(String(4000))  # PhoneNumber
    # active = Column(Integer(), index=True)  # Active

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class WebPage(BasePstg):
    """Contains all gathered info about company webpage"""
    __tablename__ = 'web_page'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('web_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber
    web_url = Column(String(4000), index=True)  # Url

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class NoResult(BasePstg):
    """Contains all gathered info about companies that don't have any data on apr"""
    __tablename__ = 'no_result'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('nores_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # RegistryNumber
    date_checked = Column(DateTime)  # DateChecked

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


# =====================================================================
#                               CODEBOOKS
# =====================================================================
class AprAccounts(BasePstg):
    """Contains all gathered info about companies that don't have any data on apr"""
    __tablename__ = 'c_apr_accounts'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('apracc_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    aname = Column(String(64))  # person account name
    surname = Column(String(64))  # person account surname
    pin = Column(String(64))  # personal identification number
    email = Column(String(500))  # account email address used for login
    password = Column(String(500))  # account password used for login
    secret_q = Column(String(1024))  # secret question
    secret_a = Column(String(1024))  # answer to secret question
    pin_code = Column(String(1024))  # code send by apr server on email addres
    blocked = Column(Integer())  # flags with 1 all that are blocked
    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class TransferColumns(BasePstg):
    """Contains column name variations for transfer between mssql and postgre"""
    __tablename__ = 'c_transfer'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('trans_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    table_name_ms = Column(String(256))
    column_name_ms = Column(String(256))
    table_name_pst = Column(String(256))
    column_name_pst = Column(String(256))
    column_desc = Column(String(4000))

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # DateCreated


class CompanyChanges(BasePstg):
    """Contains all change codes for company"""
    __tablename__ = 'c_change_ids'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('change_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    change_id = Column(Integer())  # code
    description = Column(String(1024))
    register_priority = Column(Integer())
    change_priority = Column(Integer())
    register_name = Column(String(1024))
    reporting = Column(Integer())  # flag that tells us do we send this change in email as part of statistics
    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class CodeBooksBase1:
    sha = Column(String(64), index=True)  # hash value of the row
    description_c = Column(String(1024))
    description_l = Column(String(1024))
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class CompanyTypes(BasePstg, CodeBooksBase1):
    __tablename__ = 'c_company_types'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('comptype_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)
    company_type_id = Column(Integer(), index=True)


class CompanyChangeModes(BasePstg, CodeBooksBase1):
    __tablename__ = 'c_change_modes'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('compmode_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)
    company_mode_id = Column(Integer(), index=True)


class CompanyRegProcTypes(BasePstg, CodeBooksBase1):
    __tablename__ = 'c_regproc_types'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('regproctype_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)
    regproc_id = Column(Integer(), index=True)


class CrawlerInfoLog(BasePstg):
    __tablename__ = 'l_info'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('loginfo_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    runtime_id = Column(String(64), index=True)  # hash value of single run cycle of program
    sha = Column(String(64), index=True)  # hash value of the row
    start_time = Column(DateTime)
    end_time = Column(DateTime)
    step_name = Column(String(256))
    status = Column(Integer() ) # 1 = success  2=errors
    errors = Column(Integer())  # >0 there was and error
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class CrawlerErrorLog(BasePstg):
    __tablename__ = 'l_errors'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('logerr_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    runtime_id = Column(String(64), index=True)  # hash value of single run cycle of program
    sha = Column(String(64), index=True)  # hash value of the row
    start_time = Column(DateTime)
    end_time = Column(DateTime)
    step_name = Column(String(256))
    err_type = Column(String(3072))
    err_line = Column(Integer())
    err_desc = Column(String(3072))
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class DbConnection(DbEngine):
    def __init__(self):
        DbEngine.__init__(self)
        self.connect(1)


def populate_codebooks():
    api = DbConnection()
    populate_tables = {CompanyChanges.__tablename__ : change_ids,AprAccounts.__tablename__: apr_accounts,TransferColumns.__tablename__ : transfer_tables }
    for populate_table, data in populate_tables.items():
        data_copy = data.copy()
        for k, v in data_copy.items():
            sha = hashlib.sha3_256(str(v).encode()).hexdigest()
            v.update({'sha': sha})
        api.merge(populate_table, data_copy, filters={'archive': None}, schema=schema)


def deploy(conn_id):
    db = DbEngine()
    engine = db.connect(conn_id, echo=False)
    BasePstg.metadata.drop_all(engine)
    BasePstg.metadata.create_all(engine)
    populate_codebooks()


if __name__ == '__main__':
    deploy(1)

