import re
import hashlib
import pandas as pd
from datetime import datetime
from sys import stdout
import concurrent.futures
from core_framework.db_engine import DbEngine
from crawlers_sm.serbia.apr.companies import parser

schema = 'apr'
ms_tables = ['SearchInfo','BasicInfo','ActivityPause','AddressData','AssociationExtra', 'Annotations',
             'BankAccounts','Bankruptcy', 'BranchOffices','BusinessData','BusinessName','Capital',
             'CapitalAmounts','CapitalGroups','CapitalShares','CompanyPeople','Email','EraseDetails',
             'Fax','LimitedDuration','Phone','WebPage']

class DbConnection(DbEngine):
    def __init__(self):
        DbEngine.__init__(self, primary_key='id')
        self.connect(2)


class PostgresConnection(DbEngine):
    def __init__(self):
        DbEngine.__init__(self, primary_key='id')
        self.connect(1)


class DbAprConnection(DbEngine):
    def __init__(self):
        DbEngine.__init__(self, primary_key='id')
        self.conn = self.connect(3)


class Converters:
    def __init__(self):
        big_archivename = ['SearchInfo', 'Annotations', 'BankAccounts', 'Bankruptcy', 'BranchOffices', 'Capital',
         'CapitalAmounts', 'CapitalGroups', 'CapitalShares', 'CompanyPeople', 'Decisions',
         'Fax', 'Phone']

        self.common = {'registry_number': 'RegistryNumber', 'date_created': 'DateCreated', 'sha': 'sha'}

        self.search_info = {'ctype': 'Type', 'registry': 'Registry', 'cname': 'Name',
                            'cstatus': 'Status', 'status_group': 'StatusGroup', 'sha': 'sha', 'archive': 'Archive'}

        self.basic_info = {'basic_name': 'BasicName', 'cname': 'CompanyName',
                           'cstatus': 'Status', 'legal_form': 'LegalForm',
                           'registry_number': 'RegistryNumber', 'activity_start': 'ActivityStartDate',
                           'activity_end': 'ActivityEndDate', 'foun_date': 'FounDate',
                           'erase_date': 'EraseDate', 'archive': 'archive'}

        self.activity_pause = {'pause_start': 'PauseStartDate', 'pause_end': 'PauseEndDate', 'archive': 'archive'}
        self.address_data = {'municipality': 'Municipality', 'place': 'Place', 'street': 'Street', 'address_type': 'AddressType',
                             'address_extra': 'AddressExtra', 'zip_code': 'ZipCode', 'archive': 'archive'}
        self.association_extra = {'statute_date': 'StatuteDate', 'statute_change': 'StatuteDateChange',
                                  'activity_code2': 'ActivityCode2', 'activity_name2': 'ActivityName2', 'archive': 'archive'}
        self.annotations = {'post_date': 'PostDate', 'text_lines': 'TextLines', 'submitter': 'Submitter', 'archive': 'Archive'}
        self.bank_accounts = {'bank_number': 'BankNumber', 'archive': 'Archive'}
        self.bankruptcy = {'vat_number': 'VatNumber', 'bankruptcy_status': 'BankruptcyStatus', 'cregistry_number': 'CompanyRegNum',
                           'cname': 'CompanyName', 'pname': 'Name', 'psurname': 'Surname', 'pin': 'PIN', 'debtor_reg_num': 'DebtorRegNum',
                           'debtor': 'Debtor', 'creation_type': 'CreationType',
                           'court_name': 'CourtName', 'case_number': 'CaseNumber', 'case_date': 'CaseDate',
                           'bankruptcy_value': 'BankruptcyValue', 'archive': 'Archive'}

        self.blackout = {'archive': 'Archive'}

        self.branch_offices = {'branch': 'Branch', 'bname': 'Name', 'activity_code': 'ActivityCode', 'activity_name': 'ActivityName',
                               'municipality': 'Municipality', 'place': 'Place', 'street': 'Street', 'address_extra': 'AddressExtra',
                               'description': 'Description', 'archive': 'Archive'}
        self.business_data = {'vat_number': 'VatNumber', 'rzzo': 'RZZO', 'pio': 'PIO', 'activity_code': 'ActivityCode',
                              'activity_name': 'ActivityName', 'archive': 'archive'}

        self.business_name = {'cname': 'CompanyName', 'short_name': 'ShortName', 'other_name': 'CompanyNameOther',
                              'short_other_name': 'ShortNameOther', 'language': 'language'}

        self.capital = {'share_group': 'ShareGroup', 'capital_type': 'CapitalType', 'amount': 'Amount', 'currency': 'Currency',
                        'event_date': 'EventDate', 'share_in': 'ShareIn', 'archive': 'Archive'}

        self.tables = TransferPstgr().tablenames
        self.ms_tables = ['SearchInfo', 'BasicInfo', 'ActivityPause', 'AddressData', 'AssociationExtra', 'Annotations',
                          'BankAccounts', 'Bankruptcy', 'BranchOffices', 'BusinessData', 'BusinessName', 'Capital']

    def convert_to_mssql(self, data, tablename):
        conversion = getattr(self, tablename)
        if conversion:
            conversion.update(self.common)
        return {conversion.get(k): v for k,v in data.items() if k in list(conversion.keys())}

    def convert_to_postgre(self, data,tablename):
        tablename = self.tables.get(tablename)
        conversion = getattr(self, tablename)
        common = {v.lower(): k for k,v in self.common.items()}
        conversion = {v.lower():k for k,v in conversion.items()}
        if conversion:
            conversion.update(common)
        return {conversion.get(k): v for k, v in data.items() if k in list(conversion.keys())}


def unique_status(status_eval):
    if status_eval is None or status_eval == '':
        return 'Nepoznato'
    rex = {'aktiv': 'Aktivan', 'bris': 'Brisan', 'prin.*?likvi': 'Prinudna likvidacija',
           'proc.*?likvi': 'U procesu likvidacije', 'likvi': 'U likvidaciji',
           'ste.*?post': 'U stečajnom postupku', 'ste.a': 'U stečaju'}
    for k, v in rex.items():
        if re.search(k, status_eval, re.IGNORECASE) is not None:
            status_eval = v
            break
    return status_eval


def row_sha(data, cols=None, dates=None):
    """
    :type data: dict
    :type cols: list
    :type dates: list
    """
    copy_data = data.copy()
    if cols is not None:
        copy_data = {k: str(v) for k,v in copy_data.items() if k in cols}

    # replacing nan and Nat and None
    for k, v in copy_data.items():
        if str(v).lower() in ('nan', 'nat', 'none'):
            copy_data.update({k: str(None)})

    # order columns
    if cols is not None:
        copy_data = {k:copy_data.get(k) for k in cols}

    values = ''.join(list(copy_data.values()))
    # values = re.sub(r'[\s+]','',values, re.DOTALL|re.UNICODE)
    sha = hashlib.sha3_256(values.encode()).hexdigest()
    return sha


db = DbConnection()
db_apr = DbAprConnection()
pstgrs = PostgresConnection()


class HashCreator:
    # calculates sha for mssql tables
    def __init__(self):
        self.basic_info_cols = ['registry_number', 'basic_name', 'cname', 'cstatus', 'legal_form', 'foun_date', 'erase_date', 'activity_start', 'activity_end']
        self.search_info_cols = ['registry_number', 'ctype', 'cname', 'cstatus', 'registry']
        self.annotations_cols = ['registry_number', 'post_date', 'text_lines', 'submitter']

        # self.activity_pause_cols = ['registry_number', 'pause_start', 'pause_end']
        self.activity_pause_cols = ['registry_number', 'pause_start']
        self.address_data_cols = ['registry_number', 'municipality', 'place', 'street', 'address_type', 'address_extra']

        self.association_extra_cols = ['registry_number', 'statute_date', 'statute_change', 'activity_code2', 'activity_name2']
        self.bank_account_cols = ['registry_number', 'bank_number']
        self.bankruptcy_cols = ['registry_number', 'vat_number', 'bankruptcy_status', 'cregistry_number', 'cname', 'pname', 'psurname', 'pin',
                                'debtor_reg_num', 'debtor', 'creation_type', 'court_name', 'case_number', 'case_date', 'bankruptcy_value']
        self.branch_offices_cols = ['registry_number', 'branch', 'bname', 'activity_code', 'activity_name', 'municipality', 'place', 'street', 'address_extra', 'description']
        self.busniness_data_cols = ['registry_number', 'vat_number', 'rzzo', 'pio', 'activity_code', 'activity_name']
        self.business_name_cols = ['registry_number', 'cname', 'short_name', 'other_name', 'short_other_name']
        self.capital_cols = ['registry_number', 'share_group', 'capital_type', 'amount', 'currency', 'event_date', 'share_in']
        self.capital_amount_cols = ['registry_number', 'group_id', 'person_group_id', 'owner_type', 'share_group', 'capital_type', 'amount', 'currency', 'event_date', 'share_in']
        self.capital_groups_cols = ['registry_number', 'pin', 'passport_number', 'creg_num', 'group_id', 'person_group_id', 'owner_type']
        self.capital_shares_cols = ['registry_number', 'group_id', 'person_group_id', 'in_group_share', 'owner_type', 'capital_share']
        self.company_people_cols = ['registry_number', 'pin', 'passport_number', 'passport_state', 'pname', 'pfunction', 'ptype', 'pgroup', 'represent', 'limitations', 'creg_num', 'cname', 'branch']
        self.email_cols = ['registry_number', 'email']
        self.erase_details_cols = ['registry_number', 'erase_date', 'activity_end_date', 'activity_end_reason']
        self.fax_cols = ['registry_number', 'fax_number']
        self.limited_duration_cols = ['registry_number', 'end_date']
        self.phone_cols = ['registry_number', 'phone_number']
        self.web_page_cols = ['registry_number', 'web_url']

    def assign_row_sha(self, recalc=False, tablename=None):
        # in datetime format
        tables_with_dates = {}

        basic_info_sql = ''' select id, RegistryNumber registry_number, BasicName basic_name, CompanyName cname, Status cstatus, LegalForm legal_form,
        FounDate foun_date, EraseDate erase_date, ActivityStartDate activity_start, ActivityEndDate activity_end from BasicInfo where sha is null '''

        search_info_sql = '''select id, RegistryNumber registry_number, Type ctype, Name cname, Status cstatus, Registry registry from SearchInfo where sha is null '''

        activity_pause_sql = ''' select id, RegistryNumber registry_number, PauseStartDate pause_start,PauseEndDate pause_end from ActivityPause where sha is null '''

        address_data_sql = ''' select id, RegistryNumber registry_number, Municipality municipality, Place place, Street street, 
        AddressType address_type, AddressExtra address_extra from AddressData where sha is null'''

        association_extra_sql = ''' select id, RegistryNumber registry_number, StatuteDate statute_date, StatuteDateChange statute_change,
         ActivityCode2 activity_code2, ActivityName2 activity_name2  from AssociationExtra where sha is null'''

        bank_accounts_sql = ''' select  id,  RegistryNumber registry_number, BankNumber bank_number from BankAccounts where sha is null '''

        bankruptcy_sql = ''' select  id, RegistryNumber registry_number, VatNumber vat_number, BankruptcyStatus bankruptcy_status, CompanyRegNum cregistry_number, 
        CompanyName cname, Name pname, Surname psurname, PIN pin, DebtorRegNum debtor_reg_num, Debtor debtor, CreationType creation_type, CourtName court_name,
        CaseNumber case_number, cast(CaseDate as date) case_date, BankruptcyValue bankruptcy_value  from Bankruptcy where sha is null '''

        branch_offices_sql = ''' select  id, RegistryNumber registry_number, Branch branch, Name bname, ActivityCode activity_code, ActivityName activity_name, 
        Municipality municipality, Place place, 
        Street street, AddressExtra address_extra, Description description from BranchOffices where sha is null '''

        busniness_data_sql = ''' select  id, RegistryNumber registry_number, VatNumber vat_number, RZZO rzzo, PIO pio, ActivityCode activity_code, ActivityName activity_name  
        from BusinessData where sha is null '''

        business_name_sql = '''select  id, RegistryNumber registry_number, CompanyName cname, ShortName short_name, CompanyNameOther other_name, ShortNameOther short_other_name  
        from BusinessName where sha is null '''

        capital_sql = ''' select  id, RegistryNumber registry_number, ShareGroup share_group, CapitalType capital_type, Amount amount, Currency currency, EventDate event_date, 
        ShareIn share_in from Capital where sha is null'''

        capital_amount_sql = ''' select  id, RegistryNumber registry_number, cast(GroupId as varchar) group_id, cast(PersonGroupId as varchar) person_group_id,  OwnerType owner_type, ShareGroup share_group,
        CapitalType capital_type, Amount amount, Currency currency , EventDate event_date, ShareIn share_in  from CapitalAmounts where sha is null '''

        capital_groups_sql = ''' select   id, RegistryNumber registry_number, PIN pin, PassportNum passport_number, CompanyNumber creg_num, cast(GroupId as varchar) group_id,
        cast(PersonGroupId as varchar) person_group_id,  OwnerType owner_type from CapitalGroups where sha is null '''

        capital_shares_sql = ''' select id, RegistryNumber registry_number, cast(GroupId as varchar) group_id, cast(PersonGroupId as varchar) person_group_id, 
        InGroupShare in_group_share, OwnerType owner_type, CapitalShare capital_share  from CapitalShares where sha is null '''

        company_people_sql = ''' select   id, RegistryNumber registry_number, PIN pin, PassportNum passport_number, PassportState passport_state, 
        Name pname, PersonFunction pfunction, PersonType ptype, PersonGroup pgroup, Represent represent, Limitations limitations,
         CompanyNumber creg_num, CompanyName cname, Branch branch from CompanyPeople  where sha is null '''

        email_sql = ''' select  id, RegistryNumber registry_number, Email email  from Email where sha is null  '''

        erase_details_sql = '''  select  id, RegistryNumber registry_number, EraseDate erase_date, ActivityEndDate activity_end_date, ActivityEndReason activity_end_reason 
        from EraseDetails where sha is null  '''

        fax_sql = ''' select  id, RegistryNumber registry_number,  FaxNumber fax_number from Fax where sha is null  '''
        limited_duration_sql = "select  id, RegistryNumber registry_number,  EndDate end_date from LimitedDuration where sha is null "
        phone_sql = "select  id, RegistryNumber registry_number,  PhoneNumber phone_number from Phone where sha is null "
        web_page_sql = "select  id, RegistryNumber registry_number, Url web_url  from WebPage where sha is null "

        gen_order = {
                     'SearchInfo': {'sql': search_info_sql, 'cols': self.search_info_cols},
                     'BasicInfo': {'sql': basic_info_sql, 'cols': self.basic_info_cols},
                     'ActivityPause': {'sql': activity_pause_sql, 'cols': self.activity_pause_cols},
                     'AddressData': {'sql': address_data_sql, 'cols': self.address_data_cols},
                     'AssociationExtra': {'sql': association_extra_sql, 'cols': self.association_extra_cols},
                     'BankAccounts': {'sql': bank_accounts_sql, 'cols': self.bank_account_cols},
                     'Bankruptcy': {'sql': bankruptcy_sql, 'cols': self.bankruptcy_cols},
                     'BranchOffices': {'sql': branch_offices_sql, 'cols': self.branch_offices_cols},
                     'BusinessData': {'sql': busniness_data_sql, 'cols': self.busniness_data_cols},
                     'BusinessName': {'sql': business_name_sql, 'cols': self.business_name_cols},
                     'Capital': {'sql': capital_sql, 'cols': self.capital_cols},
                     'CapitalAmounts': {'sql': capital_amount_sql, 'cols': self.capital_amount_cols},
                     'CapitalGroups': {'sql': capital_groups_sql, 'cols': self.capital_groups_cols},
                     'CapitalShares': {'sql': capital_shares_sql, 'cols': self.capital_shares_cols},
                     'CompanyPeople': {'sql': company_people_sql, 'cols': self.company_people_cols},
                     'Email': {'sql': email_sql, 'cols': self.email_cols},
                     'EraseDetails': {'sql': erase_details_sql, 'cols': self.erase_details_cols},
                     'Fax': {'sql': fax_sql, 'cols': self.fax_cols},
                     'LimitedDuration': {'sql': limited_duration_sql, 'cols': self.limited_duration_cols},
                     'Phone': {'sql': phone_sql, 'cols': self.phone_cols},
                     'WebPage': {'sql': web_page_sql, 'cols': self.web_page_cols},
                     }
        if tablename is not None:
            gen_order = {tablename: gen_order.get(tablename)}

        # association extra sha row assing
        for tablename, gen_data in gen_order.items():
            print(f"\nCalcluating sha for {tablename}")
            if recalc is True:
                print('reseting sha column')
                update_status = db_apr.update(tablename, {'sha': True}, {'sha': None})
                print(f'reset status: {update_status}')
            sql = gen_data.get('sql')
            cols = gen_data.get('cols')
            data = db_apr.select(tablename, sql=sql)
            cnt = 0
            hashes = {}
            for row in data:
                cnt += 1
                stdout.write(f'\rprogress: {cnt}/{len(data)}')
                row_id = row.get('id')
                if tablename in list(tables_with_dates.keys()):
                    date_columns = tables_with_dates.get(tablename)
                    for date_column in date_columns:
                        date_data = row.get(date_column)
                        lower_str = str(date_data).lower()
                        if date_data is not None and lower_str not in ('nan', 'nat'):
                            date_data = date_data.strftime("%Y-%m-%d %H:%M:%S")
                            row.update({date_column: date_data})
                row = {x: row.get(x) for x in cols}

                row_sha_ready = row_sha(row, cols)
                # print(row_sha_ready, row, cols)
                # if tablename == 'SearchInfo':
                #     print(tablename, row_id, row_sha_ready, row)
                #     exit()
                hashes.update({len(hashes): {'id': row_id, 'sha': row_sha_ready}})

                if recalc is False:
                    update_status = db_apr.update(tablename, {'id': row_id}, {'sha': row_sha_ready})
                    if update_status != 200:
                        print(f"error occured: {update_status}")
                        exit()
            if cnt == 0:
                stdout.write(f'\rthere is not new data')
            if recalc is True:
                stdout.write(f'\rprogress: {cnt}/{len(data)}')
                insert_status = db_apr.insert('sha_table', hashes)
                if insert_status != 200:
                    print(f"error occured during insert: {insert_status}")
                    exit()
                manual_sql = f'''\nupdate a
                set a.sha  = b.sha
                from 
                {tablename} a
                inner join
                sha_table b
                on a.id = b.id
                where a.sha is null;
                truncate table sha_table'''
                print(manual_sql)


class ArchiveCalculator:
    # calculating arhive dates for those table who didnt have archive column

    # tables AssociationExtra, BusinessData, BusinessName, Email, EraseDetails, LimitedDuration, WebPage
    def archive_date(self, tablename, data):
        sorted_data = {}
        for row in data:
            registry_number = row.get('registry_number')
            date_created = row.get('date_created')

            get_reg_data = sorted_data.get(registry_number)
            if get_reg_data is None:
                sorted_data.update({registry_number: {0: row}})
            else:
                get_reg_data.update({len(get_reg_data): row})
        cnt = 1
        for k, v in sorted_data.items():
            total_rows = len(v)
            if total_rows > 1:

                for i in range(1, total_rows):
                    previous_row_data = v.get(i - 1)
                    row_data = v.get(i)
                    row_id = row_data.get('id')
                    archive_date = previous_row_data.get('date_created')
                    if tablename in ['SearchInfo', 'Bankruptcy', 'Fax']:
                        upd_stat = db_apr.update(tablename, {'id': row_id}, {'Archive': archive_date}, freeze=True)
                    else:
                        upd_stat = db_apr.update(tablename, {'id': row_id}, {'archive': archive_date}, freeze=True)
                    if upd_stat != 200:
                        print(tablename, cnt, upd_stat, row_data)
                        exit()
                    stdout.write(f'\rprogress: {cnt}/{len(sorted_data)}')
                    # print(cnt, upd_stat, row_data)
            cnt += 1
        if cnt == 1:
            stdout.write(f'\rarchive {tablename} - no new data to process')

    def calc_archive_date(self):
        tables = ['SearchInfo', 'BasicInfo', 'AddressData', 'AssociationExtra', 'Bankruptcy',
                  'BusinessData', 'BusinessName', 'Email', 'EraseDetails', 'Fax', 'LimitedDuration', 'WebPage']
        for tablename in tables:
            print(f"\nArchiving for {tablename}")
            # update rows that are older than last record for single registry number
            if tablename == 'AddressData':
                add_types = ['CompanyHq', 'CompanyMail']
                for add_type in add_types:
                    sql = f'''select id, RegistryNumber registry_number, DateCreated date_created  from {tablename}
                        where Archive is null and AddressType = '{add_type}' and
                        RegistryNumber in (
                        select RegistryNumber from(
                        SELECT RegistryNumber, COUNT(*) cnt FROM {tablename} where Archive is null and AddressType = '{add_type}' group by RegistryNumber having COUNT(*) > 1) as a)
                        order by RegistryNumber, DateCreated desc'''
                    data = db_apr.select(tablename, sql=sql)
                    self.archive_date(tablename, data)
            else:
                sql = f'''select id, RegistryNumber registry_number, DateCreated date_created  from {tablename}  where archive is null
                and RegistryNumber in (
                select RegistryNumber from(
                SELECT RegistryNumber, COUNT(*) cnt FROM {tablename} where Archive is null group by RegistryNumber having COUNT(*) > 1) as a
                ) order by RegistryNumber, DateCreated desc'''
                data = db_apr.select(tablename, sql=sql)
                self.archive_date(tablename, data)

        tables = ['ActivityPause']
        for tablename in tables:
            print(f"\nArchiving for {tablename}")
            # archive records for registry numbers of companies that are Erased from registry
            if tablename == 'ActivityPause':
                sql = '''select distinct RegistryNumber, 
                case
                when max(EraseDate) is not null then max(EraseDate) 
                when MAX(EraseDate) IS null and  MAX(ActivityEndDate) is not null then  MAX(ActivityEndDate)
                else MAX(DateCreated) end EraseDate  
                from BasicInfo where [status] like '%bris%' 
                and RegistryNumber in (select RegistryNumber from ActivityPause where Archive is null)
                and Archive is null group by RegistryNumber'''
                erased_companies = db_apr.select('SearchInfo', sql = sql)
                cnt = 0
                for row in erased_companies:
                    registry_number = row.get('registrynumber')
                    erasedate = row.get('erasedate')
                    update_status = db_apr.update(tablename, {'RegistryNumber': registry_number, 'archive': None}, {'archive': erasedate, 'reason': 'Brisan'}, freeze=True)

                    if update_status != 200:
                        print(f"Error on update of table {tablename}")
                        exit()
                    stdout.write(f'\rprogress: {cnt}/{len(erased_companies)}')
                    cnt += 1
                if cnt == 0:
                    stdout.write(f'\r no new data to process')


class TransferPstgr:
    # class is used to transfer new data from mssql to pstgr or update archive date of existing record
    def __init__(self):
        self.tablenames = {'SearchInfo': 'search_info', 'BasicInfo': 'basic_info', 'ActivityPause': 'activity_pause', 'AddressData': 'address_data',
                           'AssociationExtra': 'association_extra', 'Annotations': 'annotations', 'BankAccounts': 'bank_accounts', 'Bankruptcy': 'bankruptcy',
                           'BranchOffices': 'branch_offices', 'BusinessData': 'business_data', 'BusinessName': 'business_name', 'Capital': 'capital',
                           'CapitalAmounts': 'capital_amounts', 'CapitalGroups': 'capital_groups', 'CapitalShares': 'capital_shares', 'CompanyPeople': 'company_people',
                           'Decisions': 'decisions','Email': 'email', 'EraseDetails': 'erase_details', 'Fax': 'fax', 'LimitedDuration': 'limited_duration',
                           'Phone': 'phone', 'WebPage': 'web_page'}


    def remove_duplicates(self):
        """Removes duplicate values from mssql database APR_SRBIJA"""
        print('\nadd first_date column if does not exist')
        for ms_table in ms_tables:
            try:
                sql = f'ALTER TABLE {ms_table} ADD first_date datetime;'
                db_apr.conn.execute(sql)
            except:
                pass

        print('updating first date')
        for ms_table in ms_tables:
            stdout.write(f'\r{ms_table}')
            db_apr.conn.execute(f'''update a set a.first_date = b.first_date
            from {ms_table} a
            inner join (
            select RegistryNumber, sha, min(DateCreated) first_date  from {ms_table}
            where  RegistryNumber in (select RegistryNumber from  {ms_table} where first_date is null ) and sha is not null
            group by RegistryNumber, Sha) b
            on a.RegistryNumber = b.RegistryNumber
            and a.sha = b.sha where a.first_date is null''')

        print('\nduplicate remover')
        for ms_table in ms_tables:
            stdout.write(f'\r{ms_table}')
            sql = f'''WITH CTE AS(
            SELECT *,
            RN = ROW_NUMBER()OVER(PARTITION BY RegistryNumber, sha ORDER BY Archive, DateCreated)
            FROM {ms_table} where RegistryNumber is Not null and sha is Not null)
            DELETE FROM CTE WHERE RN > 1;
            update {ms_table} set DateCreated = first_date where DateCreated != first_date and first_date < isnull(Archive, '3092-01-01') and first_date is not null'''
            db_apr.conn.execute(sql)

        print("\ndelete records without registry number")
        for ms_table in ms_tables:
            stdout.write(f'\r{ms_table}')
            sql = f'''delete [APR_SRBIJA].[dbo].{ms_table} where RegistryNumber is null'''
            db_apr.conn.execute(sql)

        print('\ncheck for duplicates')
        duplicates = dict()
        for ms_table in ms_tables:
            sql = f'''select RegistryNumber, sha, count(*) from {ms_table} group by RegistryNumber, sha having  count(*) > 1 and sha is not null'''
            data = db_apr.select(ms_table, sql=sql)
            if data:
                duplicates.update({ms_table: len(data)})
            if duplicates:
                stdout.write(f'\r{duplicates}')
            else:
                stdout.write(f'\rno duplicates found in {ms_table}')

    def update_status_group(self):
        print("\nupdating status group in SearchInfo")
        sql = "select id id2, * from SearchInfo where StatusGroup is null"
        data = db_apr.select('SearchInfo', sql=sql)
        cnt = 0
        for row in data:
            cnt += 1
            stdout.write(f'\rprogress {cnt}/{len(data)}')
            row_id = row.get('id2')
            cstatus = row.get('status')
            status_group = unique_status(cstatus)
            update_status = db_apr.update('SearchInfo', {'id': row_id}, {'StatusGroup': status_group}, freeze=True)
            if update_status != 200:
                print(f"Error on update_status_group for table SearchInfo")
                exit()
        if cnt == 0:
            stdout.write(f'\rno rows for update in SearchInfo')

    def import_ms_sync_data(self, tablename_ms, regnum=None, archive=False):
        converters = Converters()
        filters = {'sha': True, 'sync': None}
        if regnum is not None:
            filters.update({'registrynumber': regnum})
        if archive is not False:
            filters.update({'archive': archive})

        data = db_apr.select(tablename_ms, filters=filters)

        # spliting data per registry number
        sorted_data = {}
        only_active = dict()
        for row in data:
            registry_number = row.get('registrynumber')
            archive = row.get('archive')
            if str(archive).lower() in ['none', 'nan', 'nat'] and 'archive' in row.keys():
                only_active.update({registry_number: row})

            get_reg_data = sorted_data.get(registry_number)
            row = converters.convert_to_postgre(row, tablename_ms)

            if get_reg_data is None:
                sorted_data.update({registry_number: {0: row}})

            else:
                get_reg_data.update({len(get_reg_data): row})
        return sorted_data, only_active

    def from_ms_to_pstgr(self, first_fill=False, regnum=None):
        converters = Converters()

        if first_fill is True:
            for tablename_ms, tablename_pst in self.tablenames.items():
                # insert new data
                print(f"\nPOSTGRES: syncing data for {tablename_ms}")
                update_status = db_apr.update(tablename_ms, {'sync': True}, {'sync': None}, freeze=True)
                if update_status != 200:
                    print(f"Error on update of table {tablename_ms}")
                    exit()

        sql = f'select count(*) from apr.search_info'
        check = pstgrs.select('search_info', sql=sql)
        if check:
            print("\nclosing existing records on postgres")
            for tablename_ms, tablename_pst in self.tablenames.items():
                extra = ''
                if regnum is not None:
                    extra = f" and RegistryNumber= '{regnum}' "
                # update archive dates if archive date is null in postgre and not in mssql + sync is null for that regnum
                sql = f'''select RegistryNumber, sha, archive from {tablename_ms} where RegistryNumber in (
                select RegistryNumber where sync is null
                ) and archive is not null {extra} '''
                data = db_apr.select(tablename_ms,  sql=sql)
                cnt = 0
                for row in data:
                    cnt += 1
                    sha = row.get('sha')
                    archive = row.get('archive')
                    registry_number = row.get('registrynumber')
                    if sha is not None and registry_number is not None:
                        stdout.write(f'\rprocess {cnt}/{len(data)}')
                        filters = {'sha': sha, 'registry_number': registry_number, 'archive': None}
                        update_status = pstgrs.update(tablename_pst, filters, {'archive': archive}, schema=schema)
                        if update_status != 200:
                            print(f"\nError on update of table {tablename_pst}")
                            exit()

        single_row_tables = ['SearchInfo', 'BasicInfo', 'AddressData', 'AssociationExtra', 'Bankruptcy',
                             'BusinessData', 'BusinessName', 'Email', 'EraseDetails', 'Fax', 'LimitedDuration', 'WebPage']

        for tablename_ms, tablename_pst in self.tablenames.items():
            print(tablename_ms)
            if tablename_ms not in converters.ms_tables:
                continue

            sorted_data, only_active = self.import_ms_sync_data(tablename_ms, regnum)

            # sync and calculate last record if active record from pstg doesnt match ms
            print("syncing last record in single row tables")
            archive_name = 'archive'
            if tablename_ms in ['SearchInfo', 'Annotations', 'BankAccounts', 'Bankruptcy', 'BranchOffices', 'Capital',
                                'CapitalAmounts', 'CapitalGroups', 'CapitalShares', 'CompanyPeople', 'Decisions',
                                'Fax', 'Phone']:
                archive_name = 'Archive'

            if tablename_ms in single_row_tables:
                for registry_number, ms_data in only_active.items():
                    # stdout.write(f'\rprocess {cnt}/len({data})')
                    if registry_number is not None:
                        filters = {'registry_number': registry_number, 'archive': None}
                        pstgr_data = pstgrs.select(tablename_pst, filters=filters, schema=schema)
                        if pstgr_data and ms_data:
                            pstgr_data = pstgr_data[0]
                            ms_sha = ms_data.get('sha')
                            pst_sha = pstgr_data.get('sha')
                            ms_created = ms_data.get('datecreated')
                            pst_created = pstgr_data.get('date_created')
                            if ms_sha != pst_sha:
                                if pst_created > ms_created:
                                    filters = {'RegistryNumber': registry_number, 'sha': ms_sha}
                                    update_status = db_apr.update(tablename_ms, filters, {archive_name: datetime.now()}, freeze=True)
                                    if update_status != 200:
                                        print(f"\nError on update of table {tablename_ms}")
                                        exit()
                                    new_ms_data = converters.convert_to_mssql(pstgr_data, tablename_pst)
                                    insert_status = db_apr.insert(tablename_ms, {0: new_ms_data})
                                    if insert_status != 200:
                                        print(f"\nError on insert to table {tablename_ms}")
                                        print(new_ms_data)
                                        exit()
                                if pst_created < ms_created:
                                    new_pst_data = converters.convert_to_postgre(ms_data, tablename_ms)
                                    merge_status = pstgrs.merge(tablename_pst, {0: new_pst_data}, filters={'registry_number': registry_number, archive_name: None}, on=['sha'], schema=schema)
                                    if merge_status != 200:
                                        print(f"Error on merge of table {tablename_ms}")

                                        exit()

            print("syncing archive row")
            sync_status = list()

            sorted_data, only_active = self.import_ms_sync_data(tablename_ms, regnum, archive=True)
            for registry_number, pst_data in sorted_data.items():
                merge_status = pstgrs.merge(tablename_pst, pst_data, filters={'registry_number': registry_number, archive_name: True}, update=False, on=['sha'], schema=schema)
                if merge_status != 200:
                    print(f"Error on merge of table {tablename_ms}")
                    exit()
                if registry_number not in sync_status:
                    sync_status.append(registry_number)

            print("syncing active rows")
            sorted_data, only_active = self.import_ms_sync_data(tablename_ms, regnum, archive=None)
            for registry_number, pst_data in sorted_data.items():
                merge_status = pstgrs.merge(tablename_pst, pst_data, filters={'registry_number': registry_number, archive_name: None}, on=['sha'], schema=schema)
                if merge_status != 200:
                    print(f"Error on merge of table {tablename_ms}")
                    exit()

            for registry_number in sync_status:
                update_status = db_apr.update(tablename_ms, {'RegistryNumber': registry_number, 'sync': None}, {'sync': 1}, freeze=True)
                if update_status != 200:
                    print(f"\nError on update of table {tablename_ms}")
                    exit()


if __name__ == '__main__':
    hc = HashCreator()
    # when there is only new rows of data to calculate
    hc.assign_row_sha()
    # when there is large amount of data to calculate
    # hc.assign_row_sha(recalc=True, tablename='WebPage')

    ac = ArchiveCalculator()
    ac.calc_archive_date()

    transfer = TransferPstgr()
    transfer.remove_duplicates()
    transfer.update_status_group()
    # when postgres tables are empty
    # transfer.from_ms_to_pstgr(first_fill=True)
    # when we want to sync specific registry number
    # transfer.from_ms_to_pstgr(regnum='06020089')
    # when there ms data is already transfered
    transfer.from_ms_to_pstgr()

