import os
import gc
import sys
import math
import hashlib
import asyncio
import pickle
import threading
import getpass
from abc import ABC
from time import sleep
import concurrent.futures
from random import randint
from bs4 import BeautifulSoup
from multiprocessing import Process
from sqlalchemy.pool import NullPool
from datetime import datetime, timedelta
from core_framework.crawlers import CrawlerBase
from core_framework.db_engine import DbEngine
from crawlers_sm.serbia.crhov.parser import *

from Scripts import filepath

storage_location = filepath.set_storage()
print("base storage_location",  storage_location)

optimization_status = 'not optimized'
schema = 'crhov'
base = 'http://crhov.rs'
ownership_url = '/'.join([base, '?Opcija=1'])

today = datetime(datetime.now().year, datetime.now().month, datetime.now().day)
# crhov_folder = r'C:\Users\{}\Documents\crhov'.format(getpass.getuser())
crhov_folder = os.path.join(storage_location, 'crhov')
today_folder_name = '/'.join([crhov_folder, datetime.now().strftime('%Y%m%d')])
results_folder = '/'.join([crhov_folder, 'results'])
today_results_folder = '/'.join([results_folder, datetime.now().strftime('%Y%m%d')])
short_date = datetime(datetime.now().year, datetime.now().month, datetime.now().day)

log = False
framework = 8
crawlers_sm = 9


class Framework(DbEngine):
    def __init__(self):
        DbEngine.__init__(self)
        self.connect(framework)

    def reset_proxy_dist(self):
        self.delete('proxy_dist', filters={'web_base': base})


class MsCrhov(DbEngine):
    def __init__(self):
        DbEngine.__init__(self)
        self.connect(6)


class Crhov(CrawlerBase, ABC):
    def __init__(self):
        CrawlerBase.__init__(self, base, proxy_type=2)
        self.framework = Framework()

        # runtime log initialization
        self.runtime_errors = 0
        self.runtime_status = 1
        self.runtime_start = datetime.now()
        self.runtime_id = hashlib.sha3_256(str(self.runtime_start).encode()).hexdigest()
        self.runtime_log = {'start_time': self.runtime_start, 'step_name': 'runtime'}

        # basic crawler settings
        self.workers = 7
        self.records_per_page = 30
        self.owners_data = []
        self.total_unique_owners = 0
        self.parsing_errors = 0
        self.crawler_errors = {}
        self.crawler_err_cnt = 0
        self.crawl_status = None
        self.blocked_radar = {}

        self.pages = []
        self.links = {}

        self.exit_flag = False
        self.error_sim = False
        self.shas_release = {}
        self.same_error = 0
        self.result_expo = False

    def crawler_error_combine(self, err_data, log_common, sha_data=None):
        self.crawler_err_cnt += 1
        log_copy = log_common.copy()
        log_copy.update(err_data)
        # print(err_data)
        if sha_data is None:
            sha_data = err_data
        err_sha = hashlib.sha3_256(str(sha_data).encode()).hexdigest()
        err_sha_data = self.crawler_errors.get(err_sha)
        if err_sha_data is None:
            err_data.update({'err_cnt': 1})
            self.crawler_errors.update({err_sha: log_copy})
        else:
            err_cnt = err_sha_data.get('err_cnt')
            if err_cnt is None:
                err_cnt = 0
            err_cnt += 1
            err_sha_data.update({'err_cnt': err_cnt})

    def crawl_log(self, data=None, start=False):
        if start is True:
            date_exists = self.select('l_crawl', filters={'crawl_date': today}, schema=schema)
            if not date_exists:
                self.insert('l_crawl',  {0: {'crawl_date': today}}, schema=schema)

        if data is not None:
            self.update('l_crawl', {'crawl_date': today}, data, schema=schema)
        self.crawl_status = self.select('l_crawl', filters={'crawl_date': today}, schema=schema)[0].get('crawl_status')

    def info_log(self, data):
        print("info_log", data)
        if log is True:
            sha = hashlib.sha3_256(str(data).encode()).hexdigest()
            data.update({'sha': sha, 'runtime_id': self.runtime_id, 'end_time': datetime.now()})
            self.insert('l_info', {0: data}, schema=schema)

    def error_log(self, data):
        print("error_log", data)
        self.runtime_status = 2
        self.runtime_errors += 1
        if log is True:
            sha = hashlib.sha3_256(str(data).encode()).hexdigest()
            data.update({'sha': sha, 'runtime_id': self.runtime_id, 'end_time': datetime.now()})
            self.insert('l_errors', {0: data}, schema=schema)

    def postgresql_conn(self):
        self.connect(crawlers_sm, connect_args={"application_name": "crhov/crawler.py postgresql_conn"})

    def calculate_sha(self, record):
        return hashlib.sha3_256(str(record).encode()).hexdigest()

    def dir_creation_check(self):
        start_time = datetime.now()
        step_name = 'dir_creation_check'
        print(f"> {step_name}")
        log_common = {'start_time': start_time, 'step_name': step_name}
        try:
            folders = [crhov_folder, today_folder_name]
            if self.result_expo is True:
                folders.extend([results_folder, today_results_folder])

            for folder in folders:
                if os.path.isdir(folder) is False:
                    os.mkdir(folder)
            log_common.update({'status': 1, 'errors': 0})
            self.info_log(log_common)
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            log_common.update({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)})
            self.error_log(log_common)
            self.exit_flag = True

    def handle_http_errors(self, html_bs, sha, r):
        err_desc = html_bs.get('RequestError')
        if err_desc == 'Blocked' or 'Page server is down':
            sleep(5)
            sha_count = self.blocked_radar.get(sha)
            if sha_count is None:
                self.blocked_radar.update({sha: 1})
            else:
                sha_count += 1
                self.blocked_radar.update({sha: sha_count})
                if sha_count >= 5:
                    self.bad_proxy(sha, r.proxy_data, web_page=base)
                    self.blocked_radar.pop(sha, None)
            if err_desc == 'Blocked':
                self.bad_proxy(sha, r.proxy_data, web_page=base)
            return 400
        else:
            # else we will hold proxy locked so we can grab another one
            sha_data = self.shas_release.get(sha)
            if sha_data is None:
                self.shas_release.update({sha: datetime.now()})
            return 300

    def crawler_page_links(self, page=1):
        proc_id = randint(1, 20000)
        start_time = datetime.now()
        step_name = 'crawler_page_links'
        main_url = ownership_url

        # prepare referer
        referer = ownership_url
        if page == 2:
            main_url = ''.join([ownership_url, f'&PageNum_RShartije={page}'])
        if page > 2:
            main_url = ''.join([ownership_url, f'&PageNum_RShartije={page}'])
            referer = ''.join([ownership_url, f'&PageNum_RShartije={page - 1}'])

        log_common = {'start_time': start_time, 'step_name': step_name}
        r, sha, html_bs = None, None, BeautifulSoup('', 'lxml')

        try:
            # print("url", main_url)
            for i in range(0, 5):
                sleep(randint(2, 5))
                sha = self.new_request(proc_id=proc_id, protocols=['http'], new=True)
                r = self.requests.get(proc_id)
                r.request_type = 2

                r.ses.headers.update({'Referer': referer})
                r.go(main_url)

                html_bs = r.response.content
                if type(html_bs) is dict and i == 4:
                    error = {'err_desc': 'failed to get main page with 4 different http proxies'}
                    self.crawler_error_combine(error, log_common)
                    return 400
                if type(html_bs) is dict:
                    self.handle_http_errors(html_bs, sha, r)
                    continue
                break

            if html_bs is None:
                error = {'err_desc': 'html_bs is None - could not reach data'}
                self.crawler_error_combine(error, log_common)
                return 400

            # get total number of records in table
            rec_number = parse_number_of_records(html_bs)
            if type(rec_number) is dict:
                self.crawler_error_combine(rec_number, log_common)
                return 400

            # calculate number of pages
            number_of_pages = math.ceil(rec_number/self.records_per_page)
            # make list of pages to check in mt
            pages = list(range(2, number_of_pages+1))
            if page == 1:
                self.pages = pages

            # parsing page links
            parse_results = parse_page_links(html_bs, page)
            if type(parse_results) is dict:
                self.crawler_error_combine(parse_results, log_common)
                return 400

            owner_cnt, links = parse_results

            page_link_data = self.links.get(page)
            if page_link_data is None:
                self.links.update({page: links})
            else:
                updates = dict()
                for k, v in page_link_data.items():
                    get_status = v.get('status')
                    if get_status in [None, False]:
                        row_data = links.get(k)
                        updates.update({k: row_data})

                page_link_data.update(updates)
            self.total_unique_owners += owner_cnt
            return 200

            # print(" + number of pages is", number_of_pages, len(pages))
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)}
            self.crawler_error_combine(error, log_common)
            return 400
        finally:
            # release proxy back in use if it is not released it will be not be available for that web_base next 30min
            self.release_proxy(sha)
            # release proxies with low errors
            for sha_key, date_locked in self.shas_release.items():
                self.release_proxy(sha_key)

    def filter_unsolved(self, page):
        page_data = self.links.get(page)
        return {x: y for x, y in page_data.items() if y.get('status') in [None, False]}

    def crawler_pages(self):
        start_time = datetime.now()
        step_name = 'crawler_pages'
        print(f"> {step_name}")
        log_common = {'start_time': start_time, 'step_name': step_name}

        try:
            page = 1
            issuers = {}
            sys.stdout.write(f"\r> fetch page {page}")
            while True:
                crawl_status = self.crawler_page_links(page)
                if crawl_status == 200:
                    self.crawler_details(page)
                    break

            for page in self.pages:
                while True:
                    sys.stdout.write(f"\r> fetch page {page}")
                    crawl_status = self.crawler_page_links(page=page)
                    if crawl_status == 200:
                        break
                self.crawler_details(page)

            pages_count = len(self.links)
            popkeys = ['details_url', 'status']
            for i in range(1, pages_count+1):
                links_data = self.links.get(i)
                for k, v in links_data.items():
                    row_copy = v.copy()
                    for popkey in popkeys:
                        row_copy.pop(popkey)
                    sha = self.calculate_sha(row_copy)
                    row_copy.update({'sha': sha})
                    issuers.update({len(issuers): row_copy})

            print(f"\n> add or close issuers: {len(issuers)}")
            merge_status = self.merge('issuers', issuers, filters={'archive': None},  schema=schema)
            if merge_status != 200:
                error = {'err_desc': 'failed to merge data in crhov.issuers table'}
                self.crawler_error_combine(error, log_common)
                return 400
            return 200
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'err_cnt': 1}
            self.crawler_error_combine(error, log_common)
            return 400
        finally:
            error_status = 1
            if self.crawler_err_cnt > 0:
                error_status = 2

            for err_sha, err_data in self.crawler_errors.items():
                self.error_log(err_data)

            log_common.update({'status': error_status, 'errors': self.crawler_err_cnt})
            self.info_log(log_common)

    def crawler_details(self, page):
        run_count = 0
        step_name = 'crawler_details'
        start_time = datetime.now()
        log_common = {'start_time': start_time, 'step_name': step_name}

        while True:
            # print("-----"*10)
            solved = 0
            self.same_error = 0
            links = self.filter_unsolved(page)
            if run_count > 0:
                sleep(10)
                if links:
                    self.crawler_page_links(page)
            links = self.filter_unsolved(page)

            if not links:
                break

            sys.stdout.write(f'\r> page:{page} tries: {run_count}, progress:{solved}/{len(links)}')
            run_count += 1
            with concurrent.futures.ThreadPoolExecutor(max_workers=self.workers) as executor:
                # Start the load operations and mark each future with its URL
                future_to_url = {executor.submit(self.crawler_detail_get, key, link, page): (key,link) for key,link in links.items()}
                for future in concurrent.futures.as_completed(future_to_url):
                    url = future_to_url[future]
                    try:
                        row_id, crawl_status = future.result()
                    except AttributeError as e:
                        pass
                    except Exception as e:
                        exc_type, exc_obj, exc_tb = sys.exc_info()
                        error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)}
                        self.crawler_error_combine(error, log_common)
                    else:
                        if type(row_id) is dict:
                            self.crawler_error_combine(row_id, log_common)
                        if crawl_status == 200:
                            solved += 1
                            page_data = self.links.get(page)
                            row_data = page_data.get(row_id)
                            if row_data is not None:
                                row_data.update({'status': True})
                    sys.stdout.write(f'\r> page:{page} tries: {run_count}, progress:{solved}/{len(links)}')

        # release proxies
        self.framework.reset_proxy_dist()

    def crawler_detail_get(self, key, link_data, page):
        sha = None
        http_error = 0
        proc_id = threading.current_thread()

        isin = link_data.get('isin')
        link = link_data.get('details_url')
        registry_number = link_data.get('registry_number')

        if self.same_error > 5:
            error = {'err_desc': "to many blank pages, triggering cooldown"}
            sleep(15)
            return error, -1

        main_url = ''.join([base, link])
        try:
            sleep(randint(2, 5))
            sha = self.new_request(proc_id=proc_id, protocols=['http'], new=True)
            r = self.requests.get(proc_id)
            r.request_type = 2
            r.ses.headers.update({'Referer': ownership_url})
            r.go(main_url)

            html = r.response.content_raw
            html_bs = r.response.content
            if type(html_bs) is dict:
                handle_status = self.handle_http_errors(html_bs, sha, r)
                # print (" ", handle_status, html_bs)
                if handle_status == 300:
                    http_error = 1
                return [], -1

            # check if website returned blank
            navigacija_element = html_bs.find('table', {'class': 'navigacija'})
            hartije_element = html_bs.find('table', {'class': 'hartije'})

            if navigacija_element is None and hartije_element is None:
                error = {'err_desc': 'webpage returned blank html page'}
                self.same_error += 1
                return error, -1

            if navigacija_element is not None:
                error = {'err_desc': 'webpage returned search results instead of details'}
                self.same_error += 1
                return error, -1

            with open(f'{today_folder_name}/{registry_number}_{isin}.html', 'wb') as f:
                f.write(html)

            # parsing data
            parsed_data = parse_details(html_bs, registry_number)
            if type(parsed_data) is dict:
                return parsed_data, -1

            final_datas = ['basic_details', 'legal_shares', 'percentage_shares', 'shares']
            basic_details = getattr(parsed_data, 'basic_details')
            isin = basic_details.get('isin')
            web_reg_num = basic_details.get('registry_number')

            if web_reg_num == registry_number:
                for_pickle = {}
                for final_data in final_datas:
                    data_ready = getattr(parsed_data, final_data)
                    if final_data == 'basic_details':
                        data_ready = {0: data_ready}
                    for_pickle.update({final_data: data_ready})
                    # if final_data == 'shares':
                    #     print('shares', registry_number, isin,  data_ready)
                    if data_ready is not None:
                        self.merge(final_data, data_ready, filters={'registry_number': registry_number, 'isin': isin, 'archive': None}, on=['sha'], schema=schema)

                if self.result_expo is True:
                    today_dt = datetime.now().strftime(f'{registry_number}_%H%M')
                    pickle_filename = f'{today_results_folder}/{today_dt}.pickle'
                    with open(pickle_filename, 'wb') as handle:
                        pickle.dump(for_pickle, handle, protocol=pickle.HIGHEST_PROTOCOL)
            else:
                error = {'err_desc': 'web_reg_num is not same as registry number'}
                return error, -1

            return key, 200

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)}
            return error, -1
        finally:
            if http_error == 0:
                # release proxy back in use if it is not released it will be not be available for that web_base next 30min
                self.release_proxy(sha)

    def transfer_data(self):
        start_time = datetime.now()
        step_name = 'transfer_data'
        print(f"> {step_name}")
        log_common = {'start_time': start_time, 'step_name': step_name}
        transfer_errors = 0

        try:
            mscrhov = MsCrhov()
            print("> transferring updates for issuers")
            issuers = self.select('issuers', columns=['sha', 'registry_number', 'issuer', 'isin'], filters={'archive': None}, schema=schema)
            if issuers:
                merge_data = dict(zip(list(range(len(issuers))), issuers))
                mscrhov.merge('issuers',merge_data, filters={'archive': None}, on=['sha'] )

            transfer_tables = ['shares', 'percentage_shares', 'legal_shares', 'basic_details']
            cols = {'shares': ['sha', 'registry_number', 'isin', 'cfi', 'ordinal', 'shareholder', 'shares', 'percentage', 'note', 'share_group'],
                    'percentage_shares': ['sha', 'registry_number', 'isin', 'cfi', 'range_id', 'range_name', 'sharholders_domestic', 'sharholders_foreign',
                                          'shares_domestic', 'shares_foreign', 'percentage_domestic', 'percentage_foreign'],
                    'legal_shares': ['sha', 'registry_number', 'isin', 'legal_type', 'shares', 'percentage'],
                    'basic_details': ['sha', 'registry_number', 'issuer', 'issuer_status', 'isin', 'cfi']}

            for transfer_table in transfer_tables:

                transfer_registry_numbers = self.select(transfer_table,columns=['registry_number'], filters={'archive': None, 'transfer': None}, schema=schema)
                transfer_registry_numbers = set([row.get('registry_number') for row in transfer_registry_numbers])
                print(f"> transferring updates for {transfer_table}: {len(transfer_registry_numbers)}")
                for transfer_registry_number in transfer_registry_numbers:
                    filters = {'archive': None, 'registry_number': transfer_registry_number}
                    col = cols.get(transfer_table)

                    reg_num_data = self.select(transfer_table, columns=col, filters=filters, schema=schema)

                    if reg_num_data:
                        # fix percentages
                        if transfer_tables in ['shares', 'legal_shares']:
                            [row.update({'percentage': format(row.get('percentage'), '.5f')}) for row in reg_num_data]
                        if transfer_table in ['percentage_shares']:
                            percentage_cols = ['percentage_domestic', 'percentage_foreign']
                            for percentage_col in percentage_cols:
                                [row.update({percentage_col: format(row.get(percentage_col), '.5f')}) for row in reg_num_data]

                        merge_data = dict(zip(list(range(len(reg_num_data))), reg_num_data))
                        merge_status = mscrhov.merge(transfer_table, merge_data, filters=filters, on=['sha'])
                        if merge_status == 200:
                            self.update(transfer_table, {'registry_number': transfer_registry_number, 'transfer': None}, {'transfer': 1, 'transfer_date': datetime.now()}, schema=schema)

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)}
            log_common_copy = log_common.copy()
            log_common_copy.update(error)
            self.error_log(log_common_copy)
        finally:
            error_status = 1
            if transfer_errors > 0:
                error_status = 2

            log_common.update({'status': error_status, 'errors': transfer_errors})
            self.info_log(log_common)


def start():
    api = Crhov()
    api.postgresql_conn()

    crawl_status = api.select('l_crawl', filters={'crawl_date': short_date, 'crawl_status': 1}, schema=schema)
    if not crawl_status:
        api.dir_creation_check()
        api.framework.reset_proxy_dist()
        if api.exit_flag is False:
            api.crawler_pages()
            api.transfer_data()
            api.insert('l_crawl', {0: {'crawl_date': short_date, 'crawl_status': 1}}, schema=schema)

        # final runtime log
        api.runtime_log.update({'status': api.runtime_status, 'errors': api.runtime_errors})
        api.info_log(api.runtime_log)
        api.disconnect()


if __name__ == '__main__':
    while True:
        p = Process(target=start)
        p.start()
        p.join()
        gc.collect()
        sleep(3600)




