import os
import sys
import pickle
import getpass

from sqlalchemy import *
from subprocess import call
from sqlalchemy import event
from subprocess import Popen, PIPE
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base

if sys.version_info.major > 2:
    from core_framework.db_engine import DbEngine

schema = 'crhov'

framework_folder = r'C:\Users\{}\Documents\crawler_framework'.format(getpass.getuser())
python_path = r'{}\py_config.pkl'.format(framework_folder)

BasePstg = declarative_base()

# creating schema if it doesn't exist
event.listen(BasePstg.metadata, 'before_create', DDL("CREATE SCHEMA IF NOT EXISTS {}".format(schema)))

# ======================================================================
#                               TABLES
# ======================================================================


class Issuers(BasePstg):
    """History of all gathered owners"""
    __tablename__ = 'issuers'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('issuers_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)
    issuer = Column(String(1024), index=True)
    isin = Column(String(128), index=True)

    transfer = Column(Integer())
    transfer_date =  Column(DateTime)
    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class LegalShares(BasePstg):
    """History of changes in legal shares"""
    __tablename__ = 'legal_shares'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('legal_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)
    isin = Column(String(128), index=True)
    legal_type = Column(String(1024))
    shares = Column(BigInteger())
    percentage = Column(Numeric(18, 5))

    transfer = Column(Integer())
    transfer_date = Column(DateTime)
    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class PercentageShares(BasePstg):
    """History of changes in legal shares"""
    __tablename__ = 'percentage_shares'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('prc_share_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)
    isin = Column(String(128), index=True)
    cfi = Column(String(128))
    range_id = Column(Integer())
    range_name = Column(String(128))
    sharholders_domestic = Column(BigInteger())
    sharholders_foreign = Column(BigInteger())
    shares_domestic = Column(BigInteger())
    shares_foreign = Column(BigInteger())
    percentage_domestic = Column(Numeric(18, 5))
    percentage_foreign = Column(Numeric(18, 5))

    transfer = Column(Integer())
    transfer_date = Column(DateTime)
    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class BasicDetails(BasePstg):
    """History of changes in legal shares"""
    __tablename__ = 'basic_details'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('basic_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)
    isin = Column(String(128), index=True)
    cfi = Column(String(128))
    issuer = Column(String(1024))
    issuer_status = Column(String(1024))

    transfer = Column(Integer())
    transfer_date = Column(DateTime)
    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class Shares(BasePstg):
    """History of changes in legal shares"""
    __tablename__ = 'shares'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('shares_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)
    isin = Column(String(128), index=True)
    cfi = Column(String(128))
    ordinal = Column(Integer())
    shareholder = Column(String(1024))
    shares = Column(BigInteger())
    percentage = Column(Numeric(18, 5))
    note = Column(String(3000))
    share_group = Column(String(512))

    transfer = Column(Integer())
    transfer_date = Column(DateTime)
    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


# ======================================================================
#                               LOGGERS
# ======================================================================
class CrawlerSemaphore(BasePstg):
    __tablename__ = 'l_crawl'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('crawl_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    crawl_date = Column(Date, index=True)
    crawl_status = Column(Integer())  # NULL = nothing passed, 1 = all passed
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class CrawlerInfoLog(BasePstg):
    __tablename__ = 'l_info'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('loginfo_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    runtime_id = Column(String(64), index=True)  # hash value of single run cycle of program
    sha = Column(String(64), index=True)  # hash value of the row
    start_time = Column(DateTime)
    end_time = Column(DateTime)
    step_name = Column(String(256))
    status = Column(Integer())  # 1 = success  2=errors
    errors = Column(Integer())  # >0 there was and error
    note = Column(String(1024))
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class CrawlerErrorLog(BasePstg):
    __tablename__ = 'l_errors'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('logerr_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    runtime_id = Column(String(64), index=True)  # hash value of single run cycle of program
    sha = Column(String(64), index=True)  # hash value of the row
    start_time = Column(DateTime)
    end_time = Column(DateTime)
    step_name = Column(String(256))
    err_type = Column(String(3072))
    err_line = Column(Integer())
    err_desc = Column(String(3072))
    err_cnt = Column(Integer())
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


def deploy(conn_id):
    db = DbEngine()
    engine = db.connect(conn_id, echo=False)
    BasePstg.metadata.drop_all(engine)
    BasePstg.metadata.create_all(engine)


if __name__ == '__main__':
    deploy(1)
    # deploy(6)
