import re
import hashlib
import pandas as pd


def parse_owner_list(html_bs, page):
    table = html_bs.find('table', {'class': 'hartije'})
    if table is None:
        error = {'err_desc': 'unable to parse. cant find table with class hartije'}
        return error

    trs1 = table.find_all('tr', {'class': 'svetlije'})
    trs2 = table.find_all('tr', {'class': 'tamnije'})
    trs = trs1 + trs2

    row_data = []
    total_companies = 0
    for tr in trs:
        td = tr.find('td', {'align': 'left'})
        if td:
            owner_data = re.search('(?P<issuer>.*?)\((?P<registry_number>\d+)\)', td.text.strip(), re.DOTALL | re.MULTILINE | re.IGNORECASE)
            if owner_data:
                total_companies += 1
                td_left_data = {k: v.strip() for k, v in owner_data.groupdict().items()}

                # registry_number check and fix
                registry_number = td_left_data.get('registry_number')
                registry_number = registry_number.zfill(8)
                td_left_data.update({'registry_number': registry_number, 'page': page})

                td = tr.find('td', {'align': 'center'})
                if td:
                    ais = td.find_all('a', {'href': True, 'class': 'kontaktlink'})
                    for a in ais:
                        url = a['href'].replace('&amp;', '&')
                        isin = a.text.strip()
                        td_center_data = {'details_url': url, 'isin':isin}
                        td_center_data_c = td_center_data.copy()
                        td_center_data_c.update(td_left_data)
                        row_data.append(td_center_data_c)

    return total_companies, row_data


def parse_page_links(html_bs, page):
    table = html_bs.find('table', {'class': 'hartije'})
    if table is None:
        error = {'err_desc': 'unable to parse. cant find table with class hartije'}
        return error

    trs1 = table.find_all('tr', {'class': 'svetlije'})
    trs2 = table.find_all('tr', {'class': 'tamnije'})
    trs = trs1 + trs2

    issuers_data = {}
    total_companies = 0
    for tr in trs:
        td = tr.find('td', {'align': 'left'})
        if td:
            owner_data = re.search('(?P<issuer>.*?)\((?P<registry_number>\d+)\)', td.text.strip(), re.DOTALL | re.MULTILINE | re.IGNORECASE)
            if owner_data:
                total_companies += 1
                td_left_data = {k: v.strip() for k, v in owner_data.groupdict().items()}

                # registry_number check and fix
                registry_number = td_left_data.get('registry_number')
                registry_number = registry_number.zfill(8)
                issuer = td_left_data.get('issuer')

                td = tr.find('td', {'align': 'center'})
                if td:
                    ais = td.find_all('a', {'href': True, 'class': 'kontaktlink'})
                    for a in ais:
                        url = a['href'].replace('&amp;', '&')
                        isin = a.text.strip()
                        row_id = f'{registry_number}{isin}'
                        if issuers_data.get(row_id) is None:
                            issuers_data.update({row_id: {}})
                        issuer_data = issuers_data.get(row_id)
                        issuer_data.update({'registry_number': registry_number, 'isin':isin, 'issuer': issuer, 'details_url': url, 'status': None})
    return total_companies, issuers_data


def parse_number_of_records(html_bs):
    table_nav = html_bs.find('table', {'class': 'navigacija'})
    if table_nav:
        n_companies = re.search('\d+\s+-\s+\d+\s+od\s+(\d+)', table_nav.text, re.DOTALL | re.MULTILINE | re.IGNORECASE)
        if n_companies is not None:
            return int(n_companies.group(1))
        else:
            error = "could not find element with number of total owners"
            return {'err_desc': error}
    else:
        error = "could not find html element navigacija"
        return {'err_desc': error}


def get_next_page(html_bs):
    table_nav = html_bs.find('table', {'class': 'navigacija'})
    if table_nav:
        next_page = table_nav.find('a', {'href': True}, text=re.compile('Napred >>'))
        if next_page:
            page_search = re.search('PageNum_RShartije=(\d+)', next_page['href'])
            if page_search:
                page_search = page_search.group(1)
                next_page = page_search
                return int(next_page)
        else:
            error = "could not find next page"
            return {'err_desc': error}

    else:
        error = "could not find html element navigacija"
        return {'err_desc': error}


class DetailsParsed:
    def __init__(self):
        self.active_isins = None
        self.selected_isin = None
        self.basic_details = None
        self.legal_shares = None
        self.percentage_shares = None
        self.shares = None



def parse_details(html_bs, registry_number):

    def headers_replace(col_name):
        repl = {'Tip lica': 'legal_type', 'Broj akcija': 'shares', 'ukupne emisije': 'percentage', 'Broj lica': 'sharholders',
                'domaća': 'domestic', 'strana': 'foreign', 'Rbr': 'ordinal', 'Akcionar': 'shareholder', 'Broj glasova': 'shares',
                'Ukupna vrednost': 'shares'}
        for k, v in repl.items():
            if re.search(k, col_name) is not None:
                return v
        return col_name

    unique_data = {'registry_number': registry_number}
    content_div = html_bs.find('div', {'class': 'story'})
    if content_div is None:
        error = f"cannot find main div with details data"
        return {'err_desc': error}

    main_label = content_div.find('label')
    if main_label is None:
        error = "cannot find main_label within details data"
        return {'err_desc': error}

    issuer_name = re.search('Izdavalac:\s+<.*?>(?P<issuer>.*?)</.*?>', main_label.decode(), re.DOTALL | re.MULTILINE | re.IGNORECASE)
    if issuer_name is None:
        error = "cannot find issuer_name within details data"
        return {'err_desc': error}
    issuer_name = {k: v.strip() for k, v in issuer_name.groupdict().items()}
    unique_data.update(issuer_name)

    issuer_status = re.search('Status izdavaoca:\s+<.*?>(?P<issuer_status>.*?)</.*?>', main_label.decode(), re.DOTALL | re.MULTILINE | re.IGNORECASE)
    if issuer_status is not None:
        issuer_status = {k: v.strip() for k, v in issuer_status.groupdict().items()}
        unique_data.update(issuer_status)

    active_isins = re.findall(r'<a.*?>(.*?)(\(.*?\))</a>', main_label.decode(), re.DOTALL | re.MULTILINE | re.IGNORECASE)
    if not active_isins:
        error = "cannot find active isins and cfi data within details data"
        return {'err_desc': error}
    active_isins = [{'isin': isin[0].strip(), 'cfi': isin[1].strip('(').strip(')')} for isin in active_isins]

    selected_isin = re.search('Izabran ISIN:\s*<.*?>(?P<isin>.*?)</.*?>', main_label.decode(), re.DOTALL | re.MULTILINE | re.IGNORECASE)
    if selected_isin is not None:
        selected_isin = {k: v.strip() for k, v in selected_isin.groupdict().items()}
        unique_data.update(selected_isin)

    property_type = content_div.find('table', {'class': 'hartije'})
    if not property_type:
        error = "cannot find active property_type within details data"
        return {'err_desc': error}

    trs = property_type.find_all('tr', {'class': ['svetlije', 'tamnije']})
    ths = property_type.find_all('th')
    ths = [headers_replace(th.decode()) for th in ths if re.search('Tip lica|Broj akcija|ukupne emisije', th.text, re.DOTALL | re.MULTILINE | re.IGNORECASE) is not None]
    total_headers = len(ths)
    if total_headers != 3:
        error = "number total_headers is different than it was before within details data"
        return {'err_desc': error}

    property_data = {}
    for tr in trs:
        tds = tr.find_all('td')
        tds = [td.text.strip() for td in tds]
        row = dict(zip(ths, tds))

        row.update({'registry_number': registry_number})
        numerics = ['shares', 'percentage']

        for numeric in numerics:

            check_key = row.get(numeric)
            if check_key is None:
                row.update({numeric:'0'})
            num_value = row.get(numeric).replace(',', '')
            new_num_value = pd.to_numeric(num_value, errors='coerce', downcast='float')
            if pd.isna(new_num_value) or new_num_value == 0.0:
                new_num_value = 0
            row.update({numeric: new_num_value})

        if selected_isin is not None:
            row.update(selected_isin)

        if selected_isin is None and active_isins is not None:
            isin_copy = active_isins[0].copy()
            if isin_copy.get('cfi') is not None:
                isin_copy.pop('cfi')
            row.update(isin_copy)

        property_data.update({len(property_data): row})
        # print(row)
        # print("====="*10)

    # get capital percentage per shares
    content_tables = content_div.find_all('table', {'class': 'hartije'})
    percentage_shares = {}
    # shareholders, votes_share, capital_share = {}, {}, {}
    shareholders, votes_share, capital_share = [], [], []

    table_groups = {1: shareholders, 2: votes_share, 3: capital_share}
    table_group_names = {1: 'shareholders', 2: 'votes share', 3: 'capital share'}
    shares = {}
    for content_table in content_tables:

        # data table Broj akcionara sa ucescem u kapitalu
        percent_share = re.search('Broj\s+akcionara', content_table.text, re.DOTALL | re.MULTILINE | re.IGNORECASE)
        if percent_share is not None:
            # print(content_table)
            df = pd.read_html(content_table.decode())
            data = df[0].to_dict('records')
            ranges = {'od 0% do 5%': 1, 'od 5% do 10%': 2, 'od 10% do 25%': 3, 'od 25% do 33%': 4, 'od 33% do 50%': 5,
                      'od 50% do 66%': 6, 'od 66% do 75%': 7, 'od 75% do 100%': 8, 'Ukupno:': 9}
            for row in data:
                percentage_share = {'registry_number': registry_number}
                if selected_isin is not None:
                    isin_d = selected_isin.get('isin')
                    selected_cfi = [active_isin.get('cfi') for active_isin in active_isins if active_isin.get('isin') == isin_d][0]
                    selected_isin.update({'cfi': selected_cfi})
                    percentage_share.update(selected_isin)
                if selected_isin is None and active_isins is not None:
                    isin_copy = active_isins[0].copy()
                    percentage_share.update(isin_copy)

                for k, v in row.items():
                    main_group = k[0]
                    sub_group = k[1]
                    if main_group == 'Broj akcionara sa učešćem u kapitalu':
                        group_id = ranges.get(v)
                        percentage_share.update({'range_id': group_id, 'range_name': v})
                    else:
                        col_name = '_'.join([headers_replace(main_group), headers_replace(sub_group)])
                        new_num_value = pd.to_numeric(v, errors='coerce', downcast='float')
                        if pd.isna(new_num_value) or new_num_value == 0.0:
                            new_num_value = 0
                        percentage_share.update({col_name : new_num_value})

                row_sha = hashlib.sha3_256(str(percentage_share).encode()).hexdigest()
                percentage_share.update({'sha': row_sha})
                percentage_shares.update({len(percentage_shares): percentage_share})

        # 1 data table Redosled prvih 10 akcionara po broju akcija
        # 2 data table Redosled prvih 10 akcionara po broju glasova
        table_regex = {1:'Rbr.*?Akcionar.*?Broj\s+Akcija', 2:'Rbr.*?Akcionar.*?Broj\s+Glasova', 3: 'Rbr.*?Akcionar.*?Ukupna\s+Vrednost'}
        for table_id, regex in table_regex.items():
            shareholders_share = re.search(regex, content_table.text, re.DOTALL | re.MULTILINE | re.IGNORECASE)
            if shareholders_share is not None:
                table_group = table_groups.get(table_id)
                # print(content_table)
                # filter shareholders with note
                tds = content_table.find_all('td')

                td_notes = [td for td in tds if td.find('font') is not None]
                shareholders_notes = {}
                for td_note in td_notes:
                    note = td_note.font.text
                    td_note.font.decompose()
                    shareholder_name = td_note.text.strip()
                    note = note.strip()
                    shareholders_notes.update({shareholder_name: note})

                note_elements = ['a','div']
                td_notes = [td for td in tds if td.find(note_elements) is not None]
                for td_note in td_notes:
                    note_all = ''

                    note = td_note.a.text
                    td_note.a.decompose()
                    note = re.sub(r'\s{2,}',' ', note.strip(), re.DOTALL | re.MULTILINE | re.IGNORECASE)
                    note_all += note

                    note = td_note.div.text
                    td_note.div.decompose()
                    note = re.sub(r'\s{2,}',' ', note.strip(), re.DOTALL | re.MULTILINE | re.IGNORECASE)
                    note_all += f"##{note}"
                    note_all = note_all[:3000]

                    shareholder_name = td_note.text.strip()
                    shareholders_notes.update({shareholder_name: note_all})

                df = pd.read_html(content_table.decode())[0]
                columns = [headers_replace(col) for col in df.columns]
                df.columns = columns
                data = df.to_dict('records')
                for d in data:
                    numerics = ['percentage']
                    for numeric in numerics:
                        num_value = str(d.get(numeric))
                        new_num_value = pd.to_numeric(num_value, errors='coerce', downcast='float')
                        if pd.isna(new_num_value) or new_num_value == 0.0:
                            new_num_value = 0
                        d.update({numeric: new_num_value})

                    d.update({'registry_number': registry_number})
                    if selected_isin is not None:
                        isin_d = selected_isin.get('isin')
                        selected_cfi = [active_isin.get('cfi') for active_isin in active_isins if active_isin.get('isin') == isin_d][0]
                        selected_isin.update({'cfi': selected_cfi})
                        d.update(selected_isin)
                    if selected_isin is None and active_isins is not None:
                        isin_copy = active_isins[0].copy()
                        d.update(isin_copy)

                    # shareholder
                    if shareholders_notes:
                        shareholder = d.get('shareholder')
                        for k, v in shareholders_notes.items():
                            v = v.strip()
                            match_shareholder = re.search(k, shareholder)
                            if match_shareholder is not None:
                                d.update({'shareholder': k, 'note': v})
                                break
                    row_sha = hashlib.sha3_256(str(d).encode()).hexdigest()
                    check_note = 'note' in d.keys()
                    if check_note is False:
                        d.update({'note': None})
                    d.update({'share_group': table_group_names.get(table_id), 'sha': row_sha})
                    shares.update({len(shares): d})

    # unique data adding selected isin data
    if selected_isin is not None:
        isin_d = selected_isin.get('isin')
        selected_cfi = [active_isin.get('cfi') for active_isin in active_isins if active_isin.get('isin') == isin_d][0]
        selected_isin.update({'cfi': selected_cfi})
        unique_data.update(selected_isin)
    if selected_isin is None and active_isins is not None:
        isin_copy = active_isins[0].copy()
        unique_data.update(isin_copy)
    row_sha = hashlib.sha3_256(str(unique_data).encode()).hexdigest()
    unique_data.update({'sha': row_sha})

    # parsing done
    details_parsed = DetailsParsed()
    details_parsed.active_isins = active_isins
    details_parsed.selected_isin = selected_isin
    details_parsed.basic_details = unique_data
    details_parsed.legal_shares = property_data
    details_parsed.percentage_shares = percentage_shares
    details_parsed.shares = shares
    return details_parsed



import os
from bs4 import BeautifulSoup

# htmls_dir = r'C:\Users\draganm\Documents\crhov\20191202'
# htmls = os.listdir(htmls_dir)
# for html_path in htmls:
#     # html_path = '07606273_RSVBSEE87204.html'
#     reg_number = re.search('(\d+)_', html_path).group(1)
#
#     full_html_path = '/'.join([htmls_dir, html_path])
#     print(full_html_path)
#     with open(full_html_path, 'rb') as f:
#         html = f.read()
#         soup = BeautifulSoup(html, 'lxml')
#         result = parse_details(soup, reg_number)
#         if type(result) is dict:
#             print(result)
#         else:
#             final_datas = ['basic_details', 'legal_shares', 'percentage_shares', 'shares']
#             basic_details = getattr(result, 'basic_details')
#             isin = basic_details.get('isin')
#             web_reg_num = basic_details.get('isin')
#             for final_data in final_datas:
#                 print(final_data)
#                 data_ready = getattr(result, final_data)
#                 if final_data == 'basic_details':
#                     data_ready = {0: data_ready}
#                 print(data_ready)
#                 # for k, v in data_ready.items():
#                 #     print(k, v)
#
#
#     # exit()