import os
import gc
import sys
import pickle
import hashlib
import asyncio
import getpass
from abc import ABC
from time import sleep
import concurrent.futures
from random import randint
from Scripts import filepath
from bs4 import BeautifulSoup
from multiprocessing import Process
from sqlalchemy.pool import NullPool
from datetime import datetime, timedelta
from core_framework.crawlers import CrawlerBase
from crawlers_sm.serbia.nbs.enforced_collection.parser import *
from crawlers_sm.serbia.monitoring.update import update_monitoring_rs

storage_location = filepath.set_storage()
print("base storage_location",  storage_location)

optimization_status = 'not optimized'
schema = 'nbs_pn'
base = 'https://www.nbs.rs'
enforced_collection = '/'.join([base, 'internet', 'latinica','67','pn.html'])
ws_url = 'https://webservices.nbs.rs/CommunicationOfficeService1_0/CompanyAccountXmlService.asmx?WSDL'

today_folder_name = datetime.now().strftime('%Y%m%d')
# nbs_pn_folder = r'C:\Users\{}\Documents\nbs_pn'.format(getpass.getuser())
nbs_pn_folder = os.path.join(storage_location, 'nbs_pn')
# nbs_pn_folder = os.path.join('E:\WebDocs', 'nbs_pn')

print("nbs_pn_folder",nbs_pn_folder)

six_hours = (datetime.now() - timedelta(hours=6))
one_day = (datetime.now() - timedelta(hours=24))
two_days = (datetime.now() - timedelta(hours=48))
five_days = (datetime.now() - timedelta(days=5))
seven_days = (datetime.now() - timedelta(days=7))
thirty_days = (datetime.now() - timedelta(days=30))
sixty_days = (datetime.now() - timedelta(days=60))
over_sixty_days = (datetime.now() - timedelta(days=61))


class NbsPrinudnaNaplata(CrawlerBase, ABC):
    def __init__(self):
        CrawlerBase.__init__(self, base)

        # runtime log initialization
        self.runtime_errors = 0
        self.runtime_status = 1
        self.runtime_start = datetime.now()
        self.runtime_id = hashlib.sha3_256(str(self.runtime_start).encode()).hexdigest()
        self.runtime_log = {'start_time': self.runtime_start, 'step_name': 'runtime'}

        # webservice variables
        self.ws_errors = 0
        self.exit_flag = False

        # crawler variables
        self.daily = True  # default global flag for type of crawler
        self.crawler_list = []
        self.crawler_fetched = 0
        self.crawler_errors = 0
        self.workers = 20
        self.blocked_radar = {}
        self.pstgr_id = 9

    def info_log(self, data):
        print("info_log", data)
        sha = hashlib.sha3_256(str(data).encode()).hexdigest()
        data.update({'sha': sha, 'runtime_id': self.runtime_id, 'end_time': datetime.now()})
        self.insert('l_info', {0: data}, schema=schema)

    def error_log(self, data):
        print("error_log", data)
        self.runtime_status = 2
        self.runtime_errors += 1
        sha = hashlib.sha3_256(str(data).encode()).hexdigest()
        data.update({'sha': sha, 'runtime_id': self.runtime_id, 'end_time': datetime.now()})
        self.insert('l_errors', {0: data}, schema=schema)

    def ws_status(self):
        if self.ws_errors > 0:
            return 2
        return 1

    def webservice_error(self, log_common):
        self.error_log(log_common)
        self.ws_errors += 1
        self.exit_flag = True

    def dir_creation_check(self):
        start_time = datetime.now()
        step_name = 'dir_creation_check'
        print(f"> {step_name}")
        log_common = {'start_time': start_time, 'step_name': step_name}
        try:
            folders = [nbs_pn_folder]
            for folder in folders:
                if os.path.isdir(folder) is False:
                    os.mkdir(folder)
            log_common.update({'status': 1, 'errors': 0})
            self.info_log(log_common)
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            log_common.update({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)})
            self.error_log(log_common)
            self.exit_flag = True

    def postgresql_conn(self):
        self.connect(self.pstgr_id, connect_args={"application_name": "nbs/enforced_collection/crawler.py postgresql_conn"})

    def webservice_count(self, date_from, date_to):
        proc_id = randint(1, 20000)
        start_time = datetime.now()
        step_name = 'webservice_count'
        print(f"> {step_name}")
        log_common = {'start_time': start_time, 'step_name': step_name}

        xml_envelope = self.xml_prepare('company_account_updated_count', (date_from, date_to))
        sha = self.new_request(proc_id=proc_id, protocols=['http'], new=True)
        r = self.requests.get(proc_id)
        r.request_type = 3
        r.ses.headers.update({'Host': 'webservices.nbs.rs', 'Content-Type': 'application/soap+xml; charset=utf-8'})
        args = {'data': xml_envelope, 'headers': r.ses.headers}

        change_number = -2
        try:
            args.update({'data': xml_envelope})
            args = {'data': xml_envelope, 'headers': r.ses.headers}
            r.go(ws_url, args=args)
            html_bs = r.response.content

            if type(html_bs) is dict:
                error = html_bs.get('RequestError')
                log_common.update({'err_desc': error, 'status': 2})
                self.webservice_error(log_common)
            else:
                change_number = number_of_changes(r)
                if change_number == -1:
                    error = 'could not find xml element GetCompanyAccountUpdatedCountResult'
                    log_common.update({'err_desc': error, 'status': 2})
                    self.webservice_error(log_common)
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            log_common.update({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2})
            self.webservice_error(log_common)
        finally:
            # release proxy back in use if it is not released it will be not be available for that web_base next 30min
            self.release_proxy(sha)

        log_common.update({'status': self.ws_status(), 'errors': self.ws_errors, 'note': f'number of changes: {change_number}'})
        self.info_log(log_common)
        return change_number

    def webservice_results(self, pages, date_from, date_to):
        rpp = 100  # rpp = records per page 100 is default
        proc_id = randint(1, 20000)
        start_time = datetime.now()
        step_name = 'webservice_results'
        print(f"> {step_name}")
        log_common = {'start_time': start_time, 'step_name': step_name}

        records = list(range(1, pages + 1))
        records_split = [records[i:i + rpp] for i in range(0, len(records), rpp)]
        record_ranges = [(min(x), max(x)) for x in records_split]

        bank_accounts = []
        for record_range in record_ranges:
            sleep(5)
            sys.stdout.write(f'\r + working on range {record_range}')
            start_item, end_item = record_range
            xml_envelope = self.xml_prepare('company_account_updated', (date_from, date_to, start_item, end_item))
            sha = self.new_request(proc_id=proc_id, protocols=['http'], new=True)
            r = self.requests.get(proc_id)
            r.request_type = 3
            r.ses.headers.update({'Host': 'webservices.nbs.rs', 'Content-Type': 'application/soap+xml; charset=utf-8'})
            args = {'data': xml_envelope, 'headers': r.ses.headers}

            try:
                args.update({'data': xml_envelope})
                args = {'data': xml_envelope, 'headers': r.ses.headers}
                r.go(ws_url, args=args)
                html_bs = r.response.content

                if type(html_bs) is dict:
                    log_copy = log_common.copy()
                    error = html_bs.get('RequestError')
                    error = f"error occured during geting records in range from {start_item} to {end_item}. {error}"
                    log_copy.update({'err_desc': error, 'status': 2})
                    self.webservice_error(log_copy)
                else:
                    company_account_data = ws_page_results(r)
                    if type(company_account_data) is dict:
                        log_copy = log_common.copy()
                        log_copy.update(company_account_data)
                        self.webservice_error(log_copy)
                    else:
                        if company_account_data:
                            bank_accounts.extend(company_account_data)
                        else:
                            log_copy = log_common.copy()
                            error = f"Returned no results for range from {start_item} to {end_item}"
                            log_copy.update({'err_desc': error, 'status': 2})
                            self.webservice_error(log_copy)
            except Exception as e:
                print("\n")
                log_copy = log_common.copy()
                exc_type, exc_obj, exc_tb = sys.exc_info()
                log_copy.update({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2})
                self.webservice_error(log_copy)
            finally:
                # release proxy back in use if it is not released it will be not be available for that web_base next 30min
                self.release_proxy(sha)
        print("\n")
        log_common.update({'status': self.ws_status(), 'errors': self.ws_errors, 'note': f'number of gathered records: {len(bank_accounts)}'})
        self.info_log(log_common)
        return bank_accounts

    def webservice(self):
        if self.exit_flag is True:
            return

        note = None
        current_hour = datetime.now().hour
        proc_id = randint(1, 20000)
        start_time = datetime.now()
        step_name = 'webservice'
        print(f"> {step_name}")
        log_common = {'start_time': start_time, 'step_name': step_name}
        if current_hour == 6:
            date_from = (datetime.now()-timedelta(days=2)).strftime('%Y%m%d')
        if current_hour < 21 and current_hour != 6:
            date_from = (datetime.now() - timedelta(days=1)).strftime('%Y%m%d')
        else:
            date_from = datetime.now().strftime('%Y%m%d')
        date_to = datetime.now().strftime('%Y%m%d')

        # manual date set if necessary
        # date_from = '20200110'
        # date_to = '20200110'

        change_number = self.webservice_count(date_from, date_to)
        complete_status = None
        if self.exit_flag is True:
            log_common.update({'status': self.ws_status(), 'errors': self.ws_errors, 'note': 'could not get number of changes'})
            self.info_log(log_common)
            return

        if change_number == 0:
            log_common.update({'status': self.ws_status(), 'errors': self.ws_errors, 'note': f'could not find any changes for period {date_from}-{date_to}'})
            self.info_log(log_common)
            return

        all_records = self.webservice_results(change_number, date_from, date_to)

        count_results = len(all_records)
        if change_number != count_results:
            note = f'number of changes({change_number}) is different from number of gathered records ({count_results})'
            complete_status = 3

        self.webservice_data_distribution(all_records)

        if complete_status is None:
            complete_status = self.ws_status()
        log_common.update({'status': complete_status, 'errors': self.ws_errors, 'note': note})
        self.info_log(log_common)

    def webservice_data_distribution(self, all_records):
        print("> ws data distribution")
        # data distribution
        if all_records:
            cnt = 0
            for record in all_records:
                cnt += 1
                sys.stdout.write(f"\r + progress: {cnt}/{len(all_records)}")
                # print(record)
                # all changes
                sha = record.get('sha')
                self.merge('account_changes', {0: record}, filters={'sha': sha}, schema=schema)

                # only account changes with enforced collection
                blockade_record = status_extract(record)
                if blockade_record is not None:
                    sha = blockade_record.get(sha)
                    self.merge('account_blockade', {0: blockade_record}, filters={'sha': sha}, schema=schema)

                # company_blockade = insert or update only company if one or more accounts is with status of enforced collection
                # company_changes = insert or update companies with any change

                company_inserts = {'company_blockade': {'filter_id': 2}, 'company_changes': {'filter_id': 2, 'blockade': False}}
                for table, filters in company_inserts.items():

                    company_record = status_extract(record, **filters)
                    if company_record is not None:
                        sha = company_record.get('sha')
                        registry_number = company_record.get('registry_number').strip()
                        vat_number = company_record.get('vat_number')
                        company_id = company_record.get('company_id')
                        update_date = company_record.get('update_date')

                        company_check = self.select(table, filters={'registry_number': registry_number}, schema=schema)
                        if company_check != [] and company_check is not None:
                            company_check_sha = company_check[0].get('sha')
                            last_upd_date = company_check[0].get('update_date')
                            last_upd_date = datetime(last_upd_date.year, last_upd_date.month, last_upd_date.day)
                            if company_check_sha != sha:
                                if table == 'company_blockade' and update_date > last_upd_date:
                                    if vat_number is not None:
                                        self.update(table, {'registry_number': registry_number}, {'update_date': update_date, 'night_check': None, 'vat_number': vat_number, 'company_id' :company_id, 'sha': sha}, schema=schema)
                                    else:
                                        self.update(table, {'registry_number': registry_number}, {'update_date': update_date, 'night_check': None, 'sha': sha}, schema=schema)
                                if table == 'company_changes' and update_date > last_upd_date:
                                    if vat_number is not None:
                                        self.update(table, {'registry_number': registry_number}, {'update_date': update_date, 'pu_check': None, 'rir_check': None, 'vat_number': vat_number, 'company_id' :company_id, 'sha': sha}, schema=schema)
                                    else:
                                        self.update(table, {'registry_number': registry_number}, {'update_date': update_date, 'pu_check': None, 'rir_check': None, 'sha': sha}, schema=schema)
                        if company_check == [] and company_check is not None:
                            self.insert(table, {0: company_record}, schema=schema)

    def xml_prepare(self, xml_name, args):

        xml_body = self.xml_loader(xml_name)
        xml_body = xml_body.format(*args)

        xml_envelope = self.xml_loader('master')
        xml_envelope = xml_envelope.format(xml_body)
        return xml_envelope

    def xml_loader(self, xml_name):
        with open(f'xml/{xml_name}.xml') as f:
            return f.read()

    def crawler(self):

        if self.exit_flag is True:
            return

        error_counter = 0
        start_time = datetime.now()
        step_name = 'crawler'
        print(f"> {step_name}")
        log_common = {'start_time': start_time, 'step_name': step_name}

        daily = True
        daily_min_h, daily_max_h = 7, 18
        night_min_h, night_max_h = 6, 19
        # get current hour of a day
        h = datetime.now().hour

        if daily_min_h <= h <= daily_max_h:
            print(">>>DAY MODE<<<")
            daily = True
        if h >=night_max_h or h <= night_min_h:
            print(">>>NIGHT MODE<<<")
            daily = False
        if daily is not None:
            self.crawler_prepare_list(daily)
        self.daily = daily
        # self.crawler_prepare_list(True)

        with concurrent.futures.ThreadPoolExecutor(max_workers=self.workers) as executor:
            # Start the load operations and mark each future with its URL
            future_to_url = {executor.submit(self.crawler_execute, page): page for page in self.crawler_list}
            for future in concurrent.futures.as_completed(future_to_url):
                url = future_to_url[future]
                try:
                    registry_number, crawl_status = future.result()
                except AttributeError as e:
                    print("-- 5", str(e))
                    pass
                except Exception as e:
                    exc_type, exc_obj, exc_tb = sys.exc_info()
                    print({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)})
                    log_copy = log_common.copy()
                    log_copy.update({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)})
                    self.error_log(log_copy)
                    error_counter += 1
                else:
                    # print("-", registry_number, crawl_status)
                    if type(registry_number) is dict:
                        log_copy = log_common.copy()
                        log_copy.update(registry_number)
                        self.error_log(log_copy)
                        error_counter += 1
        print("\r\n")
        log_type = log_common.copy()
        status = 1
        if error_counter > 0:
            status = 2
        log_type.update({'status': status, 'errors': error_counter})
        self.info_log(log_type)

    def crawler_prepare_list(self, daily):
        print(f"> preparing list daily: {daily}")
        table = 'company_blockade'

        if daily is True:
            priorities_24 = [51, 50, 49, 48, 47]
            priorities_5 = [46, 45, 44, 43, 42]
            priorities = priorities_24+priorities_5

            # reset daily check for 24h priority group
            for priority_24 in priorities_24:
                self.update(table, {'priority': priority_24, 'daily_check': True, 'check_date': f'<={one_day}'}, {'daily_check': None}, schema=schema)

            # reset daily check for every 5 days group
            for priority_5 in priorities_5:
                self.update(table, {'priority': priority_5, 'daily_check': True, 'check_date': f'<={five_days}'}, {'daily_check': None}, schema=schema)

            exteded_list = []
            for priority in priorities:
                self.crawler_list = self.select(table, filters={'daily_check': None, 'priority': priority}, schema=schema)
                if len(exteded_list) > 300:
                    self.crawler_list = exteded_list
                    break

                if len(self.crawler_list) < 300:
                    exteded_list.extend(self.crawler_list)
                    continue

            if not self.crawler_list:
                # reset daily check for 24h priority group if there is no old or new data to check
                for priority_24 in priorities_24:
                    self.update(table, {'priority': priority_24, 'daily_check': True, 'check_date': f'<={six_hours}'}, {'daily_check': None}, schema=schema)
                self.crawler_list = self.select(table, filters={'daily_check': None}, schema=schema)

            if not self.crawler_list:
                # this are accoutn bans only to check further if there is time
                self.update(table, {'priority': 42, 'daily_check': True, 'check_date': f'<={one_day}'}, {'daily_check': None}, schema=schema)
                self.crawler_list = self.select(table, filters={'daily_check': None}, schema=schema)

            if not self.crawler_list:
                wait_time = 3600
                for i in range(wait_time):
                    sys.stdout.write(f"\r> nothing left to do waiting: {i}/{wait_time}")
                    sleep(1)

        if daily is False:
            priorities_night = [51, 50]
            if datetime.now().hour < 21 and datetime.now().hour > 3:
                priorities_night = [51, 50, 49]
            self.crawler_list = self.select(table, filters={'night_check': None}, schema=schema)

            if not self.crawler_list:
                # update night_check column
                # those who have today update date
                self.update(table, {'update_date': datetime(datetime.now().year, datetime.now().month,datetime.now().day), 'night_check': True}, {'night_check': None}, schema=schema)
                # those who are in monitoring companies
                for priority_night in priorities_night:
                    self.update(table, {'priority': priority_night, 'night_check': True}, {'night_check': None}, schema=schema)
                self.crawler_list = self.select(table, filters={'night_check': None}, schema=schema)

            if not self.crawler_list:
                wait_time = 900
                for i in range(wait_time):
                    sys.stdout.write(f"\r> nothing left to do waiting: {i}/{wait_time}")
                    sleep(1)

        print(f"> list full size is: {len(self.crawler_list)} companies")
        if self.crawler_list:
            self.crawler_list = self.crawler_list[:300]

    def crawler_execute(self, data):
        # sleep(randint(10, 15))
        proc_id = randint(150, 55555)
        registry_number = data.get('registry_number')
        if registry_number is not None:
            registry_number = registry_number.strip()

        self.crawler_fetched += 1
        sys.stdout.write(f'\r> crawler progress: {self.crawler_fetched}/{len(self.crawler_list)}')

        try:
            regnum_folder = os.path.join(nbs_pn_folder, registry_number)
            today_regnum_folder = os.path.join(regnum_folder, today_folder_name)
            dir_create = [regnum_folder, today_regnum_folder]
            for directory in dir_create:
                if os.path.isdir(directory) is False:
                    os.mkdir(directory)
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)}
            print(error)
            return error, -1

        sha = None
        try:
            sleep(randint(2, 5))
            sha = self.new_request(proc_id=proc_id, protocols=['https'], new=True)
            sleep(randint(10, 20))
            r = self.requests.get(proc_id)
            r.request_type = 4
            # for row in self.crawler_list:

            payload = {'matbr': registry_number, 'type': 'pn_results'}
            r.payload = payload
            r.go('https://www.nbs.rs/rir_pn/pn_rir.html.jsp?type=pn_results&lang=SER_CIR&konverzija=no')

            html = r.response.content_raw
            html_bs = r.response.content
            if type(html_bs) is dict:
                # print("|", sha, html_bs)
                err_desc = html_bs.get('RequestError')
                # if err_desc != '' and err_desc != 'Page server is down':
                if err_desc == 'Blocked':
                    sleep(180)
                    sha_count = self.blocked_radar.get(sha)
                    if sha_count is None:
                        self.blocked_radar.update({sha: 1})
                    else:
                        sha_count += 1
                        self.blocked_radar.update({sha: sha_count})
                        if sha_count >= 5:
                            self.bad_proxy(sha, r.proxy_data, web_page=base)
                            self.blocked_radar.pop(sha, None)
                    self.bad_proxy(sha, r.proxy_data, web_page=base)
                if err_desc == 'Page server is down':
                    sleep(600)
                    error = {'err_desc': err_desc}
                    return error, -1
                return registry_number, -1

            # try to write html on disc if an error occurred just skip
            try:
                html_name = f"{datetime.now().strftime('%H%M')}.html"
                html_path = os.path.join(today_regnum_folder, html_name)
                with open(html_path, 'wb') as f:
                    f.write(html)
            except Exception as e:
                exc_type, exc_obj, exc_tb = sys.exc_info()
                error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)}
                print(error)
                return error, -1

            # for tests - be sure to comment above part where sending request
            # with open('path to html', 'rb') as f:
            #     html = f.read()
            # with open(f'htmls/test2/step1.html', 'rb') as f:
            #     html = f.read()

            # parsing html
            final_data = web_parser(html, registry_number)

            if type(final_data) is dict:
                print(final_data)
                return final_data, -1

            company_status = final_data.company_status
            company_data = final_data.company_data

            blockade_dates = final_data.dates_data
            blockade_amounts = final_data.blockade_amount
            blockade_stats = final_data.blockade_stats
            blockade_bans = final_data.ban_data

            history_blockade_dates = final_data.history_dates_data

            # if company status is 102 or 103 then there is no open blockades
            now = pd.datetime.now()

            amount = '0'
            if blockade_amounts:
                get_amount = blockade_amounts.get('blockade_amount')
                if get_amount is not None:
                    amount = get_amount

            if company_status in (102, 103):
                # add in blockade_amounts that amout is 0
                blockade_amounts2 = {}
                blockade_amounts_copy = {'registry_number': registry_number, 'blockade_amount': '0'}
                amount_data = self.select('blockade_amounts', filters={'registry_number': registry_number}, schema=schema)
                for row in amount_data:
                    blockade_amounts2.update({len(blockade_amounts2): row})
                df = pd.DataFrame.from_dict(blockade_amounts2, orient='index')
                if df.empty is False:
                    # get last closed blockade date range from webpage and his sha
                    df = df.where(df['date_created'] == df['date_created'].max())
                    df.dropna(how='all', inplace=True)
                    last_row = df.iloc[0]
                    ban_id_db = last_row.get('ban_id')
                    blockade_id_db = last_row.get('blockade_id')
                    blockade_amounts_copy.update({'ban_id': ban_id_db})
                    blockade_amounts_copy.update({'blockade_id': blockade_id_db})
                block_amount_sha = final_data.get_sha(blockade_amounts_copy)
                blockade_amounts_copy.update({'sha': block_amount_sha})
                self.merge('blockade_amounts', {0: blockade_amounts_copy}, filters={'registry_number': registry_number, 'archive': None}, schema=schema)

                # add new record in blockade_status
                self.merge('blockade_status', {0: blockade_stats}, {'registry_number': registry_number, 'archive': None}, schema=schema)

                # close any open blockade date range, because we don't know when it is closed we are closing on today day
                new_sha_data = {'registry_number': registry_number, 'blockade_to': now}
                new_sha = final_data.get_sha(new_sha_data)
                self.update('blockade_dates',  {'registry_number': registry_number, 'archive': None, 'blockade_to': None}, {'blockade_to': now, 'sha': new_sha, 'removed': company_status}, schema=schema)

                # close all records for company bans and amounts
                # self.update('blockade_amounts', {'registry_number': registry_number, 'archive': None}, {'archive': now, 'removed': company_status}, schema=schema)
                self.update('blockade_bans', {'registry_number': registry_number, 'archive': None}, {'archive': now, 'removed': company_status}, schema=schema)

                # insert company data if there is any
                if company_data:
                    self.merge('blockade_company', {0: company_data}, {'registry_number': registry_number, 'archive': None}, schema=schema)

                # update blockade amount of currently open blockade in database and close the date of blockade, asign new sha
                self.update('blockade_latest', {'registry_number': registry_number, 'archive': None, 'blockade_to': None}, {'blockade_to': now, 'sha': new_sha, 'blockade_amount': '0'}, schema=schema)

                # update blockade amount of closed blockades in database
                self.update('blockade_latest', {'registry_number': registry_number, 'blockade_to': True}, {'blockade_amount': '0'}, schema=schema)

                # update company_blockade table
                self.update('company_blockade', {'registry_number': registry_number}, {'daily_check': 1, 'check_date': now, 'blockade_id': 2, 'acc_ban_id': 2, 'blockade_amount': '0'}, schema=schema)
                self.update('company_blockade', {'registry_number': registry_number}, {'night_check': 1, 'check_date': now, 'blockade_id': 2, 'acc_ban_id': 2, 'blockade_amount': '0'}, schema=schema)
            else:
                # if there is and open blockade date
                if blockade_dates:
                    op_block_sha = blockade_dates.get('sha')
                    # check is there any record in database with same sha and archive is null (active record)
                    open_block_data = self.select('blockade_dates', filters={'registry_number': registry_number, 'archive': None, 'sha': op_block_sha}, schema=schema)
                    # if there is record with that sha we compare the current blockade_id with existing
                    if open_block_data:
                        # TODO: if there is more than one sha for some weird reason
                        open_block_data = open_block_data[0]
                        db_blockade_id = open_block_data.get('blockade_id')
                        new_blockade_id = blockade_dates.get('blockade_id')
                        # if database and new blockade id are not the same use db blockade id and update that id in blockade_amounts and blockade_dates
                        if db_blockade_id != new_blockade_id:
                            blockade_dates.update({'blockade_id': db_blockade_id})
                            if blockade_amounts:
                                blockade_amounts.update({'blockade_id': db_blockade_id})

                        # update blockade amount of currently open blockade in database
                        self.update('blockade_latest', {'registry_number': registry_number, 'archive': None, 'blockade_to': None}, {'blockade_amount': amount}, schema=schema)
                    else:
                        # if it doesn't exist in db, close current open blockade in db if exist for some reason and insert new one
                        new_sha_data = {'registry_number': registry_number, 'blockade_to': now}
                        new_sha = final_data.get_sha(new_sha_data)

                        self.update('blockade_dates', {'registry_number': registry_number, 'archive': None, 'blockade_to': None},
                                    {'blockade_to': now, 'sha': new_sha, 'removed': 404}, schema=schema)

                        self.insert('blockade_dates', {0: blockade_dates}, schema=schema)

                        # close current open blockade date record if exists and add new one
                        copy_blockade_dates = blockade_dates.copy()
                        copy_blockade_dates.update({'blockade_amount': amount})
                        copy_blockade_dates.pop('blockade_id')
                        self.update('blockade_latest', {'registry_number': registry_number, 'archive': None, 'blockade_to': None},
                                    {'blockade_to': now, 'sha': new_sha}, schema=schema)

                        self.insert('blockade_latest', {0: copy_blockade_dates}, schema=schema)

                # if there is no open blockade on webpage
                if not blockade_dates:
                    # and if there is open blockade date in database and history dates exists
                    open_block_data = self.select('blockade_dates', filters={'registry_number': registry_number, 'archive': None, 'blockade_to': None}, schema=schema)
                    if open_block_data and history_blockade_dates:
                        df = pd.DataFrame.from_dict(history_blockade_dates, orient='index')
                        if df.empty is False:
                            # get last closed blockade date range from webpage and his sha
                            df = df.where(df['blockade_to'] == df['blockade_to'].max())
                            df.dropna(how='all', inplace=True)
                            last_sha = df['sha'].iloc[0]
                            blockade_to = df['blockade_to'].iloc[0]

                            # close the open blockade range with data available from last closed blockade date
                            self.update('blockade_dates', {'registry_number': registry_number, 'archive': None, 'blockade_to': None},
                                        {'blockade_to': blockade_to, 'sha': last_sha}, schema=schema)

                            # close open currently open blockade date and update blockade amount
                            self.update('blockade_latest', {'registry_number': registry_number, 'archive': None, 'blockade_to': None},
                                        {'blockade_to': blockade_to, 'sha': last_sha, 'blockade_amount': '0'}, schema=schema)

                if history_blockade_dates:
                    # compare website history data with db history dates, add new ones don't close old ones
                    self.merge('blockade_dates', history_blockade_dates, filters={'registry_number': registry_number, 'archive': None, 'blockade_to': True}, update=False, on=['sha'], schema=schema)
                    self.merge('blockade_latest', history_blockade_dates, filters={'registry_number': registry_number, 'archive': None, 'blockade_to': True}, update=False, schema=schema)

                if blockade_bans:
                    self.merge('blockade_bans', {0: blockade_bans}, filters={'registry_number': registry_number, 'archive': None}, schema=schema)
                else:
                    self.update('blockade_bans', {'registry_number': registry_number, 'archive': None}, {'archive': now}, schema=schema)

                # calculate sha
                keys = ['blockade_amount', 'registry_number', 'ban_id', 'blockade_id']
                if blockade_amounts:
                    blockade_amounts_copy = blockade_amounts.copy()
                    amount_data = self.select('blockade_amounts', filters={'registry_number': registry_number, 'archive': None}, schema=schema)
                    if amount_data:
                        amount_data = amount_data[0]
                        ban_id_db = amount_data.get('ban_id')
                        blockade_id_db = amount_data.get('blockade_id')
                        if blockade_amounts_copy.get('ban_id') is None and ban_id_db is not None:
                            blockade_amounts_copy.update({'ban_id': ban_id_db})
                        if blockade_amounts_copy.get('blockade_id') is None  and blockade_id_db is not None:
                            blockade_amounts_copy.update({'blockade_id': blockade_id_db})

                    blockade_amounts_copy = {k:v for k,v in blockade_amounts_copy.items() if k in keys}
                    block_sha = final_data.get_sha(blockade_amounts_copy)
                    blockade_amounts_copy.update({'sha': block_sha})

                    if blockade_dates or blockade_bans:
                        self.merge('blockade_amounts', {0: blockade_amounts_copy}, filters={'registry_number': registry_number, 'archive': None}, schema=schema)
                    else:
                        amount_to_num = pd.to_numeric(amount, errors='coerce')
                        if not pd.isna(amount_to_num):
                            if amount_to_num > 0:
                                company_status = 204
                        # updating data for blockade_status
                        blockade_stats.update({'company_status': company_status})
                        blockade_stats_copy = blockade_stats.copy()
                        if blockade_stats_copy.get('sha') is not None:
                            blockade_stats_copy.pop('sha')
                        block_stat_sha = final_data.get_sha(blockade_stats_copy)
                        blockade_stats.update({'sha': block_stat_sha})

                        blockade_amounts_copy = blockade_amounts.copy()
                        # get last record that is created in blockade_amounts and his blockade_id and ban_id
                        blockade_amounts2 = {}
                        block_amount_data_hist = self.select('blockade_amounts', filters={'registry_number': registry_number}, schema=schema)
                        for row in block_amount_data_hist:
                            blockade_amounts2.update({len(blockade_amounts2): row})
                        df = pd.DataFrame.from_dict(blockade_amounts2, orient='index')
                        if df.empty is False:
                            # get last closed blockade date range from webpage and his sha
                            df = df.where(df['date_created'] == df['date_created'].max())
                            df.dropna(how='all', inplace=True)
                            last_row = df.iloc[0]
                            last_row = last_row.to_dict()
                            ban_id_last = last_row.get('ban_id')
                            blockade_id_last = last_row.get('blockade_id')
                            blockade_amounts_copy.update({'ban_id': ban_id_last, 'blockade_id': blockade_id_last})

                        blockade_amounts_copy = {k: v for k, v in blockade_amounts_copy.items() if k in keys}
                        block_sha = final_data.get_sha(blockade_amounts_copy)
                        blockade_amounts_copy.update({'sha': block_sha})

                        # update and close record if it is not archived yet
                        # self.update('blockade_amounts', {'registry_number': registry_number, 'archive': None}, {'sha': block_sha, 'archive': datetime.now(), 'removed': company_status}, schema=schema)

                        # add new record if it doesn't match with any existing closed record
                        self.merge('blockade_amounts', {0: blockade_amounts_copy}, filters={'registry_number': registry_number, 'archive': None}, update=False, schema=schema)

                # insert company data if there is any
                if company_data:
                    self.merge('blockade_company', {0: company_data}, {'registry_number': registry_number, 'archive': None}, schema=schema)

                    copy_company_data = company_data.copy()
                    if copy_company_data.get('sha') is not None:
                        copy_company_data.pop('sha')
                    self.update('blockade_latest', {'registry_number': registry_number, 'archive': None}, copy_company_data, schema=schema)

                # if there is no open blockade ranges and there is no open company ban restrictions
                if not blockade_dates and not blockade_bans:
                    # close all records for company bans and amounts
                    # self.update('blockade_amounts', {'registry_number': registry_number, 'archive': None}, {'archive': now, 'removed': company_status}, schema=schema)
                    self.update('blockade_bans', {'registry_number': registry_number, 'archive': None}, {'archive': now, 'removed': company_status}, schema=schema)

                # add new record in blockade_status
                self.merge('blockade_status', {0: blockade_stats}, {'registry_number': registry_number, 'archive': None}, schema=schema)

            self.update('blockade_latest', {'registry_number': registry_number, 'archive': None}, {'transfer': None}, schema=schema)

            # update company_blockade table
            blockade_stat_combos = {200: (2, 2), 201: (2, 1), 202: (1, 2), 203: (1, 1), 204: (2, 2)}
            combo_id = blockade_stat_combos.get(company_status)
            if combo_id is not None:
                blockade_id, acc_ban_id = combo_id
                update_data = {'check_date': now, 'blockade_id': blockade_id, 'acc_ban_id': acc_ban_id, 'blockade_amount': amount}
                company_filter = {'registry_number': registry_number}
                if self.daily is True:
                    update_data.update({'daily_check': 1})
                    self.update('company_blockade', company_filter, update_data, schema=schema)
                if self.daily is False:
                    update_data.update({'night_check': 1})
                    self.update('company_blockade', company_filter, update_data, schema=schema)

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            print({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)})
            return registry_number, -1
        finally:
            # release proxy back in use if it is not released it will be not be available for that web_base next 30min
            self.release_proxy(sha)
        return registry_number, 200

    def transfer_data(self):
        note = None
        start_time = datetime.now()
        step_name = 'transfer_data'
        print(f"> {step_name}")
        log_common = {'start_time': start_time, 'step_name': step_name}

        transfer_errors = 0
        recheck_companies = []
        solved, transferred, trans_acc_change, ban_sync, company_stats = [], [], [], [], []
        company_stats_fail = []
        current_hour = datetime.now().hour

        try:
            blockade_latest = self.select('blockade_latest', filters={'transfer': None, 'archive': None}, schema=schema)
            if 3 < current_hour < 6:
                return
            # if hours are 7-23 then transfer data if total data size is over 2k of records
            if current_hour > 19 and len(blockade_latest) < 2000:
                return
            self.update('account_changes', {'blockade_id': None}, {'blockade_id': 0}, schema=schema)
            closed_blokades = self.select('blockade_status', filters={'transfer': None, 'archive': None, 'company_status': [102, 103]}, schema=schema)
            company_status = self.select('blockade_status', filters={'transfer': None, 'archive': None}, schema=schema)
            account_changes = self.select('account_changes', filters={'transfer': None}, schema=schema)
            ban_changes = self.select('blockade_bans', filters={'transfer': None}, schema=schema)

            date_change = datetime.now() - timedelta(days=1)
            bans = self.select('blockade_bans', filters={'archive': f'>={date_change}'}, schema='nbs_pn')
            ban_companies = []
            for ban in bans:
                registry_number_ban = ban.get('registry_number')
                ban_data = self.select('blockade_bans', filters={'registry_number': registry_number_ban, 'archive': None}, schema='nbs_pn')
                replace_ban = {'registry_number': 'mb', 'ban_date': 'dat_zabr', 'archive': 'dat_pro', 'sha': 'sha', 'date_created': 'date_created'}
                new_ban = {}
                for k, v in ban.items():
                    col_name = replace_ban.get(k)
                    if col_name is not None:
                        new_ban.update({col_name: v})
                if not ban_data:
                    if new_ban.get('mb') is not None:
                        ban_companies.append(new_ban)

            print("> preparing data")
            # order data from oldest update made on bank account towards new
            df = pd.DataFrame.from_records(account_changes)
            if df.empty is False:
                df.sort_values(by=['update_date'], inplace=True)
                account_changes = df.to_dict(orient='index')
            if not account_changes:
                account_changes = {}

            # Populate transfer sha in blockade_latest pstg
            columns = ['registry_number', 'blockade_from', 'blockade_to', 'blockade_amount']
            repl_names = {'registry_number': 'mb', 'blockade_from': 'dat_od', 'blockade_to': 'dat_do', 'blockade_amount': 'iznos_blokade',
                          'debtor_name': 'duznik', 'address': 'adresa', 'place': 'mjesto', 'transfer_sha': 'transfer_sha', 'vat_number': 'pib'}
            parser_data = ParserData()
            updated = 0
            group_data = {}
            for row in blockade_latest:
                sha = row.get('sha')
                registry_number = row.get('registry_number')

                row_copy = row.copy()
                row_copy.pop('sha')
                row_copy = {k: v for k,v in row_copy.items() if k in columns}
                transfer_sha = parser_data.get_sha(row_copy)

                self.update('blockade_latest', {'transfer': None, 'sha': sha}, {'transfer_sha': transfer_sha}, schema=schema)
                row.update({'transfer_sha': transfer_sha})

                reg_data = group_data.get(registry_number)
                insert_row = {}
                for k, v in row.items():
                    new_name = repl_names.get(k)
                    if new_name is not None:
                        if pd.isna(v) is True:
                            v = None
                        insert_row.update({new_name: v})
                if insert_row:
                    if reg_data is None:
                        group_data.update({registry_number: {0: insert_row}})
                    else:
                        reg_data.update({len(reg_data): insert_row})
                updated += 1
                sys.stdout.write(f'\r + updated {updated}/{len(blockade_latest)}')
            sys.stdout.write(f'\r + updated {updated}/{len(blockade_latest)}')

            # update web_change date
            print("\n> update web change")
            updated = 0
            for company_stat in company_status:
                registry_number = company_stat.get('registry_number')
                solved_sha = company_stat.get('sha')
                change_date = company_stat.get('date_created')
                company_status_col = company_stat.get('company_status')
                update_status = self.update('company_blockade', {'registry_number': registry_number}, {'web_change': change_date}, schema=schema)
                if company_status_col not in [102, 103, 200]:
                    if update_status == 200:
                        company_stats.append(solved_sha)
                        updated += 1
                sys.stdout.write(f'\r + updated {updated}/{len(company_status)}')
            sys.stdout.write(f'\r + updated {updated}/{len(company_status)}')

            # update mssql tables
            self.connect(2, archive_date='dat_pro')
            print("\n> sync bans")
            bans_upd = 0
            for ban in ban_companies:
                bans_upd += 1
                registry_number_ban = ban.get('registry_number')
                merge_status = self.merge('NBS_PN_ZABRANE', {0: ban}, filters={'dat_pro': None, 'mb': registry_number_ban}, on=['sha'])
                sys.stdout.write(f"\r + check {bans_upd}/{len(ban_companies)}")
            sys.stdout.write(f"\r + check {bans_upd}/{len(ban_companies)}")

            print('\n> sync blockades')
            if blockade_latest and group_data:
                self.transfer_mssql_updates()
                # merge data into mssql tables
                print("\n> updating nbs_prinudna_naplata table")
                for registry_number, data in group_data.items():
                    merge_status = self.merge('NBS_PRINUDNA_NAPLATA', data, filters={'mb': registry_number, 'dat_pro': None}, on=['transfer_sha'])
                    if merge_status == 200:
                        update_status = self.update('NBS_PRINUDNA_NAPLATA',  {'MB': registry_number, 'DAT_PRO': None}, {'DATE_CHECKED': datetime.now()}, freeze=True)
                        if update_status == 200:
                            transferred.append(registry_number)
                        sys.stdout.write(f'\r + updated {len(transferred)}/{len(group_data)}')
            sys.stdout.write(f'\r + updated {len(transferred)}/{len(group_data)}')

            print ("\n> inserting status blockades")
            updated = 0
            for company_stat in company_status:
                registry_number = company_stat.get('registry_number')
                solved_sha = company_stat.get('sha')
                date_created = company_stat.get('date_created')
                pn_stat = company_stat.get('company_status')
                if pn_stat in [200, 201, 202, 203, 204]:
                    pn_stat = 200
                insert_payload = {'mb': registry_number, 'pn_stat': pn_stat, 'date_created': date_created}
                insert_status = self.insert('NBS_SKIDANJE_STATUS2', {0: insert_payload})

                if insert_status != 200:
                    company_stats_fail.append(solved_sha)
                else:
                    if solved_sha not in company_stats:
                        company_stats.append(solved_sha)
                    updated += 1
                sys.stdout.write(f'\r + updated {updated}/{len(company_status)}')
            sys.stdout.write(f'\r + updated {updated}/{len(company_status)}')

            print('\n> close 102 and 103 cases')
            for close_date in closed_blokades:
                registry_number = close_date.get('registry_number')
                solved_sha = close_date.get('sha')
                update_status = self.update('NBS_PRINUDNA_NAPLATA', {'MB': registry_number, 'DAT_PRO': None}, {'DAT_PRO': datetime.now(), 'UPD': None}, freeze=True)
                update_status2 = self.update('NBS_PN_ZABRANE', {'MB': registry_number, 'DAT_PRO': None}, {'DAT_PRO': datetime.now()}, freeze=True)
                if update_status == 200 and update_status2 == 200:
                    solved.append(solved_sha)
                sys.stdout.write(f'\r + updated {len(solved)}/{len(closed_blokades)}')
            sys.stdout.write(f'\r + updated {len(solved)}/{len(closed_blokades)}')

            print("\n> sync ban dates")
            # ban_changes switch colum names to mssql
            repl_col = {'registry_number': 'mb', 'sha': 'sha', 'ban_date': 'dat_zabr'}
            for ban_change in ban_changes:
                ban_sha = ban_change.get('sha')
                ban_change = {repl_col.get(k):v for k,v in ban_change.items() if k in repl_col.keys()}
                if len(ban_change) == len(repl_col):
                    mb = ban_change.get('mb')
                    merge_status = self.merge('NBS_PN_ZABRANE', {0: ban_change}, filters={'mb': mb, 'dat_pro': None})
                    if merge_status == 200:
                        ban_sync.append(ban_sha)
                        sys.stdout.write(f'\r + updated {len(ban_sync)}/{len(ban_changes)}')
            sys.stdout.write(f'\r + updated {len(ban_sync)}/{len(ban_changes)}')

            print(f'\n> sync account changes {len(account_changes)}')
            self.connect(2, archive_date='archive_date')
            for rowid, account_change in account_changes.items():
                account_id = account_change.get('account_id')
                row_sha = account_change.get('sha')

                col_repl = {'registry_number': 'NATIONAL_IDENTIFICATION_NUMBER', 'vat_number': 'TAX_IDENTIFICATION_NUMBER', 'bank_account': 'ACCOUNT',
                            'bank_code': 'BANKCODE', 'account_number': 'ACCOUNT_NUMBER', 'control_number': 'CONTOROL_NUMBER',
                            'company_name': 'COMPANY_NAME', 'address': 'ADDRESS', 'city': 'CITY', 'activity_code': 'ACTIVITY_CODE',
                            'municipality_code': 'MUNICIPALITY_CODE', 'municipality_name': 'MUNICIPALITY_NAME', 'activity_name': 'ACTIVITY_NAME',
                            'bank_name': 'BANK_NAME', 'status_id': 'COMPANY_ACCOUNT_STATUS_ID', 'blockade_id': 'COMPANY_ACCOUNT_BLOCKADE_STATUS_ID',
                            'type_id': 'COMPANY_ACCOUNT_TYPE_ID', 'company_type_id': 'LEGAL_USER_TYPE_ID', 'initialization_date': 'INITIALIZATION_DATE',
                            'change_date': 'CHANGE_DATE', 'update_date': 'UPDATE_DATE', 'account_id': 'account_sha', 'transfer_sha': 'sha',
                            'date_created': 'DATE_CREATED'}

                account_data = {}
                account_data_keys = ['company_id', 'transfer', 'transfer_date', 'sha']
                for k, v in account_change.items():
                    if k in account_data_keys:
                        continue
                    key_name = col_repl.get(k)
                    if key_name is not None:
                        account_data.update({key_name: v})

                done_count = 0
                account_data.update({'CREATED_BY': 'transfer'})
                precheck_record = self.select('COMPANY_ACCOUNT_UPDATES', filters={'account_sha': account_id, 'date_created': account_data.get('DATE_CREATED')}, schema='NBS_WS')
                merge_status = 200
                if not precheck_record:
                    merge_status = self.merge('COMPANY_ACCOUNT_UPDATES', {0: account_data}, filters={'account_sha': account_id, 'archive_date': None}, schema='NBS_WS')

                if merge_status == 200:
                    done_count += 1

                success = 200
                account_data.update({'ENTRY_DATE_UPDATED': datetime.now(), 'ENTRY_UPDATED_BY': 'transfer'})
                db_account_data = self.select('COMPANY_ACCOUNTS',  filters={'account_sha': account_id}, schema='NBS_WS')
                if not db_account_data:
                    success = self.insert('COMPANY_ACCOUNTS', {0: account_data}, schema='NBS_WS')
                else:
                    db_account_data = db_account_data[0]
                    row_data_sha = db_account_data.get('sha')
                    transfer_sha = account_data.get('sha')
                    updatable_columns = ['COMPANY_NAME', 'ADDRESS', 'CITY', 'MUNICIPALITY_CODE', 'ACTIVITY_CODE', 'MUNICIPALITY_NAME',
                                         'ACTIVITY_NAME', 'BANK_NAME', 'COMPANY_ACCOUNT_STATUS_ID', 'COMPANY_ACCOUNT_BLOCKADE_STATUS_ID',
                                         'COMPANY_ACCOUNT_TYPE_ID', 'LEGAL_USER_TYPE_ID', 'INITIALIZATION_DATE', 'CHANGE_DATE', 'UPDATE_DATE',
                                         'sha', 'ENTRY_DATE_UPDATED', 'ENTRY_UPDATED_BY']
                    account_data = {k:v for k,v in account_data.items() if k in updatable_columns}
                    if row_data_sha != transfer_sha:
                        success = self.update('COMPANY_ACCOUNTS', {'account_sha': account_id}, account_data, schema='NBS_WS', freeze=True)
                if success == 200:
                    done_count += 1

                if done_count == 2:
                    trans_acc_change.append(row_sha)
                sys.stdout.write(f'\r + transfering: {len(trans_acc_change)}/{len(account_changes)}')
            sys.stdout.write(f'\r + transfering: {len(trans_acc_change)}/{len(account_changes)}')

            # remove failed transfers
            print("\n> reseting failed transfers")
            for company_fail in company_stats_fail:
                if company_fail in company_stats:
                    company_stats.remove(company_fail)
                if company_fail in solved:
                    solved.remove(company_fail)

            # rsservis integration -- fugly peace of code -- work on it ones rsservis is upgraded
            recheck_companies = []
            companies_regs = []
            recheck_companies1 = self.select('rsservis_nbs_nerijeseni', view=True)
            recheck_companies2 = self.select('rsservis_nbs_nerijeseni2', view=True)

            check_companies = recheck_companies1 + recheck_companies2
            for recheck_company in check_companies:
                registry_number = recheck_company.get('registry_number')
                if registry_number not in companies_regs:
                    recheck_companies.append(recheck_company)
                    companies_regs.append(registry_number)

            print(f"> rs servis sync: {len(recheck_companies)}")
            # relevant procedures in mssql: proc_update_source
            self.postgresql_conn()
            updatable = []
            for recheck_company in recheck_companies:
                registry_number = recheck_company.get('registry_number')
                vat_number = recheck_company.get('vat_number')
                last_date = recheck_company.get('nbs')
                date_created = recheck_company.get('date_created')

                if last_date is not None:
                    last_date = pd.Timestamp(last_date)

                registry_data = self.select('company_blockade', filters={'registry_number': registry_number}, schema=schema)
                # case when registry data exists
                if registry_data:
                    registry_data = registry_data[0]
                    check_date = registry_data.get('check_date')
                    if last_date is None and check_date is not None:
                        date_created = pd.Timestamp(date_created)
                        if date_created < check_date:
                            updatable.append([{'mbr': registry_number}, {'nbs': check_date}])
                        else:
                            self.update('company_blockade', {'registry_number': registry_number}, {'daily_check': None, 'priority': 50}, schema=schema)
                    if last_date is not None and check_date is not None:
                        if last_date < check_date:
                            updatable.append([{'mbr': registry_number}, {'nbs': check_date}])
                        else:
                            self.update('company_blockade', {'registry_number': registry_number}, {'daily_check': None, 'priority': 50}, schema=schema)
                    if last_date is None and check_date is None:
                        self.update('company_blockade', {'registry_number': registry_number}, {'daily_check': None, 'priority': 50}, schema=schema)
                # case when registry number does not exist
                else:
                    if vat_number is not None:
                        company_id_keys = ['registry_number', 'vat_number']
                        company_id = {x: recheck_company.get(x) for x in company_id_keys}
                        company_id = hashlib.sha3_256(str(company_id).encode()).hexdigest()
                        payload = {'sha': 'rsservis', 'company_id': company_id, 'registry_number': registry_number,
                                   'vat_number': vat_number, 'update_date': datetime.now(), 'night_check': 1}
                        self.insert('company_blockade', {0: payload}, schema=schema)

            self.connect(2, archive_date='dat_pro')
            for upd in updatable:
                filters, update_data = upd
                update_status = self.update('subject_source_update', filters, update_data)

            self.proc('proc_rs_servis_update')

        except Exception as e:
            self.postgresql_conn()
            transfer_errors += 1
            exc_type, exc_obj, exc_tb = sys.exc_info()
            print({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)})
            log_copy = log_common.copy()
            log_copy.update({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)})
            self.error_log(log_copy)
        finally:
            print('> set complete for every successful transferred row')
            self.postgresql_conn()

            # mark as transferred
            for reg_num_trans in transferred:
                self.update('blockade_latest', {'registry_number': reg_num_trans, 'transfer': None}, {'transfer': 1, 'transfer_date': datetime.now()}, schema=schema)

            for solved_sha in solved:
                self.update('blockade_status', {'sha': solved_sha, 'transfer': None}, {'transfer': 1, 'transfer_date': datetime.now()}, schema=schema)

            for row_sha_id in trans_acc_change:
                self.update('account_changes', {'sha': row_sha_id, 'transfer': None}, {'transfer': 1, 'transfer_date': datetime.now()}, schema=schema)

            for ban_sha in ban_sync:
                self.update('blockade_bans', {'sha': ban_sha, 'transfer': None}, {'transfer': 1, 'transfer_date': datetime.now()}, schema=schema)

            for solved_sha in company_stats:
                self.update('blockade_status', {'sha': solved_sha, 'transfer': None}, {'transfer': 1, 'transfer_date': datetime.now()}, schema=schema)

        error_status = 1
        if transfer_errors > 0:
            error_status = 2

        log_common.update({'status': error_status, 'errors': transfer_errors, 'note': note})
        self.info_log(log_common)

    def transfer_mssql_updates(self):
        print('> mssql_sha_updates')
        # update transfer_sha in mssql db table NBS_PRINUDNA_NAPLATA
        nbs_pn_mssql = self.select('v_nbs_pn_transfer', view=True)
        parser_data = ParserData()
        updated = 0
        for row in nbs_pn_mssql:
            col_rep = {'mb': 'registry_number', 'dat_od': 'blockade_from', 'dat_do': 'blockade_to', 'iznos_blokade': 'blockade_amount'}
            val_rep = {'dat_od': pd.NaT, 'dat_do': pd.NaT, 'iznos_blokade': 0.0}
            for k, v in val_rep.items():
                k_val = row.get(k)
                if k_val is None:
                    row.update({k: v})
                k_val = row.get(k)
                if k in ['dat_od', 'dat_do'] :
                    if pd.isna(k_val) is False:
                        row.update({k: pd.Timestamp(k_val)})
            row2 = {col_rep.get(k): v for k, v in row.items() if k in col_rep.keys()}
            transfer_sha = parser_data.get_sha(row2)
            id = row.get('id2')
            self.update('NBS_PRINUDNA_NAPLATA', {'id': id}, {'transfer_sha': transfer_sha})
            updated += 1
            sys.stdout.write(f'\r> updated {updated}/{len(nbs_pn_mssql)}')
        self.disconnect()
        sys.stdout.write(f'\r> updated {updated}/{len(nbs_pn_mssql)}')

    def priority_update(self):
        note = None
        start_time = datetime.now()
        step_name = 'priority_update'
        print(f"> {step_name}")
        log_common = {'start_time': start_time, 'step_name': step_name}

        # update company list in monitoring
        print("> update monitoring rs")
        update_monitoring_rs(self.pstgr_id)

        print("> fetch companies in monitoring")
        monitoring_companies = self.select('rs_companies', filters={'archive': None}, schema='monitoring')
        monitoring_companies = [row.get('registry_number') for row in monitoring_companies]

        # assign priority to all blockades that are not checked yet in daily check
        print("> daily company priorities ")
        table = 'company_blockade'

        print(" + clean priorities")
        self.update(table, {'priority': True,}, {'priority': None}, schema=schema)

        print(" + monitoring companies")
        #  top priority goes to companies that are in monitoring list
        companies = self.select(table, filters={'or': [{'daily_check': None}, {'blockade_id': 1}, {'acc_ban_id': 1}]}, schema=schema)
        for company in companies:
            registry_number = company.get('registry_number')
            company_filter = {'registry_number': registry_number}
            if registry_number in monitoring_companies:
                self.update(table, company_filter, {'priority': 51}, schema=schema)

        print(" + new companies")
        # second priority goes to companies that are not check not even once
        self.update(table, {'daily_check': None, 'blockade_id': None, 'priority': None}, {'priority': 50}, schema=schema)

        print(" + last 2 days companies")
        # next priority goes to companies changes that occurred in last 2 days on webservice or web page
        self.update(table, {'update_date': f'>={two_days}', 'priority': None}, {'priority': 49}, schema=schema)
        # self.update(table, {'web_change': f'>={two_days}', 'check_date': f'<={one_day}', 'priority': None}, {'priority': 48}, schema=schema)

        print(" + closed blockades update")
        self.update(table, {'blockade_id': 2, 'acc_ban_id': 2, 'priority': None}, {'priority': -100}, schema=schema)

        print(" + last 7 days companies")
        # self.update(table, {'web_change': f'>={seven_days}', 'priority': None}, {'priority': 47}, schema=schema)

        print(" + older blockades segmentation")
        # next priority goes to companies changes that occurred in last 7 days except this previous
        self.update(table, {'update_date': f'>={seven_days}', 'blockade_id': 1, 'priority': None}, {'priority': 46}, schema=schema)

        # next priority goes to companies changes that occurred in last 30 days except this previous
        self.update(table, {'update_date': f'>={thirty_days}', 'blockade_id': 1, 'priority': None}, {'priority': 45}, schema=schema)

        # next priority goes to companies changes that occurred in last 60 days except this previous
        self.update(table, {'update_date': f'>={sixty_days}', 'blockade_id': 1, 'priority': None}, {'priority': 44}, schema=schema)

        # next priority goes to companies changes that occurred in last 60 days except this previous
        self.update(table, {'update_date': f'<={over_sixty_days}', 'blockade_id': 1, 'priority': None}, {'priority': 43}, schema=schema)

        print("+ company account bans")
        self.update(table, {'priority': None}, {'priority': 42}, schema=schema)

        log_common.update({'status': 1, 'errors': 0, 'note': note})
        self.info_log(log_common)


def start(n_run):
    upd_id = n_run / 10
    api = NbsPrinudnaNaplata()
    api.postgresql_conn()
    api.dir_creation_check()

    if upd_id.is_integer() is True:
        current_hour = datetime.now().hour
        if 22 > current_hour > 19:
            # fetch changes from webservice
            api.webservice()

        # crawler priority setup
        api.priority_update()

    # crawl website
    api.crawler()
    # transfer data from pstg to mssql
    api.transfer_data()

    # final runtime log
    api.runtime_log.update({'status': api.runtime_status, 'errors': api.runtime_errors})
    api.info_log(api.runtime_log)
    api.disconnect()


if __name__ == '__main__':
    # start(10)
    run_counter = 10
    while True:
        p = Process(target=start, args=(run_counter,))
        p.start()
        p.join()
        gc.collect()
        run_counter += 1
