import os
import sys
import pickle
import getpass

from sqlalchemy import *
from subprocess import call
from sqlalchemy import event
from subprocess import Popen, PIPE
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base

if sys.version_info.major > 2:
    from core_framework.db_engine import DbEngine

schema = 'nbs_pn'

framework_folder = r'C:\Users\{}\Documents\crawler_framework'.format(getpass.getuser())
python_path = r'{}\py_config.pkl'.format(framework_folder)

BasePstg = declarative_base()

# creating schema if it doesn't exist
event.listen(BasePstg.metadata, 'before_create', DDL("CREATE SCHEMA IF NOT EXISTS {}".format(schema)))

# ======================================================================
#                               TABLES
# ======================================================================


class AccountChanges(BasePstg):
    """History of all gathered company bank account changes that came over NBS webservice"""
    __tablename__ = 'account_changes'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('acc_change_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        Integer, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # original mbr
    vat_number = Column(String(128), index=True)  # original pib
    bank_account = Column(String(128), index=True)
    bank_code = Column(Integer())
    account_number = Column(Numeric(18, 0))
    control_number = Column(Integer())
    company_name = Column(String(3072))
    address = Column(String(3072))
    city = Column(String(3072))
    municipality_code = Column(Integer())
    activity_code = Column(Integer())
    municipality_name = Column(String(3072))
    activity_name = Column(String(3072))
    bank_name = Column(String(3072))
    status_id = Column(Integer(), index=True)
    blockade_id = Column(Integer())  # dali polijeze blokadi racun 2= ne, 0 = da
    type_id = Column(Integer())
    company_type_id = Column(Integer())
    initialization_date = Column(DateTime)
    change_date = Column(DateTime)
    update_date = Column(DateTime)

    account_id = Column(String(64), index=True)  # hash (registry_number, vat_number, bank_account)
    company_id = Column(String(64), index=True)  # hash (registry_number, vat_number)

    transfer = Column(Integer(), index=True)  # if null data is not transferred  1 = had bank account status changes, 2 = had other changes (city, activity..) 3: new bank account
    transfer_sha = Column(String(64), index=True)  # hash (status_id, blockade_id, type_id, company_type_id)
    transfer_date = Column(DateTime)  # date and time when data was transferred

    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class AccountBlockade(BasePstg):
    """History of all company bank account changes with enforced collection that came over NBS webservice"""
    __tablename__ = 'account_blockade'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('acc_block_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        Integer, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # original mbr
    vat_number = Column(String(128))  # original pib
    bank_account = Column(String(128))
    update_date = Column(Date)  # date when this blockade occurred
    account_id = Column(String(64), index=True)  # hash (registry_number, vat_number, bank_account)
    company_id = Column(String(64), index=True)  # hash (registry_number, vat_number)
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class CompanyBlockade(BasePstg):
    """Unique list of companies that have enforced collection over NBS webservice"""
    __tablename__ = 'company_blockade'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('comp_block_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        Integer, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # original mbr
    vat_number = Column(String(128))  # original pib
    update_date = Column(Date)  # last date when enforced collection occurred

    night_check = Column(Integer())  # did we successfully crawled nbs_pn web during night time  1 = yes, null = not checked
    night_countdown = Column(Integer())  # countdown number of times this registry number has to be checked before night_check gets 1=yes
    daily_check = Column(Integer())  # did we successfully crawled nbs_pn web during daytime 1 = yes, null = not checked
    priority = Column(Integer())  # higher the number higher the priority for checking that registry number
    web_change = Column(DateTime)  # last date when some change happened

    blockade_id = Column(Integer())   # is he in enforced collection on webpage 1 = yes, 2 = no, null = not checked
    acc_ban_id = Column(Integer())  # does he have ban on disposing of bank accounts 1 = yes, 2 = no, null = not checked
    blockade_amount = Column(Numeric(18, 2))  # if there is amount of blockade

    check_date = Column(DateTime)  # last date when it was successfully crawled on web

    company_id = Column(String(64), index=True)  # hash (registry_number, vat_number)
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class CompanyChanges(BasePstg):
    """Unique list of companies that had any kind of change on bank accounts over NBS webservice it serves for gathering data from other sources"""
    __tablename__ = 'company_changes'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('comp_chang_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        Integer, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row
    registry_number = Column(String(128), index=True)  # original mbr
    vat_number = Column(String(128))  # original pib
    update_date = Column(Date)  # last date of change
    pu_check = Column(Integer())  # did we successfully crawled poreska uprava web  1 = yes, null = not checked
    pu_date = Column(DateTime)  # last date when it was successfully crawled on web
    rir_check = Column(Integer())  # did we successfully gathered all bank accounts over NBS webservice
    rir_date = Column(DateTime)  # last date when it was successfully checked over NBS ws
    company_id = Column(String(64), index=True)  # hash (registry_number, vat_number)
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class BlockadeCompany(BasePstg):
    """History of all company basic info gathered from webpage"""
    __tablename__ = 'blockade_company'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('block_comp_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        Integer, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row without column company_id, archive and date_created
    registry_number = Column(String(128), index=True)  # original mbr
    vat_number = Column(String(128))  # original pib
    debtor_name = Column(String(4000))
    address = Column(String(4000))
    place = Column(String(4000))
    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class BlockadeStatus(BasePstg):
    """History of all company basic info gathered from webpage"""
    __tablename__ = 'blockade_status'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('block_stat_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        Integer, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row without column company_id, archive and date_created
    registry_number = Column(String(128), index=True)  # original mbr
    company_status = Column(Integer)  # status of company 102 and 103 no blockade 200 it has blockade data but not in blockade, 201 only ban, 202 only blockade, 203 both, 204 not in blockade but has amount
    transfer = Column(Integer, index=True)
    transfer_date = Column(DateTime)
    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class BlockadeDates(BasePstg):
    """History of all company blockade dates gathered from webpage"""
    __tablename__ = 'blockade_dates'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('block_dates_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        Integer, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row without column blockade_from and company_id
    blockade_id = Column(String(64), index=True)  # hash value of the
    registry_number = Column(String(128), index=True)  # original mbr
    blockade_from = Column(DateTime)  # date and time when blockade started
    blockade_to = Column(DateTime)  # date and time when blockade ended
    removed = Column(Integer)  # when company status is 102 or 103 records are removed from webpage, date when they re removed is equal to blockade_to date
    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class BlockadeBans(BasePstg):
    """History of all company ban dates of using bank accounts gathered from webpage"""
    __tablename__ = 'blockade_bans'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('block_banes_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        Integer, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row without column
    registry_number = Column(String(128), index=True)  # original mbr
    ban_date = Column(DateTime)  # date and time when company was banned
    removed = Column(Integer)  # when company status is 102 or 103 records are removed from webpage, date when they re removed is equal to archive date
    transfer = Column(Integer, index=True)
    transfer_date = Column(DateTime)
    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class BlockadeAmounts(BasePstg):
    """History of all company blockade amounts gathered from webpage"""
    __tablename__ = 'blockade_amounts'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('block_amount_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        Integer, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row without column
    registry_number = Column(String(128), index=True)  # original mbr
    blockade_amount = Column(Numeric(18, 2))  # amount of blockade presented in locale currency
    ban_id = Column(String(64), index=True)  # hash value connection on table blockade_bans column sha
    blockade_id = Column(String(64), index=True)  # hash value connection on table  blockade_dates column sha
    removed = Column(Integer)  # when company status is 102 or 103 records are removed from webpage, date when they re removed is equal to archive date
    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class BlockadeLatest(BasePstg):
    """History of all company blockade amounts gathered from webpage"""
    __tablename__ = 'blockade_latest'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('block_lat_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        Integer, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64), index=True)  # hash value of the row without column
    registry_number = Column(String(128), index=True)  # original mbr

    blockade_from = Column(DateTime)  # date and time when blockade started
    blockade_to = Column(DateTime)  # date and time when blockade ended
    blockade_amount = Column(Numeric(18, 2), server_default='0')  # amount of blockade presented in locale currency

    vat_number = Column(String(128))  # original pib
    debtor_name = Column(String(4000))
    address = Column(String(4000))
    place = Column(String(4000))

    transfer = Column(Integer(), index=True)  # if null data is not transferred  1 = had bank account status changes, 2 = had other changes (city, activity..) 3: new bank account
    transfer_sha = Column(String(64), index=True)  # hash (status_id, blockade_id, type_id, company_type_id)
    transfer_date = Column(DateTime)  # date and time when data was transferred

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


# ======================================================================
#                               LOGGERS
# ======================================================================

class CrawlerInfoLog(BasePstg):
    __tablename__ = 'l_info'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('loginfo_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        Integer, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    runtime_id = Column(String(64), index=True)  # hash value of single run cycle of program
    sha = Column(String(64), index=True)  # hash value of the row
    start_time = Column(DateTime)
    end_time = Column(DateTime)
    step_name = Column(String(256))
    status = Column(Integer())  # 1 = success  2=errors
    errors = Column(Integer())  # >0 there was and error
    note = Column(String(1024))
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class CrawlerErrorLog(BasePstg):
    __tablename__ = 'l_errors'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('logerr_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        Integer, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    runtime_id = Column(String(64), index=True)  # hash value of single run cycle of program
    sha = Column(String(64), index=True)  # hash value of the row
    start_time = Column(DateTime)
    end_time = Column(DateTime)
    step_name = Column(String(256))
    err_type = Column(String(3072))
    err_line = Column(Integer())
    err_desc = Column(String(3072))
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


def deploy(conn_id):
    db = DbEngine()
    engine = db.connect(conn_id, echo=False)
    BasePstg.metadata.drop_all(engine)
    BasePstg.metadata.create_all(engine)


if __name__ == '__main__':
    deploy(1)


