import re
import sys
import hashlib
import pandas as pd
from time import sleep
from html import unescape
from dateutil import parser
from bs4 import BeautifulSoup
from datetime import datetime


def number_of_changes(r):
    data = unescape(r.response.content_raw.decode())
    soup = BeautifulSoup(data, 'xml')

    # geting number of changes
    change_number = soup.findAll('GetCompanyAccountUpdatedCountResult')
    if change_number:
        change_number = change_number[0].text.strip()
        if change_number == "":
            change_number = 0
        else:
            change_number = int(change_number)
    else:
        return -1
    return change_number


def ws_page_results(r):
    try:
        bank_accounts = []
        data = unescape(r.response.content_raw.decode())
        soup = BeautifulSoup(data, 'xml')
        company_account_sets = soup.findAll('CompanyAccountDataSet')
        if not company_account_sets:
            return {'err_desc': 'could not find xml element CompanyAccountDataSet', 'status': 2}
        for company_account_set in company_account_sets:
            company_accounts = company_account_set.findAll('CompanyAccount')
            if not company_accounts:
                return {'err_desc': 'could not find xml element CompanyAccount', 'status': 2}
            for company_account in company_accounts:
                company_account_data = {}
                ws_columns = {'Account': 'bank_account', 'BankCode': 'bank_code', 'AccountNumber': 'account_number',
                              'ControlNumber': 'control_number', 'CompanyName': 'company_name',
                              'NationalIdentificationNumber': 'registry_number', 'TaxIdentificationNumber': 'vat_number',
                              'Address': 'address', 'City': 'city', 'MunicipalityCode': 'municipality_code',
                              'ActivityCode': 'activity_code', 'MunicipalityName': 'municipality_name',
                              'ActivityName': 'activity_name', 'BankName': 'bank_name',
                              'CompanyAccountStatusID': 'status_id', 'CompanyAccountBlockadeStatusID': 'blockade_id',
                              'CompanyAccountTypeID': 'type_id', 'LegalUserTypeID': 'company_type_id',
                              'InitializationDate': 'initialization_date', 'ChangeDate': 'change_date', 'UpdateDate': 'update_date'}

                date_columns = ['InitializationDate', 'ChangeDate', 'UpdateDate']

                for k, v in ws_columns.items():
                    column_data = company_account.find(k)
                    if column_data is not None:
                        column_data = column_data.text.strip()
                        if k in date_columns:
                            column_data = parser.parse(column_data, ignoretz=True, yearfirst=True)

                        # registry_number fix
                        if k == 'NationalIdentificationNumber':
                            diff = abs(column_data.__len__() - 8)
                            if diff is not 0:
                                column_data = "{0}{1}".format("0" * diff, column_data)

                        company_account_data.update({v: column_data})

                # bank account check and fix if needed
                account_data = company_account_data.get('bank_account')
                if len(account_data) != 20:
                    bank_code = company_account_data.get('bank_code')
                    account_number = company_account_data.get('account_number')
                    control_number = company_account_data.get('control_number')
                    account = f"{bank_code}-{account_number.zfill(13)}-{control_number.zfill(2)}"
                    company_account_data.update({'bank_account': account})

                sha = hashlib.sha3_256(str(company_account_data).encode()).hexdigest()
                company_account_data.update({'sha': sha})

                account_id_keys = ['registry_number', 'vat_number', 'bank_account']
                account_id = {x: company_account_data.get(x) for x in account_id_keys}
                account_sha = hashlib.sha3_256(str(account_id).encode()).hexdigest()
                company_account_data.update({'account_id': account_sha})

                company_id_keys = ['registry_number', 'vat_number']
                company_id = {x: company_account_data.get(x) for x in company_id_keys}
                company_sha = hashlib.sha3_256(str(company_id).encode()).hexdigest()
                company_account_data.update({'company_id': company_sha})

                transfer_keys = ['status_id', 'blockade_id', 'type_id', 'company_type_id']
                transfer_id = {x: company_account_data.get(x) for x in transfer_keys}
                transfer_sha = hashlib.sha3_256(str(transfer_id).encode()).hexdigest()
                company_account_data.update({'transfer_sha': transfer_sha})

                bank_accounts.append(company_account_data)
        return bank_accounts
    except Exception as e:
        exc_type, exc_obj, exc_tb = sys.exc_info()
        error = f"parser ws_page_results error {str(e)}"
        return {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': error, 'status': 2}


def status_extract(record, filter_id=1, blockade=True):
    # if blockade is True:
    #     status_id = record.get('status_id')
    #     blockade_id = record.get('blockade_id')
    #     if (status_id == '3' or status_id == '103') and blockade_id == '0':
    #         pass
    #     else:
    #         return

    keys_filters = {1: ['registry_number', 'vat_number', 'bank_account', 'update_date', 'account_id', 'company_id'],
                    2: ['registry_number', 'vat_number', 'update_date', 'company_id']}

    keys = keys_filters.get(filter_id)
    filtered_record = {x: record.get(x) for x in keys}
    update_date = filtered_record.get('update_date')
    if update_date is not None:
        update_date = datetime.strptime(update_date.strftime("%Y-%m-%d"), "%Y-%m-%d")
        filtered_record.update({'update_date': update_date})
    else:
        update_date = datetime.strptime(datetime.now().strftime("%Y-%m-%d"), "%Y-%m-%d")
        filtered_record.update({'update_date': update_date})
    sha = hashlib.sha3_256(str(filtered_record).encode()).hexdigest()
    filtered_record.update({'sha': sha})
    return filtered_record


def fix_letters(text):
    if type(text) is str:
        replacements = {'^': 'Č','[': 'Š', ']': 'Ć', '@': 'Ž', '!': 'Ć', "'": '"', '$': 'Č', '&#039;': '"', '&#034;': '"'}
        for k, v in replacements.items():
            text = text.replace(k, v)
    return text


def cirilic_to_latin(text):
    rep = {'А': 'A', 'Б': 'B', 'В': 'V', 'Г': 'G', 'Д': 'D', 'Ђ': 'Đ', 'Е': 'E', 'Ж': 'Ž', 'З': 'Z',
           'И': 'I', 'Ј': 'J', 'К': 'K', 'Л': 'L', 'Љ': 'Lj', 'М': 'M', 'Н': 'N', 'Њ': 'Nj', 'О': 'O',
           'П': 'P', 'Р': 'R', 'С': 'S', 'Т': 'T', 'Ћ': 'Ć', 'У': 'U', 'Ф': 'F', 'Х': 'H', 'Ц': 'C',
           'Ч': 'Č', 'Џ': 'Dž', 'Ш': 'Š', 'а': 'a', 'б': 'b', 'в': 'v', 'г': 'g', 'д': 'd', 'ђ': 'đ',
           'е': 'e', 'ж': 'ž', 'з': 'z', 'и': 'i', 'ј': 'j', 'к': 'k', 'л': 'l', 'љ': 'lj', 'м': 'm',
           'н': 'n', 'њ': 'nj', 'о': 'o', 'п': 'p', 'р': 'r', 'с': 's', 'т': 't', 'ћ': 'ć', 'у': 'u',
           'ф': 'f', 'х': 'h', 'ц': 'c', 'ч': 'č', 'џ': 'dž', 'ш': 'š', 'à': 'a', '&amp;': '&',
           'Ω': 'Omega', "'": '', 'º': '0', 'Å': 'A', 'å': 'a', 'A': 'A', 'È': 'E', 'Ì': 'I', 'Ò': 'O', 'Ù': 'U',
           'è': 'e', 'ì': 'i', 'Á': 'A', 'É': 'E', 'Í': 'I', 'Ó': 'O', 'Ú': 'U', 'Ý': 'Y', 'Â': 'A', 'Ê': 'E', 'Î': 'I', 'Ô': 'O',
           'Û': 'U', 'á': 'a', 'é': 'e', 'í': 'i', 'ó': 'o', 'ú': 'u', 'ý': 'y', 'â': 'a', 'ê': 'e', 'î': 'i', 'ô': 'o', 'û': 'u',
           'Ã': 'A', 'Ñ': 'N', 'Õ': 'O', 'ã': 'a', 'ñ': 'n', 'õ': 'o', 'Ä': 'A', 'Ë': 'E', 'Ï': 'I', 'Ö': 'O', 'Ü': 'U', 'Ÿ': 'Y',
           'ä': 'a', 'ë': 'e', 'ï': 'i', 'ö': 'o', 'ü': 'u', 'ÿ': 'y', 'Ç': 'C', 'ç': 'c', 'Þ': 'P', 'þ': 'p', 'Ø': 'O', 'ø': 'o',
           'ª': '', 'ß': 'ss', 'ẞ': 'SS', 'Œ': 'OE', 'œ': 'oe', 'Æ': 'Ć', 'æ': 'ć', 'ð': 'đ', 'і': 'i', 'S': 'S', '&quot;': '"',
           'ӧ': 'o', "ʽ": "", 'ќ': 'k', 'ѕ': 's', 'ĭ': 'l', 'й': 'j', 'я': 'a', 'ы': 'i', 'ΡΜ': 'PM', '″': '', 'Ð': 'Đ', 'Й': 'J',
           'Ӧ': 'O', 'đ': 'đ', '&#231;': 'c', 'ğ': 'g', '&#193;': 'A', '&#201;': 'E', '&#220;': 'U', '&#211;': 'O', 'Ő': 'O',
           'Ũ': 'U', 'β': 'SS', '׳': '', 'ž': 'ž', 'a': 'a', 'č': 'č', '&#208;': 'Đ', 'ÄŤ': 'č', 'İ': 'I', 'Ĺľ': 'ž', 'ș': 's', '΄': ''
           }
    rep = dict((re.escape(k), v) for k, v in rep.items())
    pattern = re.compile("|".join(rep.keys()))
    text = pattern.sub(lambda m: rep[re.escape(m.group(0))], text)
    return text


def web_cleanup(data, db_reg_num):
    data_clean = {}
    for k, v in data.items():
        v = v.strip()
        if k == 'blockade_amount':
            v = ''.join(v.replace('-', '').strip().split('.')).replace(',', '.')
        if k == 'vat_number':
            v = v.strip()
            if v is not None:
                if len(v) > 12:
                    v = None
            if v is not None:
                if len(v) < 9:
                    v = None
        if k == 'registry_number':
            if type(v) is str:
                if len(v) < 8:
                    v = v.zfill(8)
            if v != db_reg_num:
                print("wrong registry number returned")
                return {'err_desc': 'Service returned data for wrong registry number'}, -1
        if k == 'ban_date':
            v = re.search('(\d{2}\.)\s*(\d{2}\.)\s*(\d{4}\.)', v, re.MULTILINE | re.IGNORECASE | re.DOTALL)
            if v is not None:
                v = ''.join(v.groups())
                v = pd.to_datetime(v, dayfirst=True, errors='coerce')
        if k == 'blockade_amount' and v == '':
            v = None
        data_clean.update({k:v})
    return data_clean, 200


class ParserData:
    def __init__(self):
        self.company_id = None
        self.company_status = 200
        self.dates_data = dict()
        self.history_dates_data = dict()
        self.base_data = dict()
        self.ban_data = dict()
        self.blockade_amount = dict()
        self.company_data = dict()
        self.blockade_stats = dict()
        self.blockade_id = None

    def get_sha(self, data):
        return hashlib.sha3_256(str(data).encode()).hexdigest()

    def compose(self, keys, data):
        ready_data = {k: fix_letters(v) for k,v in data.items() if k in keys}
        sha = self.get_sha(ready_data)
        ready_data.update({'sha': sha})
        return ready_data

    def extract_company(self, data):
        debtor_name = data.get('debtor_name')
        keys = ['registry_number', 'vat_number', 'debtor_name', 'address', 'place']
        if debtor_name is not None:
            self.company_data = self.compose(keys, data)

    def extract_ban_data(self, data):
        ban_date = data.get('ban_date')
        keys = ['registry_number', 'ban_date']
        if ban_date is not None:
            self.ban_data = self.compose(keys, data)

    def extract_blockade(self, data):
        blockade_amount = data.get('blockade_amount')
        keys = ['registry_number', 'blockade_amount']
        if blockade_amount is not None:
            self.blockade_amount = self.compose(keys, data)
            ban_id = self.ban_data.get('sha')
            self.blockade_amount.update({'ban_id': ban_id})

    def prepare_blockade_stats(self, registry_number):
        self.blockade_stats.update({'registry_number': registry_number, 'company_status': self.company_status})
        sha = self.get_sha(self.blockade_stats)
        self.blockade_stats.update({'sha': sha})

    def split_data(self, data):
        self.extract_company(data)
        self.extract_ban_data(data)
        self.extract_blockade(data)


def web_parser(html, db_reg_num):
    try:
        data = {}
        parser_data = ParserData()

        no_blockade = {'Du.nik nema evidentirane neizmirene obaveze': 102, 'Tra.eni du.nik nije evidentiran u prinudnoj naplati': 103}
        html = cirilic_to_latin(html.decode())

        for search_text, status_code in no_blockade.items():
            re_result = re.search(search_text, html, re.MULTILINE | re.IGNORECASE | re.DOTALL)
            if re_result is not None:
                parser_data.company_status = status_code

        duznik_rex = f'''<td.*?>Dužnik:.*?</td>.*?<td.*?"><strong>(?P<debtor_name>.*?)</strong></td>.*?</tr>'''
        x = re.search(duznik_rex, html, re.MULTILINE | re.IGNORECASE | re.DOTALL)
        if x is not None:
            data.update(x.groupdict())

        rex_keys = {'Adresa:': 'address', 'Mesto:': 'place', 'Ukupan iznos blokade:': 'blockade_amount',
                    'Poreski broj:': 'vat_number', 'Zabrana prenosa:.*?': 'ban_date'}

        for k, v in rex_keys.items():
            rex = f'''<td.*?>{k}</td>.*?<td.*?>(?P<{v}>.*?)</td>'''
            x = re.search(rex, html, re.MULTILINE | re.IGNORECASE | re.DOTALL)
            if x is not None:
                data.update(x.groupdict())

        registry_number_rex = '''<input type="hidden" name="matbr".*?value="(?P<registry_number>.*?)" />'''
        x = re.search(registry_number_rex, html, re.MULTILINE | re.IGNORECASE | re.DOTALL)
        if x is not None:
            data.update(x.groupdict())
        if x is None:
            registry_number_rex = '''<tr><td>Matični broj:</td><td>(?P<registry_number>.*?)</td></tr>'''
            x = re.search(registry_number_rex, html, re.MULTILINE | re.IGNORECASE | re.DOTALL)
            if x is not None:
                data.update(x.groupdict())
            else:
                sleep(20)
                return {'err_desc': f'Could not find registry number {db_reg_num}'}

        # cleaning, formatting and separating data
        clean_data, parser_status = web_cleanup(data, db_reg_num)
        if parser_status == -1:
            return clean_data

        parser_data.split_data(clean_data)

        blockade_stat = False
        blockade_days_s = re.search('Dani blokade(.*?Ukupno:.*?</td>.*?<td>.*?</td>)', html, re.MULTILINE | re.DOTALL)
        if blockade_days_s is not None:
            blockade_days = re.findall(r'''<tr>.*?<td>.*?(\d+?\.\d+?\.\d+?\.).*?</td>.*?<td>(.*?)</td>.*?<td>\s*?(\d*?)\s*?</td>.*?</tr>''', blockade_days_s.group(1), re.MULTILINE | re.DOTALL)
            for blockade_day in blockade_days:
                blockade_day = (x.replace('-', '').strip() for x in blockade_day)
                blockade_from, blockade_to, blockade_days = blockade_day
                blockade_from = pd.to_datetime(blockade_from, dayfirst=True, errors='coerce')
                blockade_to = pd.to_datetime(blockade_to, dayfirst=True, errors='coerce')

                blockade_data = {'registry_number': db_reg_num, 'blockade_to': blockade_to}
                sha = parser_data.get_sha(blockade_data)

                # blockade_open = {'registry_number': db_reg_num, 'blockade_to': pd.NaT}
                # blockade_open_sha = parser_data.get_sha(blockade_open)
                if pd.isna(blockade_to) is True:
                    blockade_stat = True
                    copy_blockade_data = blockade_data.copy()
                    copy_blockade_data.update({'blockade_from': blockade_from})
                    blockade_id = parser_data.get_sha(copy_blockade_data)
                    copy_blockade_data.update({'sha': sha, 'blockade_id': blockade_id})
                    copy_blockade_data.update({'blockade_to': None})
                    parser_data.dates_data.update(copy_blockade_data)
                    parser_data.blockade_amount.update({'blockade_id': blockade_id})
                else:
                    blockade_data.update({'sha': sha, 'blockade_from': blockade_from, 'blockade_id': None})
                    parser_data.history_dates_data.update({len(parser_data.history_dates_data): blockade_data})

        # define type of blockade data
        if parser_data.company_status not in [102,103]:
            if blockade_stat is False and parser_data.ban_data != {}:
                parser_data.company_status = 201
            if blockade_stat is True and parser_data.ban_data == {}:
                parser_data.company_status = 202
            if blockade_stat is True and parser_data.ban_data != {}:
                parser_data.company_status = 203

        # update hash value for blockade_amount
        # sha = parser_data.get_sha(parser_data.blockade_amount)
        # parser_data.blockade_amount.update({'sha': sha})

        parser_data.prepare_blockade_stats(db_reg_num)
        return parser_data
    except Exception as e:
        exc_type, exc_obj, exc_tb = sys.exc_info()
        return {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)}

# with open(r'C:\Users\draganm\Documents\nbs_pn\20492155\20191112\0849.html', 'rb') as f:
#     html = f.read()
#     print(html)
#     data = web_parser(html, '20492155')
#     print(data.company_status)
