import re
import math
import hashlib
import pandas as pd
from sys import stdout
import concurrent.futures
from core_framework.db_engine import DbEngine


class DbConnection(DbEngine):
    def __init__(self):
        DbEngine.primary_key = 'id'
        DbEngine.__init__(self)
        self.connect(2)


class PostgresConnection(DbEngine):
    def __init__(self):
        DbEngine.primary_key = 'id'
        DbEngine.__init__(self)
        self.connect(1)


class DbAprConnection(DbEngine):
    def __init__(self):
        DbEngine.primary_key = 'id'
        DbEngine.__init__(self)
        self.connect(3)


def unique_status(status_eval):
    if status_eval is None or status_eval == '':
        return 'Nepoznato'
    rex = {'aktiv': 'Aktivan', 'bris': 'Brisan', 'prin.*?likvi': 'Prinudna likvidacija',
           'proc.*?likvi': 'U procesu likvidacije', 'likvi': 'U likvidaciji',
           'ste.*?post': 'U stečajnom postupku', 'ste.a': 'U stečaju'}
    for k, v in rex.items():
        if re.search(k, status_eval, re.IGNORECASE) is not None:
            status_eval = v
            break
    return status_eval


def row_sha(data, cols=None, dates=None):
    """
    :type data: dict
    :type cols: list
    :type dates: list
    """
    copy_data = data.copy()
    # #date conversion
    # if dates is not None:
    #     for k, v in copy_data.items():
    #         if k in dates:
    #             print(k, v, type(v))
    #             if type(v) is not str and v is not None:
    #                 copy_data.update({k: v})
    #                 exit()

    if cols is not None:
        copy_data = {k: str(v).lower() for k,v in copy_data.items() if k in cols and v is not None}

    values = ''.join(list(copy_data.values()))
    values = re.sub(r'[\s+]','',values, re.DOTALL|re.UNICODE)
    # values = ''.join(sorted(values))
    sha = hashlib.sha3_256(values.encode()).hexdigest()
    return sha


db = DbConnection()
db_apr = DbAprConnection()
pstgrs = PostgresConnection()


class SearchInfoFill:
    def __init__(self):
        # SHA columns
        self.search_info = ['registry_number', 'ctype', 'cname', 'cstatus']

    def update_sha_apr_subjekti(self):
        # populate search_info_sha that will be same as sha in [APR_SRBIJA].[dbo].[SearchInfo]
        # UPDATE SHA APR_SUBJEKTI
        print('updating search_info_sha in table APR_SUBJEKTI')
        sql = '''select distinct id , MB registry_number, PRAVNA_FORMA ctype, case when NAZIV is null then NAZIV_KRATKI else NAZIV end cname, STATUS cstatus, DAT_PRO archive, DATE_CREATED date_created, DAT_CHECK  date_checked
        from Bisnode_RS.dbo.APR_SUBJEKTI where search_info_sha is null order by id'''
        data = db.select('APR_SUBJEKTI', sql=sql)

        # # for first update we created tmp table with cols id and sha populate that data and
        # # used that data to update search_info_sha in table APR_SUBJEKTI
        # # query for update is
        # sha_data_id = {}
        # for row in data:
        #     row_sha_ready = row_sha(row, self.search_info)
        #     id = row.get('id')
        #     sha_data_id.update({len(sha_data_id): {'id': id, 'sha': row_sha_ready}})
        # merge_status = db.merge('tmp_search_info_sha_ids', sha_data_id,filters={}, update=False)
        # print(merge_status)

        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
            # Start the load operations and mark each future with its URL
            future_to_url = {executor.submit(self.update_sha_apr_subjekti_thread, row): row for row in data}
            for future in concurrent.futures.as_completed(future_to_url):
                url = future_to_url[future]
                try:
                    status = future.result()
                except AttributeError as e:
                    print(str(e))
                except Exception as e:
                    print(str(e))

    def update_sha_apr_subjekti_thread(self, row):
        row_sha_ready = row_sha(row, self.search_info)
        id = row.get('id')
        x = db.update('APR_SUBJEKTI', {'ID': id, 'search_info_sha': None}, {'search_info_sha': row_sha_ready}, freeze=True, primary_key='ID')
        # print('Update status', x, row)
        return 200

    def update_search_sha_ms(self):
        print('Update sha in SearchInfo that is in APR_SRBIJA db')
        sql = '''select id, RegistryNumber registry_number, [TYPE] ctype, [Name] cname, [Status] cstatus from SearchInfo where sha is null'''
        data = db_apr.select('SearchInfo', sql=sql)

        with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
            # Start the load operations and mark each future with its URL
            future_to_url = {executor.submit(self.update_search_sha_ms_thread, row): row for row in data}
            for future in concurrent.futures.as_completed(future_to_url):
                url = future_to_url[future]
                try:
                    status = future.result()
                except AttributeError as e:
                    pass
                except Exception as e:
                    print(str(e))

    def update_search_sha_ms_thread(self, row):
        row_sha_ready = row_sha(row, self.search_info)
        id = row.get('id')
        x = db_apr.update('SearchInfo', {'id': id, 'sha': None}, {'sha': row_sha_ready}, freeze=True)
        print('Update status', x, row)
        row.update({'sha': row_sha_ready})

    def archive_data_search_info(self):
        print('updating archive dates')
        # update rows that are older than last record for single registry number
        sql = '''select id, RegistryNumber registry_number, [TYPE] ctype, [Name] cname, [Status] cstatus, DateCreated date_created
         from SearchInfo a where a.Archive is null 
         and a.DateCreated >= getdate()-1 
         order by RegistryNumber, DateCreated desc'''
        data = db_apr.select('SearchInfo', sql=sql)
        sorted_data = {}
        for row in data:
            registry_number = row.get('registry_number')
            date_created = row.get('date_created')

            get_reg_data = sorted_data.get(registry_number)
            if get_reg_data is None:
                sorted_data.update({registry_number: {0: row}})
            else:
                get_reg_data.update({len(get_reg_data): row})
        cnt = 0
        for k, v in sorted_data.items():
            total_rows = len(v)
            if total_rows > 1:
                for i in range(1, total_rows):
                    previous_row_data = v.get(i-1)
                    row_data = v.get(i)
                    row_id = row_data.get('id')
                    archive_date = previous_row_data.get('date_created')
                    upd_stat = db_apr.update('SearchInfo', {'id': row_id}, {'Archive': archive_date}, freeze=True)
                    print(cnt, upd_stat, row_data)
            cnt += 1

    def first_transfer_apr_subjekti(self):
        # INSERT DATA TO POSTGRES SearchInfo FROM APR_SUBJEKTI
        print('insert into postgres search_info FROM APR_SUBJEKTI')
        check_data = pstgrs.select('search_info', schema='apr')
        if not check_data:
            sql = '''select * from v_toSearchInfo a
                     where not exists (select 1 from APR_SRBIJA.dbo.SearchInfo b where a.sha = b.sha and a.registry_number = b.RegistryNumber)
                     order by date_created'''
            data = db.select('APR_SUBJEKTI', sql=sql)
            for row in data:
                get_status = row.get('cstatus')
                row_sha_ready = row_sha(row, self.search_info)
                row.update({'status_group': unique_status(get_status), 'sha': row_sha_ready})

            merge_data = dict(enumerate(data))
            pstgrs.merge('search_info', merge_data, filters={}, schema='apr')

    def mix_new_data(self):
        "combine new data with old data"
        print('insert into postgres search_info new data from SearchInfo')
        # regnums_sync = [x.get('registrynumber') for x in db_apr.select('SearchInfo', sql="select distinct RegistryNumber from SearchInfo where sync is null and RegistryNumber='06000037' ")]
        regnums_sync = [x.get('registrynumber') for x in db_apr.select('SearchInfo', sql="select distinct RegistryNumber from SearchInfo where sync is null ")]
        # compare only all records SearchInfo mssql with all records in search_info postgres
        print("compare only all records SearchInfo mssql with all records in search_info postgres")
        cnt = 0
        for regnum in regnums_sync:
            cnt += 1
            stdout.write(f'\r>progress: {cnt}/{len(regnums_sync)}')
            sql = f'''select RegistryNumber registry_number, [TYPE] ctype, [Name] cname, [Status] cstatus, Registry registry, DateCreated date_created, Archive archive, sha , DateChecked date_checked  
            from SearchInfo where RegistryNumber = '{regnum}' '''
            regnum_data = db_apr.select('SearchInfo', sql=sql)
            for regnum_d in regnum_data:
                get_status = regnum_d.get('cstatus')
                regnum_d.update({'status_group': unique_status(get_status)})

            regnum_data = dict(enumerate(regnum_data))
            merge_status = pstgrs.merge('search_info',regnum_data, filters={'registry_number': regnum}, schema='apr', on=['sha'], update=False)

        # get all registry numbers that have more than one record open in postgres now and add archive date
        print("\nget all registry numbers that have more than one record open in postgres now and add archive date")
        sql = '''select registry_number from (
        select registry_number, count(*) from apr.search_info where archive is null group by registry_number having count(*) > 1) a'''
        regnums_archive = [x.get('registry_number') for x in pstgrs.select('search_info', sql=sql, schema='apr')]
        cnt = 0
        for regnum in regnums_archive:
            cnt += 1
            stdout.write(f'\r>progress: {cnt}/{len(regnums_archive)}')
            sql = f"select * from apr.search_info where registry_number= '{regnum}' order by date_created desc"
            # print(sql)
            regnum_data = pstgrs.select('search_info', sql=sql, schema='apr')
            sorted_data = {}
            for row in regnum_data:
                registry_number = row.get('registry_number')
                date_created = row.get('date_created')
                get_reg_data = sorted_data.get(registry_number)
                if get_reg_data is None:
                    sorted_data.update({registry_number: {0: row}})
                else:
                    get_reg_data.update({len(get_reg_data): row})

            for k, v in sorted_data.items():
                total_rows = len(v)
                if total_rows > 1:
                    for i in range(1, total_rows):
                        previous_row_data = v.get(i - 1)
                        row_data = v.get(i)
                        row_id = row_data.get('id')
                        archive_date = previous_row_data.get('date_created')
                        # print(row_id, archive_date)
                        upd_stat = pstgrs.update('search_info', {'id': row_id}, {'archive': archive_date}, schema='apr')
                        # print(cnt, upd_stat, row_data)

        # compare only active records SearchInfo mssql with active records in search_info postgres
        print('\ncompare only active records SearchInfo mssql with active records in search_info postgres')
        cnt = 0
        for regnum in regnums_sync:
            cnt += 1
            stdout.write(f'\r>progress: {cnt}/{len(regnums_sync)}')
            sql = f'''select RegistryNumber registry_number, [TYPE] ctype, [Name] cname, [Status] cstatus, Registry registry,  DateCreated date_created, Archive archive, sha, DateChecked date_checked   
            from SearchInfo where RegistryNumber = '{regnum}' and archive is null'''
            regnum_data = db_apr.select('SearchInfo', sql=sql)
            for regnum_d in regnum_data:
                get_status = regnum_d.get('cstatus')
                regnum_d.update({'status_group': unique_status(get_status)})

            regnum_data = dict(enumerate(regnum_data))
            merge_status = pstgrs.merge('search_info',regnum_data, filters={'registry_number': regnum, 'archive': None}, schema='apr', on=['sha'])
            if merge_status == 200:
                db_apr.update('SearchInfo', {'RegistryNumber': regnum, 'sync': None}, {'sync': 1}, freeze=True)


class BasicInfoFill:
    def __init__(self):
        # SHA columns
        self.basic_info = ['registry_number', 'basic_name', 'cname', 'cstatus', 'status_group', 'legal_form', 'foun_date', 'erase_date', 'activity_start', 'activity_end']
        self.cnt, self.total = 0, 0

    def update_sha_apr_subjekti(self):
        # populate search_info_sha that will be same as sha in [APR_SRBIJA].[dbo].[SearchInfo]
        # UPDATE SHA APR_SUBJEKTI
        print('updating basic_info_sha in table APR_SUBJEKTI')
        sql = '''select distinct id , MB registry_number, NAZIV_KRATKI basic_name, NAZIV cname, STATUS cstatus, PRAVNA_FORMA legal_form,
        DAT_OSN foun_date, DAT_BRISANJA erase_date, DAT_PRESTANKA_DJEL activity_end
        from Bisnode_RS.dbo.APR_SUBJEKTI where basic_info_sha is null order by id'''
        data = db.select('APR_SUBJEKTI', sql=sql)

        # for first update we created tmp table with cols id and sha populate that data and
        # used that data to update search_info_sha in table APR_SUBJEKTI
        # query for update is
        sha_data_id = {}
        print(f"total size: {len(data)}")
        self.cnt = 0
        self.total = len(data)

        # for row in data:
        #     stdout.write(f"\rprocessing: {self.cnt}/{self.total}")
        #     row_sha_ready = row_sha(row, self.basic_info)
        #     id = row.get('id')
        #     sha_data_id.update({len(sha_data_id): {'id': id, 'sha': row_sha_ready}})
        #     self.cnt += 1
        # print("\ninserting data")
        # merge_status = db.insert('tmp_basic_info_sha_ids', sha_data_id)
        # print(merge_status)

        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
            # Start the load operations and mark each future with its URL
            future_to_url = {executor.submit(self.update_sha_apr_subjekti_thread, row): row for row in data}
            for future in concurrent.futures.as_completed(future_to_url):
                url = future_to_url[future]
                try:
                    status = future.result()
                except AttributeError as e:
                    print(str(e))
                except Exception as e:
                    print(str(e))

    def update_sha_apr_subjekti_thread(self, row):
        self.cnt += 1
        stdout.write(f"\r progress: {self.cnt}/{self.total}")
        row_sha_ready = row_sha(row, self.basic_info)
        id = row.get('id')
        x = db.update('APR_SUBJEKTI', {'ID': id, 'basic_info_sha': None}, {'basic_info_sha': row_sha_ready}, freeze=True, primary_key='ID')
        # print('Update status', x, row)
        return 200

    def update_basic_sha_ms(self):
        print('update sha in BasicInfo that is in APR_SRBIJA db')
        sql = '''select id, RegistryNumber registry_number, BasicName basic_name, CompanyName cname, [Status] cstatus, 
        LegalForm legal_form, FounDate foun_date, EraseDate erase_date, ActivityStartDate activity_start, ActivityEndDate activity_end 
        from BasicInfo where sha is null'''
        data = db_apr.select('BasicInfo', sql=sql)

        self.cnt = 0
        self.total = len(data)

        with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
            # Start the load operations and mark each future with its URL
            future_to_url = {executor.submit(self.update_search_sha_ms_thread, row): row for row in data}
            for future in concurrent.futures.as_completed(future_to_url):
                url = future_to_url[future]
                try:
                    status = future.result()
                except AttributeError as e:
                    pass
                except Exception as e:
                    print(str(e))

    def update_search_sha_ms_thread(self, row):
        row_sha_ready = row_sha(row, self.basic_info)
        id = row.get('id')
        x = db_apr.update('BasicInfo', {'id': id, 'sha': None}, {'sha': row_sha_ready}, freeze=True)
        self.cnt += 1
        stdout.write(f"\r>progress: {self.cnt}/{self.total}")
        row.update({'sha': row_sha_ready})

    def archive_data_basic_info(self):
        print('\nupdating archive dates')
        # update rows that are older than last record for single registry number

        # check last 2 days
        sql = '''select id, RegistryNumber registry_number, BasicName basic_name, CompanyName cname, [Status] cstatus,
        LegalForm legal_form, FounDate foun_date, EraseDate erase_date, ActivityStartDate activity_start,
        ActivityEndDate activity_end, DateCreated date_created from BasicInfo a where a.Archive is null
        and a.DateCreated >= getdate()-1 order by RegistryNumber, DateCreated desc'''

        # # check all
        # sql = '''select id, RegistryNumber registry_number, BasicName basic_name, CompanyName cname, [Status] cstatus,
        # LegalForm legal_form, FounDate foun_date, EraseDate erase_date, ActivityStartDate activity_start,
        # ActivityEndDate activity_end, DateCreated date_created from BasicInfo a where a.Archive is null
        # order by RegistryNumber, DateCreated desc'''

        data = db_apr.select('BasicInfo', sql=sql)
        sorted_data = {}
        for row in data:
            registry_number = row.get('registry_number')
            date_created = row.get('date_created')

            get_reg_data = sorted_data.get(registry_number)
            if get_reg_data is None:
                sorted_data.update({registry_number: {0: row}})
            else:
                get_reg_data.update({len(get_reg_data): row})

        cnt = 0
        for k, v in sorted_data.items():
            cnt += 1
            stdout.write(f'\r> progress: {cnt}/{len(sorted_data)}')
            total_rows = len(v)
            if total_rows > 1:
                for i in range(1, total_rows):
                    previous_row_data = v.get(i-1)
                    row_data = v.get(i)
                    row_id = row_data.get('id')
                    archive_date = previous_row_data.get('date_created')
                    upd_stat = db_apr.update('BasicInfo', {'id': row_id}, {'archive': archive_date}, freeze=True)
                    print("\n", cnt, upd_stat, row_data)

    def first_transfer_apr_subjekti(self):
        # INSERT DATA TO POSTGRES BasicInfo FROM APR_SUBJEKTI
        print('insert into postgres basic_info FROM APR_SUBJEKTI')
        check_data = pstgrs.select('basic_info', schema='apr')
        if not check_data:
            sql = '''select * from v_toBasicInfo a
                     where not exists (select 1 from APR_SRBIJA.dbo.BasicInfo b where a.sha = b.sha and a.registry_number = b.RegistryNumber)
                     order by date_created'''
            data = db.select('APR_SUBJEKTI', sql=sql)
            for row in data:
                get_status = row.get('cstatus')
                row_sha_ready = row_sha(row, self.basic_info)
                row.update({'status_group': unique_status(get_status), 'sha': row_sha_ready})

            merge_data = dict(enumerate(data))
            pstgrs.merge('basic_info', merge_data, filters={}, schema='apr')

    def mix_new_data(self):
        "combine new data with old data"
        print('insert into postgres basic_info new data from BasicInfo')
        # regnums_sync = [x.get('registrynumber') for x in db_apr.select('BasicInfo', sql="select distinct RegistryNumber from BasicInfo where sync is null and RegistryNumber='06000037' ")]
        regnums_sync = [x.get('registrynumber') for x in db_apr.select('BasicInfo', sql="select distinct RegistryNumber from BasicInfo where sync is null ")]
        # compare only all records BasicInfo mssql with all records in basic_info postgres
        print("compare only all records BasicInfo mssql with all records in basic_info postgres")
        cnt = 0
        for regnum in regnums_sync:
            cnt += 1
            stdout.write(f'\r>progress: {cnt}/{len(regnums_sync)}')

            sql = f'''
            select RegistryNumber registry_number, BasicName basic_name, CompanyName cname, [Status] cstatus,
            LegalForm legal_form, FounDate foun_date, EraseDate erase_date, ActivityStartDate activity_start,
            ActivityEndDate activity_end, DateCreated date_created, Archive archive, sha  from BasicInfo a 
            where RegistryNumber = '{regnum}'
            '''
            regnum_data = db_apr.select('BasicInfo', sql=sql)
            for regnum_d in regnum_data:
                get_status = regnum_d.get('cstatus')
                regnum_d.update({'status_group': unique_status(get_status)})

            regnum_data = dict(enumerate(regnum_data))
            merge_status = pstgrs.merge('basic_info',regnum_data, filters={'registry_number': regnum}, schema='apr', on=['sha'], update=False)

        # get all registry numbers that have more than one record open in postgres now and add archive date
        print("\nget all registry numbers that have more than one record open in postgres now and add archive date")
        sql = '''select registry_number from (
        select registry_number, count(*) from apr.basic_info where archive is null group by registry_number having count(*) > 1) a'''
        regnums_archive = [x.get('registry_number') for x in pstgrs.select('basic_info', sql=sql, schema='apr')]
        cnt = 0
        for regnum in regnums_archive:
            cnt += 1
            stdout.write(f'\r>progress: {cnt}/{len(regnums_archive)}')
            sql = f"select * from apr.basic_info where registry_number= '{regnum}' order by date_created desc"
            # print(sql)
            regnum_data = pstgrs.select('basic_info', sql=sql, schema='apr')
            sorted_data = {}
            for row in regnum_data:
                registry_number = row.get('registry_number')
                date_created = row.get('date_created')
                get_reg_data = sorted_data.get(registry_number)
                if get_reg_data is None:
                    sorted_data.update({registry_number: {0: row}})
                else:
                    get_reg_data.update({len(get_reg_data): row})

            for k, v in sorted_data.items():
                total_rows = len(v)
                if total_rows > 1:
                    for i in range(1, total_rows):
                        previous_row_data = v.get(i - 1)
                        row_data = v.get(i)
                        row_id = row_data.get('id')
                        archive_date = previous_row_data.get('date_created')
                        # print(row_id, archive_date)
                        upd_stat = pstgrs.update('basic_info', {'id': row_id}, {'archive': archive_date}, schema='apr')
                        # print(cnt, upd_stat, row_data)

        # compare only active records SearchInfo mssql with active records in basic_info postgres
        print('\ncompare only active records SearchInfo mssql with active records in basic_info postgres')
        cnt = 0
        for regnum in regnums_sync:
            cnt += 1
            stdout.write(f'\r>progress: {cnt}/{len(regnums_sync)}')
            sql = f'''
            select RegistryNumber registry_number, BasicName basic_name, CompanyName cname, [Status] cstatus,
            LegalForm legal_form, FounDate foun_date, EraseDate erase_date, ActivityStartDate activity_start,
            ActivityEndDate activity_end, DateCreated date_created, Archive archive, sha  from BasicInfo a 
            where RegistryNumber = '{regnum}' and archive is null
            '''
            regnum_data = db_apr.select('BasicInfo', sql=sql)
            for regnum_d in regnum_data:
                get_status = regnum_d.get('cstatus')
                regnum_d.update({'status_group': unique_status(get_status)})

            regnum_data = dict(enumerate(regnum_data))
            merge_status = pstgrs.merge('basic_info',regnum_data, filters={'registry_number': regnum, 'archive': None}, schema='apr', on=['sha'])
            if merge_status == 200:
                db_apr.update('BasicInfo', {'RegistryNumber': regnum, 'sync': None}, {'sync': 1}, freeze=True)


class ActivityPause:
    def __init__(self):
        # SHA columns
        self.activity_cols = ['registry_number', 'pause_start', 'pause_end']
        self.cnt, self.total = 0, 0

    def add_history_records(self):
        print("Adding records from Bisnode_RS.dbo.APR_SUBJEKTI  that doesnt exist in ActivityPause")
        sql = '''
        select distinct MB RegistryNumber, DAT_OD_PP PauseStartDate, DAT_DO_PP PauseEndDate, DATE_CREATED DateCreated from Bisnode_RS.dbo.APR_SUBJEKTI a
        where (DAT_OD_PP is not null or DAT_DO_PP is not null)
        and not exists (select 1 from ActivityPause b where a.mb = b.RegistryNumber 
        and isnull(a.DAT_OD_PP, '1900-01-01') = isnull(b.PauseStartDate, '1900-01-01') 
        and isnull(a.DAT_DO_PP, '1900-01-01') = isnull(b.PauseEndDate, '1900-01-01'))
        '''
        data = db_apr.select('APR_SUBJEKTI', sql=sql, freeze=True)
        if data:
            ins_data = dict(enumerate(data))
            print("inserting")
            insert_status = db_apr.insert('ActivityPause', ins_data)
            print(f"insert status: {insert_status}")

    def update_sha_activity(self):
        print("Adding hash value in ActivityPause")
        sql = '''select id, RegistryNumber registry_number,  PauseStartDate pause_start, PauseEndDate pause_end from ActivityPause where sha is null'''
        data = db_apr.select('ActivityPause', sql=sql)
        for row in data:
            row_id = row.get('id')
            row_sha_ready = row_sha(row, self.activity_cols)
            update_status = db_apr.update('ActivityPause', {'id': row_id}, {'sha': row_sha_ready})
            if update_status != 200:
                print(f"error occured: {update_status}")
                exit()

    def calculate_endate(self):
        # sql = '''select * from APR_SRBIJA.dbo.ActivityPause where RegistryNumber = '56032061' order by DateCreated'''
        # sql = '''select * from APR_SRBIJA.dbo.ActivityPause where RegistryNumber = '60587744' order by DateCreated'''
        # sql = '''select * from APR_SRBIJA.dbo.ActivityPause where RegistryNumber = '63741370' order by DateCreated'''
        # sql = '''select * from APR_SRBIJA.dbo.ActivityPause where RegistryNumber = '56441921' order by DateCreated'''
        # sql = '''select * from APR_SRBIJA.dbo.ActivityPause where RegistryNumber = '63045225' order by DateCreated'''
        sql = '''select * from APR_SRBIJA.dbo.ActivityPause where PauseEndDate is null order by DateCreated'''
        data = db_apr.select('ActivityPause', sql=sql)
        for row in data:
            id = row.get('id')
            registrynumber = row.get('registrynumber')
            datecreated = row.get('datecreated').strftime("%Y-%m-%d")
            pausestartdate = row.get('pausestartdate')
            print("orginal", registrynumber, "datecreated", datecreated, "pausestartdate", pausestartdate)

            # sql = f'''
			# select  NATIONAL_IDENTIFICATION_NUMBER, MIN(GROUP_DATE_CHANGE) GROUP_DATE_CHANGE, DATE_CREATED from (
            # select  NATIONAL_IDENTIFICATION_NUMBER, GROUP_DATE_CHANGE, min(cast(DATE_CREATED as date)) DATE_CREATED  from (
            # select NATIONAL_IDENTIFICATION_NUMBER, GROUP_DATE_CHANGE, cast(DATE_CREATED as date) DATE_CREATED from Bisnode_RS.[APR_WS].[ENTITY_CHANGES] where  GROUP_ID = '1065'
            # union all
            # select  NATIONAL_IDENTIFICATION_NUMBER, GROUP_DATE_CHANGE,cast(DATE_CREATED as date) DATE_CREATED  from Bisnode_RS.[APR_WS].[ENTITY_CHANGES_HISTORY] where GROUP_ID = '1065') a
            # where NATIONAL_IDENTIFICATION_NUMBER = '{registrynumber}'
            # group by NATIONAL_IDENTIFICATION_NUMBER, GROUP_DATE_CHANGE) b
			# group by NATIONAL_IDENTIFICATION_NUMBER, DATE_CREATED
            # order by GROUP_DATE_CHANGE
            # '''
            # changes_data = db_apr.select('ENTITY_CHANGES', sql=sql)

            sql = f'''
            select registry_number national_identification_number, change_date group_date_change from(
            select distinct registry_number, change_date from [APR_SRBIJA].[dbo].[CompanyChanges] 
            where [change_id] = '1065' and registry_number = '{registrynumber}'
            ) a
            order by change_date
            '''
            changes_data = db_apr.select('CompanyChanges', sql=sql)
            cnt = 0
            status = 0

            # for change in changes_data:
            #     registry_number = change.get('national_identification_number')
            #     date_created = change.get('date_created')
            #     pause_start_date = change.get('group_date_change').strftime("%Y-%m-%d")
            #     # print(registry_number, date_created, pause_start_date)
            #
            #     # when pause_start_date dates match and date_created doesnt match
            #     if status == 0 and pause_start_date == pausestartdate:
            #         changes_len = len(changes_data)
            #         if changes_len > cnt+1:
            #             next_change = changes_data[cnt + 1]
            #             end_date = next_change.get('date_created')
            #             print("> pausedate", registry_number, date_created, end_date)
            #             update_status = db_apr.update('ActivityPause', {'id': id, 'PauseEndDate': None}, {'PauseEndDate': end_date}, freeze=True)
            #             status = 1
            #             if update_status != 200:
            #                 print("update failed check logs")
            #                 exit()
            #
            #     # when date_created dates match
            #     if status == 0 and datecreated == date_created:
            #         changes_len = len(changes_data)
            #         if changes_len > cnt+1:
            #             next_change = changes_data[cnt+1]
            #             end_date = next_change.get('date_created')
            #             print(">datecreated", registry_number, date_created, end_date)
            #             update_status = db_apr.update('ActivityPause', {'id': id, 'PauseEndDate': None}, {'PauseEndDate': end_date}, freeze=True)
            #             status = 1
            #             if update_status != 200:
            #                 print("update failed check logs")
            #                 exit()
            #     cnt += 1

            # when pause_start_date and date_created doesnt match but we need to find change date closest to pause_start_date
            cnt = 0
            if status == 0 and len(changes_data) > 1:
                difference = {}
                for change in changes_data:
                    try:
                        day_diff = pd.to_datetime(row.get('pausestartdate'), yearfirst=True)-change.get('group_date_change')
                    except:
                        continue
                    # print(day_diff)
                    days = math.floor(((day_diff.total_seconds() / 60) / 60) / 24)
                    if day_diff.total_seconds() > 0 or days >= -7:
                        difference.update({abs(day_diff.total_seconds()): cnt})
                    cnt += 1
                if difference:
                    min_diff = min(difference.keys())
                    days = math.floor(((min_diff/60)/60)/24)
                    if days < 14:
                        key_to_use = difference.get(min_diff)
                        end_date = key_to_use+1
                        changes_len = len(changes_data)
                        if changes_len > end_date:
                            next_change = changes_data[end_date]
                            end_date = next_change.get('group_date_change')
                            print("> closest date", id, registrynumber, pausestartdate, end_date)
                            update_status = db_apr.update('ActivityPause', {'id': id, 'PauseEndDate': None}, {'PauseEndDate': end_date}, freeze=True)
                            status = 1
                            if update_status != 200:
                                print("update failed check logs")
                                exit()
                    # else:
                    #     print("differance between dates it is bigger than 14 days. cant be matched")


# TODO: remove all duplicate hash values that are in final database known cases are search_info and basic_info tables


if __name__ == '__main__':
    # SYNC SearchInfo data
    # sif = SearchInfoFill()
    # sif.update_sha_apr_subjekti()
    # sif.update_search_sha_ms()
    # sif.archive_data_search_info()
    # sif.first_transfer_apr_subjekti()
    # sif.mix_new_data()

    # SYNC BasicInfo data
    # bif = BasicInfoFill()
    # bif.update_sha_apr_subjekti()
    # bif.update_basic_sha_ms()
    # bif.archive_data_basic_info()
    # bif.first_transfer_apr_subjekti()
    # bif.mix_new_data()

    # SYNC ActivityPause data
    ap = ActivityPause()
    # ap.add_history_records()
    # ap.update_sha_activity()
    ap.calculate_endate()